/**
 * Project: MentDB
 * Description: Mentalese Database Engine
 * Website: http://www.mentdb.org
 * Twitter: https://twitter.com/mentalese_db
 * LinkedIn: https://www.linkedin.com/company/mentdb
 * Facebook: https://www.facebook.com/mentdb
 * Author: Jimmitry Payet
 * Website: http://www.jpayet.re
 * Mail: contact@jpayet.re
 * Locality: Reunion Island (French)
 */

package re.jpayet.mentdb.core.db.file.data;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

//Manage the data file
public class DataFile {
	
	public static int BLOCK_SIZE = 0;
	public static String EMPTY_BLOCK = "";
	public static int ADD_TO_BLOCK = 24;//+2;+2bytes+length(65500)+length(9,223,372,036,854,775,807)=26
	public static long MAX_BLOCK = 0;
	public static int BLOCK_SIZE_ADD_TO_BLOCK = 0;
	
	//Local properties
	public String fileName = "";
	public int id = 0;
	public RandomAccessFile dataFile = null;
	public UnusedFile unusedFile = null;
	
	//Constructor
	public DataFile(int id, String fileName) {

		this.id = id;
		this.fileName = fileName;
		
	}
	
	//Get the max block size of the data file
	public long current_max_block_size() throws IOException {
		
		//Initialization
		long nb = (dataFile.length()/(BLOCK_SIZE_ADD_TO_BLOCK));
		
		//return the current max size
		if (nb<DataFile.MAX_BLOCK) return DataFile.MAX_BLOCK;
		else return nb;
		
	}
	
	//Get the size of the data file
	public long get_size() throws IOException {
		
		return dataFile.length();
		
	}
	
	//Get the current block size of the data file
	public long current_block_size() throws IOException {
		
		//return the current size
		return (dataFile.length()/(BLOCK_SIZE_ADD_TO_BLOCK));
		
	}
	
	//Get the percent of used of the data file
	public int percent_used() throws IOException {
		
		//return the current max size
		return (int) (((current_block_size())*100/current_max_block_size()));
		
	}
	
	//Get the per30 of used of the data file
	public String per30_used() throws IOException {
		
		//Initialization
		int per30 = (int) (((current_block_size()*30)/current_max_block_size()));
		
		//return the per20 in string format
		return new String(new char[per30]).replace('\0', '|')+
				new String(new char[30-per30]).replace('\0', '-');
		
	}
	
	//Get all blocks into the data file
	public String getAllBlocks(long position) throws IOException {
		
		//Initialization
		String result = "";
		int counter = 1;
		
		try {
			
			while (true) {
				
				if (counter==26) {
					
					break;
					
				}
				
				result += "<br>"+counter+" Position="+position+" "+Block.read(this, position)+"<br>";
				
				position+= (BLOCK_SIZE_ADD_TO_BLOCK);
				counter++;
				
			}
			
		} catch (Exception e) {};
		
		if (!result.equals("")) {
			
			result = result.substring(4);
			
		}
		
		return result;
		
	}
	
	//Open the data file
	public void open() throws FileNotFoundException {
		
		//Create the data/data/id folder if does not exist
		if (!(new File("data"+File.separator+"data"+File.separator+id).exists())) {
			
			(new File("data"+File.separator+"data"+File.separator+id)).mkdir();
			
		}

		//Open the data file
		dataFile = new RandomAccessFile(fileName, "rw");
		unusedFile = new UnusedFile(id);
		unusedFile.open();
		
	}
	
	//Close the data file
	public void close() throws IOException {

		//Close the data file
		dataFile.close();
		unusedFile.close();
		
	}

}
