/**
 * Project: MentDB
 * Description: Mentalese Database Engine
 * Website: http://www.mentdb.org
 * Twitter: https://twitter.com/mentalese_db
 * LinkedIn: https://www.linkedin.com/company/mentdb
 * Facebook: https://www.facebook.com/mentdb
 * Author: Jimmitry Payet
 * Website: http://www.jpayet.re
 * Mail: contact@jpayet.re
 * Locality: Reunion Island (French)
 */

package re.jpayet.mentdb.core.entity.relation;

import java.util.LinkedHashMap;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

import re.jpayet.mentdb.core.basic.BasicNode;
import re.jpayet.mentdb.core.brain.Brain;
import re.jpayet.mentdb.core.cognition.CognSearch;
import re.jpayet.mentdb.core.command.Command;
import re.jpayet.mentdb.core.command.MQLValue;
import re.jpayet.mentdb.core.db.Record;
import re.jpayet.mentdb.core.db.file.experience.ExperienceManagerL2;
import re.jpayet.mentdb.core.db.file.node.NodeManager;
import re.jpayet.mentdb.core.entity.circle.CircleManager;
import re.jpayet.mentdb.core.entity.concentration.ConcentrationManager;
import re.jpayet.mentdb.core.entity.thought.ThoughtManager;
import re.jpayet.mentdb.core.entity.word.WordManager;
import re.jpayet.mentdb.core.fx.MathFx;
import re.jpayet.mentdb.core.search.DeepSearchEngine;
import re.jpayet.mentdb.core.sequence.SequenceManager;
import re.jpayet.mentdb.core.session.SessionThread;
import re.jpayet.mentdb.core.session.env.EnvManager;
import re.jpayet.mentdb.core.statement.Statement;
import re.jpayet.mentdb.tools.Misc;

public class RelationManager {
	
	//Get info
	public static String getInfo(String user, String relationId) throws Exception {
		
		String result = relationId+": ";
		
		JSONArray words = RelationManager.showWords(user, relationId);
		for(int i=0;i<words.size();i++) {
			
			Random ran = new Random();
			JSONArray wlist = ((JSONArray) words.get(i));
			String selectedWord = (""+((JSONObject) wlist.get(ran.nextInt(wlist.size()))).get("word"));
			
			String[] splitedWord = selectedWord.substring(2, selectedWord.length()-1).split(" ");
			
			for(int z=0;z<splitedWord.length;z++) {
				
				//Get the splited word
				String w = splitedWord[z];
				
				result += w+" ";
				
			}
			
		}
		result += "\n";

		result += "thoughts: ";
		JSONArray thoughts = RelationManager.showThoughtsRecursively(user, relationId);
		for(int i=0;i<thoughts.size();i++) {
			String key = thoughts.get(i)+"";
			result += key+" ";
		}
		result += "\n";
		
		result += "nodes: ";
		thoughts = RelationManager.showThoughtNodes(user, relationId);
		for(int i=0;i<thoughts.size();i++) {
			String key = thoughts.get(i)+"";
			result += key+" ";
		}
		result += "\n";
		
		return result;
		
	}
	
	//Show network
	public static int firstNotSomethingPosition(String user, String relationId) throws Exception {
		
		//Initialization
		int result = 0;
		JSONArray a = RelationManager.showThoughtNodes(user, relationId);
		
		for(int i=0;i<a.size();i++) {
			
			//Get the current value
			String curVal = a.get(i)+"";
			if (!curVal.equals("TH[]")) {
				result = i+1;
				break;
			}
			
		}
		
		return result;
		
	}
	

	
	//Show network
	@SuppressWarnings("unchecked")
	public static void showNetwork_getTabLinks(String user, String relationId, String type, JSONObject tlResult) throws Exception {
		
		//Get all tab links
		JSONArray tl = RelationManager.showTabLinks(user, relationId, type, ConcentrationManager.getConcentrationDepth("C[relation]"));
		for(int i=0;i<tl.size();i++) {
			
			String key = ""+tl.get(i);
			key = Misc.atom(key, 1, " ").substring(1);
			
			if (tlResult.containsKey(key)) {
				tlResult.put(key, ((Integer) tlResult.get(key))+1);
			} else {
				tlResult.put(key, 1);
			}
			
		}
		
	}
	
	//Show network
	@SuppressWarnings("unchecked")
	public static JSONObject showNetwork(String user, String relationId) throws Exception {
		
		//Initialization
		JSONObject result = new JSONObject();
		JSONObject tlResult = new JSONObject(new LinkedHashMap<String, Integer>());

		showNetwork_getTabLinks(user, relationId, "G", tlResult);
		showNetwork_getTabLinks(user, relationId, "L", tlResult);
		showNetwork_getTabLinks(user, relationId, "I", tlResult);
		showNetwork_getTabLinks(user, relationId, "O", tlResult);
		showNetwork_getTabLinks(user, relationId, "U", tlResult);
		showNetwork_getTabLinks(user, relationId, "S", tlResult);
		showNetwork_getTabLinks(user, relationId, "D", tlResult);
		showNetwork_getTabLinks(user, relationId, "A", tlResult);
		showNetwork_getTabLinks(user, relationId, "R", tlResult);
		
		//Initialization
		JSONObject thResult = new JSONObject(new LinkedHashMap<String, Integer>());
		
		//Get all thoughts
		JSONArray tl = RelationManager.showThoughtNodes(user, relationId);
		
		for(int i=0;i<tl.size();i++) {
			
			String key = ""+tl.get(i);
			
			if (key.startsWith("R")) {
				
				//Only if this is a relation
				
				if (thResult.containsKey(key)) {
					thResult.put(key, ((Integer) thResult.get(key))+1);
				} else {
					thResult.put(key, 1);
				}
				
			}
			
		}
		
		result .put("tl", tlResult);
		result .put("th", thResult);
		
		return result;
		
	}
	
	//Show words with thought
	@SuppressWarnings("unchecked")
	public static JSONArray showWordThoughts(String user, String relationId, int[] toRemove) throws Exception {
		
		//Initialization
		JSONArray result = new JSONArray();
		
		//Get thoughts
		JSONArray thoughts = showThoughtNodes(user, relationId);
		
		for(int i=0;i<toRemove.length;i++) {
			thoughts.set(toRemove[i], "");
		}
		
		JSONObject bd = Record.getNode(relationId);
		String lang = ""+bd.get("l");
		
		//Get all words
		for(int i=0;i<thoughts.size();i++) {
			
			if (!(thoughts.get(i)+"").equals("")) {
			
				if (!(""+thoughts.get(i)).startsWith("R")) result.add(ThoughtManager.getWords(""+thoughts.get(i), lang));
				else {
					JSONArray array = new JSONArray();
					JSONObject obj = new JSONObject();
					obj.put("lang", lang);
					
					//A TESTER ...
					String w = showWordsRecurcive(user, ""+thoughts.get(i));
					
					if (w.endsWith(" ")) w = w.substring(0, w.length()-1);
					obj.put("word", "W["+w+"]");
					array.add(obj);
					result.add(array);
				}
				
			}

		}

		return result;
		
	}
	
	//Show words in a relation
	@SuppressWarnings("unchecked")
	public static JSONArray showWords(String user, String relationId) throws Exception {
		
		//Initialization
		JSONArray result = new JSONArray();
		
		//Get thoughts
		JSONArray thoughts = showThoughtNodes(user, relationId);
		
		JSONObject bd = Record.getNode(relationId);
		String lang = ""+bd.get("l");
		
		//Get all words
		for(int i=0;i<thoughts.size();i++) {
			
			if (!(""+thoughts.get(i)).startsWith("R")) result.add(ThoughtManager.getWords(""+thoughts.get(i), lang));
			else {
				JSONArray array = new JSONArray();
				JSONObject obj = new JSONObject();
				obj.put("lang", lang);
				
				String w = showWordsRecurcive(user, ""+thoughts.get(i));
				if (w.endsWith(" ")) w = w.substring(0, w.length()-1);
				obj.put("word", "W["+w+"]");
				array.add(obj);
				result.add(array);
			}

		}

		return result;
		
	}
	
	//Show words in a relation
	public static String showWordsRecurcive(String user, String relationId) throws Exception {
		
		//Initialization
		String result = "";
		
		//Get thoughts
		JSONArray thoughts = showThoughtNodes(user, relationId);
		
		JSONObject bd = Record.getNode(relationId);
		String lang = ""+bd.get("l");
		
		//Get all words
		for(int i=0;i<thoughts.size();i++) {
			
			if (!(""+thoughts.get(i)).startsWith("R")) {
				
				JSONArray wjson = ThoughtManager.getWords(""+thoughts.get(i), lang);
				int pos = Integer.parseInt(MathFx.random(""+wjson.size()));
				
				String w = ((JSONObject) wjson.get(pos)).get("word")+"";
				result +=w.substring(2, w.length()-1)+" ";
				
			} else {
				result += showWordsRecurcive(user, ""+thoughts.get(i));
			}

		}
		
		return result;
		
	}
	
	//Show thought nodes in a relation
	@SuppressWarnings("unchecked")
	public static JSONArray showThoughtNodes(String user, String relationId) throws Exception {
		
		//Initialization
		JSONArray result = new JSONArray();
		
		//Get the node
		JSONObject recNode = Record.getNode(relationId);
		
		if (recNode==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		String[] thoughts = (""+recNode.get("lth")).split(" ", -1);
		
		for(int i=0;i<thoughts.length;i++) {
			
			result.add(thoughts[i]);
			
		}
		
		return result;
		
	}
	
	//Show thoughts recursively in a relation
	@SuppressWarnings("unchecked")
	public static JSONArray showThoughtsRecursively(String user, String relationId) throws Exception {
		
		//Initialization
		JSONArray result = new JSONArray();
		
		//Get the node
		JSONObject recNode = Record.getNode(relationId);
		
		if (recNode==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		String[] thoughts = (""+recNode.get("lth")).split(" ", -1);
		
		for(int i=0;i<thoughts.length;i++) {
			
			if (thoughts[i].startsWith("T")) result.add(thoughts[i]);
			else {
				JSONArray array = showThoughtsRecursively(user, thoughts[i]);
				for(int z=0;z<array.size();z++) {
					result.add(array.get(z)+"");
				}
			}
			
		}
		
		return result;
		
	}
	
	//Show a relation
	@SuppressWarnings("unchecked")
	public static JSONArray listTabLinks(String user, String relationId) throws Exception {
		
		//Initialization
		JSONArray result = new JSONArray();
		
		//Get the node
		JSONObject recNode = Record.getNode(relationId);
		
		if (recNode==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		long nb = (Long) recNode.get("n");
		
		for(int i=1;i<=nb;i++) {
			
			result.add("["+relationId+" "+i+"/"+nb+"]");
			
		}
		
		return result;
		
	}
	
	//Show a translation of a relation
	@SuppressWarnings("unchecked")
	public static String translate(String user, String relationId, String lang) throws Exception {
		
		//Initialization
		Random ran = new Random();
		String selectedRelationId = relationId;
		JSONArray positives = new JSONArray();
		String parentRelationId = null;
		
		//Circle
		JSONObject circle = CircleManager.showCircle(user, relationId);
		Set<String> circleKeyset = circle.keySet();
		for(String key : circleKeyset) {

			String curlang = Misc.atom(key, 1, " ");
			
			if (curlang.equals(lang)) {
				
				JSONObject o = new JSONObject();
				o.put("type", "direct");
				o.put("key", key);
				positives.add(o);
				
			}
			
		}
		
		//Try to get the parent id
		JSONObject recRelation = Record.getNode(relationId);
		
		if (recRelation.containsKey("p")) {
			
			parentRelationId = ""+recRelation.get("p");
			
			//Circle
			JSONObject parentCircle = CircleManager.showCircle(user, parentRelationId);
			Set<String> parentCircleKeyset = parentCircle.keySet();
			for(String key : parentCircleKeyset) {
				
				String curlang = Misc.atom(key, 1, " ");
				
				if (curlang.equals(lang)) {
					
					JSONObject o = new JSONObject();
					o.put("type", "parent");
					o.put("key", key);
					o.put("positions", parentCircle.get(key)+"");
					positives.add(o);
					
				}
				
			}
			
		}
		
		if (positives.size()>0) {
		
			//Generate a random index
			JSONObject o = (JSONObject) positives.get(ran.nextInt(positives.size()));
			
			if (((String) o.get("type")).equals("direct")) {
				
				selectedRelationId = Misc.atom(((String) o.get("key")), 3, " ");
				
			} else {
				
				//Get thoughts
				JSONArray thoughts = showThoughtNodes(user, Misc.atom(((String) o.get("key")), 3, " "));
				String[] positionValues = ((String) o.get("positions")).split(" ");
				
				//Build the sentence from the parent relation
				String result = "";
				
				//Get and parse all ids
				JSONArray ids = RelationManager.listTabLinks(user, relationId);
				for(int i=0;i<ids.size();i++) {
					
					String currentId = (String) ids.get(i);
					
					JSONObject idNode = Record.getNode(currentId);
					
					JSONObject udRelation = (JSONObject) idNode.get("ud");
					
					String ud = "";
					for(Object udObj : udRelation.keySet()) {
						ud = (String) udObj;
					}
					
					if (ud.startsWith("[")) {
						
						int cur_index = Integer.parseInt(Misc.atom(Misc.atom(ud, 2, " "), 1, "/"))-1;
						
						String[] blockPosition = positionValues[cur_index].split("|");
						
						for(int iBlock=0;iBlock<blockPosition.length;iBlock++) {
							
							int parent_cur_index = Integer.parseInt(blockPosition[iBlock])-1;
							
							if (!(""+thoughts.get(parent_cur_index)).startsWith("R")) {
								
								String selectedWord = (""+((JSONObject) ThoughtManager.getWords(""+thoughts.get(parent_cur_index), lang).get(0)).get("word"));
								
								String[] splitedWord = selectedWord.substring(2, selectedWord.length()-1).split(" ");
								
								for(int z=0;z<splitedWord.length;z++) {
									
									//Get the splited word
									String w = splitedWord[z];
									
									result+=" "+w;
									
								}
								
							} else {
								
								String w = showWordsRecurcive(user, ""+thoughts.get(parent_cur_index));
								if (w.endsWith(" ")) w = w.substring(0, w.length()-1);
								
								result+=" "+w;
								
							}
							
						}
						
					} else {
						
						if (!ud.startsWith("R")) {
							
							String selectedWord = (""+((JSONObject) ThoughtManager.getWords(ud, lang).get(0)).get("word"));
							
							String[] splitedWord = selectedWord.substring(2, selectedWord.length()-1).split(" ");
							
							for(int z=0;z<splitedWord.length;z++) {
								
								//Get the splited word
								String w = splitedWord[z];
								
								result+=" "+w;
								
							}

						} else {
							
							String w = translate(user, ud, lang);
							
							result+=" "+w;
							
						}
						
					}
					
				}
				
				if (!result.equals("")) result = result.substring(1);
				
				return result;
				
			}
			
		}
		
		//Build the sentence
		String result = "";
		JSONArray words = showWords(user, selectedRelationId);
		for(int i=0;i<words.size();i++){
			
			JSONArray wlist = ((JSONArray) words.get(i));

			String selectedWord = (""+((JSONObject) wlist.get(ran.nextInt(wlist.size()))).get("word"));
			
			String[] splitedWord = selectedWord.substring(2, selectedWord.length()-1).split(" ");
			
			for(int z=0;z<splitedWord.length;z++) {
				
				//Get the splited word
				String w = splitedWord[z];
				
				result+=" "+w;
				
			}
			
		}
		
		if (!result.equals("")) result = result.substring(1);
		
		return result;
		
	}
	
	//Show sentence
	public static String showSentence(String user, String relationId) throws Exception {

		if (relationId.length()==0) {

			throw new Exception("Sorry, the relation id is required.");

		}
		
		//Try to get the relation
		JSONObject recRelation = Record.getNode(relationId);

		if (recRelation==null) {

			throw new Exception("Sorry, the relation "+relationId+" does not exist.");

		}
		
		String result = "";
		JSONArray words = showWords(user, relationId);
		for(int i=0;i<words.size();i++){
			
			JSONArray wlist = ((JSONArray) words.get(i));
			Random ran = new Random();
			String selectedWord = (""+((JSONObject) wlist.get(ran.nextInt(wlist.size()))).get("word"));
			
			String[] splitedWord = selectedWord.substring(2, selectedWord.length()-1).split(" ");
			
			for(int z=0;z<splitedWord.length;z++) {
				
				//Get the splited word
				String w = splitedWord[z];
				
				result+=" "+w;
				
			}
			
		}
		
		if (!result.equals("")) result = result.substring(1);
		
		return result;
		
	}
	
	//Show sentence
	public static String showThoughtSentence(String user, String relationId, int[] toRemove) throws Exception {
		
		//Try to get the relation
		JSONObject recRelation = Record.getNode(relationId);

		if (recRelation==null) {

			throw new Exception("Sorry, the relation "+relationId+" does not exist.");

		}
		
		String result = "";
		
		JSONArray words = showWordThoughts(user, relationId, toRemove);
		for(int i=0;i<words.size();i++){
			
			JSONArray wlist = ((JSONArray) words.get(i));
			Random ran = new Random();
			String selectedWord = (""+((JSONObject) wlist.get(ran.nextInt(wlist.size()))).get("word"));
			
			String[] splitedWord = selectedWord.substring(2, selectedWord.length()-1).split(" ");
			
			for(int z=0;z<splitedWord.length;z++) {
				
				//Get the splited word
				String w = splitedWord[z];
				
				result+=" "+w;
				
			}
			
		}
		
		if (!result.equals("")) result = result.substring(1);
		
		return result;
		
	}
	
	//Delete a basic relation
	public static void delete(String user, String relationId) throws Exception {

		if (relationId.length()==0) {

			throw new Exception("Sorry, the relation id is required.");

		}
		
		//Get all tab link id into a relation
		JSONArray tabLinkIds = RelationManager.listTabLinks(user, relationId);
		String type = relationId.substring(1, 2);
		
		//Parse all tab link id in the relation
		for(int i=0;i<tabLinkIds.size();i++) {

			//Get the current parent id
			String relationTL = tabLinkIds.get(i)+"";
			JSONObject bd = Record.getNode(relationTL);
			String parentId = "";
			JSONObject ud = ((JSONObject) bd.get("ud"));
			for(Object key : ud.keySet()) {
				parentId = (String)key;
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("G")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (G).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("L")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (L).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("I")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (I).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("O")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (O).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("U")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (U).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("S")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (S).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("D")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (D).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("A")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (A).");
				
			}
			
			if (((JSONObject) ((JSONObject) bd.get("fl")).get("R")).get("ftl")!=null) {

				throw new Exception("Sorry, there is a sub relation for "+relationTL+" (R).");
				
			}

			ExperienceManagerL2.deleteNodeTabLink(user, parentId, relationTL, type);

		}
		
		JSONObject bd = Record.getNode(relationId);
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("G")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("L")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("I")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("O")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("U")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("S")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("D")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("A")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		if (((JSONObject) ((JSONObject) bd.get("fl")).get("R")).get("ftl")!=null) {

			throw new Exception("Sorry, the relation is already used by another.");
			
		}
		
		//Delete the relation
		Record.remove(user, relationId);
		
	}

	//Stimulate a relation
	@SuppressWarnings("unchecked")
	public static void stimulate(String user, String relationId, boolean stimulateTheParent) throws Exception {

		if (relationId.length()==0) {

			throw new Exception("Sorry, the relation id "+relationId+" is not valid [required].");

		}

		//Try to get the relation
		JSONObject recRelation = Record.getNode(relationId);

		if (recRelation==null) {

			throw new Exception("Sorry, the relation "+relationId+" does not exist.");

		}
		
		//Stimulation +1
		recRelation.put("w", Integer.parseInt(""+recRelation.get("w"))+1);
		Record.update(user, relationId, recRelation.toJSONString());
		String parentRelationId = "";
		
		//Only if not a parent relation
		if (stimulateTheParent) {
			
			//Stimulate the parent relation
			
			if (recRelation.containsKey("p")) {
				
				parentRelationId = ""+recRelation.get("p");
				
			}
			
		}
		
		JSONArray tabLinkIds = RelationManager.listTabLinks(user, relationId);
		String type = relationId.substring(1, 2);
		
		//Parse all tab link id in the relation
		for(int i=0;i<tabLinkIds.size();i++) {

			//Get the current parent id
			String relationTL = tabLinkIds.get(i)+"";
			JSONObject bd = Record.getNode(relationTL);
			String parentId = "";
			JSONObject ud = ((JSONObject) bd.get("ud"));
			for(Object key : ud.keySet()) {
				parentId = (String)key;
			}
			
			Long parentPosition = -1L;
			if (bd.containsKey("pp")) {
				parentPosition = (Long) bd.get("pp");
			}
			
			ExperienceManagerL2.deleteNodeTabLink(user, parentId, relationTL, type);
			
			ExperienceManagerL2.addStimulationNodeTabLink(user, parentId, relationTL, ((JSONObject) bd.get("fl")), parentPosition, type);
			
		}
		
		if (!parentRelationId.equals("")) {
			stimulate(user, parentRelationId, false);
		}

	}
	
	//Create a basic relation from a sentence
	public static String createFromSentence(SessionThread session, EnvManager env, String user, String sentence, String type, String lang) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}
		
		//Split the sentence
		Vector<MQLValue> s = Misc.splitCommand(sentence).get(0);
		String thoughts = "";
		
		//Parse all words
		for(int i=0;i<s.size();i++) {
			
			String wordOrRelation = s.get(i).value;
			boolean isNotRelation = true;
			
			if (wordOrRelation.length()>4 && wordOrRelation.substring(0, 3).startsWith("R") && wordOrRelation.substring(0, 3).endsWith("[") && wordOrRelation.endsWith("]") && NodeManager.existNode(user, wordOrRelation)) {
				isNotRelation = false;
				thoughts += " "+wordOrRelation;
			} else if (wordOrRelation.length()>=4 && wordOrRelation.substring(0, 3).startsWith("TH") && wordOrRelation.substring(0, 3).endsWith("[") && wordOrRelation.endsWith("]") && NodeManager.existNode(user, wordOrRelation)) {
				isNotRelation = false;
				thoughts += " "+wordOrRelation;
			}
			
			if (isNotRelation) {
			
				//Create the word if not exist
				if (WordManager.exist(user, wordOrRelation, lang)) {
					
					//Get all thoughts
					JSONArray a = ThoughtManager.list(user, wordOrRelation, lang);
					String th = "";
					for(int j=0;j<a.size();j++) {
						th += "|"+((String) a.get(j));
					}
					if (!th.equals("")) th = th.substring(1);
					thoughts += " "+th;
					
				} else {
					
					thoughts += " "+WordManager.create(user, wordOrRelation, "", lang);
				
				}
				
			}
			
		}

		if (!thoughts.equals("")) thoughts = thoughts.substring(1);
		
		String mql = "relation create \""+thoughts+"\" "+type+" "+lang+";";
		
		try {
			
			return Statement.eval(session, mql, env);
			
		} catch (Exception e) {
			
			throw new Exception(e.getMessage()+" ["+mql+"]");
			
		}
		
	}
	
	//Create a basic relation
	@SuppressWarnings("unchecked")
	public static String create(String user, String thoughtOrRelations, String lang, String type) throws Exception {
		
		if (thoughtOrRelations.length()==0) {

			throw new Exception("Sorry, at least one thought/relation id needed.");

		}
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}
		
		//The default
		String relationId = "R"+type+"["+SequenceManager.incr(user, "relation-"+type)+"]";
		while (Record.getNode(relationId)!=null) {
			
			relationId = "R"+type+"["+SequenceManager.incr(user, "relation-"+type)+"]";
			
		}
		
		int nbThought = Misc.size(thoughtOrRelations, " ");
		
		//Parse all words in the relation
		for(int i=1;i<=nbThought;i++) {

			//Get the current thought
			String id = Misc.atom(thoughtOrRelations, i, " ");

			if (Record.getNode(id)==null) {

				throw new Exception("Sorry, the thought/relation id "+id+" does not exist.");

			} else {
				
				ExperienceManagerL2.addNodeTabLink(user, id, "["+relationId+" "+i+"/"+nbThought+"]", -1, type);
			
			}

		}
		
		//Add the record
		BasicNode br = null;
		br = new BasicNode(relationId, BasicNode.RELATION_TYPE, lang);

		br.dataNode.put("n", nbThought);
		br.dataNode.put("st", new JSONArray());
		br.dataNode.put("l", lang);
		
		String[] array = thoughtOrRelations.split(" ", -1);
		String lth = "";
		for(int i=0;i<array.length;i++) {
			if (array[i].startsWith("T")) lth+=" "+array[i];
			else {
				JSONArray ar = RelationManager.showThoughtsRecursively(user, array[i]);
				for(int z=0;z<ar.size();z++) lth+=" "+ar.get(z);
			}
		}
		lth = lth.substring(1);
		br.dataNode.put("rlth", lth);
		br.dataNode.put("lth", thoughtOrRelations);
		br.dataNode.put("dco", new JSONObject());
		br.dataNode.put("c", null);
		
		Record.add(user, relationId, br.dataNode.toJSONString());
		
		return relationId;
		
	}
	
	//Get a condition
	public static JSONObject getDco(String user, String relationId, String type) throws Exception {
		
		//Get the relation tab link id
		JSONObject rec = Record.getNode(relationId);
		
		return (JSONObject) rec.get(type);
		
	}
	
	//Get an action/elseAction
	public static String getDa(String user, String relationId, String type) throws Exception {
		
		String mql = "";
		
		//Get the relation tab link id
		JSONObject rec = Record.getNode(relationId);
		
		String m = (String) rec.get(type);
		
		if (m!=null && !m.equals("")) mql = m;
		
		return mql;
		
	}
	
	//Set a condition
	@SuppressWarnings("unchecked")
	public static void setCondition(SessionThread session, String user, String relationId, String target, String mql, EnvManager env) throws Exception {
		
		relationId = Command.executeAllCommands(false, session, Misc.splitCommand(relationId), env);
		
		if (relationId.length()==0) {

			throw new Exception("Sorry, the relation id is required.");

		}

		if (target==null ||target.length()==0) {

			throw new Exception("Sorry, the target is required (R|TH1|TH2|...).");

		}
		
		if (!target.equals("R")) {
			
			if (target.length()<2) throw new Exception("Sorry, the target is not valid (R|TH1|THN).");
			
			try {
				
				Integer.parseInt(target.substring(2));
				
			} catch (Exception e) {
				throw new Exception("Sorry, the target is not valid (R|TH1|THN).");
			}
			
		}

		if (mql.length()==0) {

			throw new Exception("Sorry, the MQL condition is required.");

		}
		
		//Try to get the relation tab link id
		JSONObject recRelationId = Record.getNode(relationId);

		if (recRelationId==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		((JSONObject) recRelationId.get("dco")).put(target, mql);
		
		Record.update(user, relationId, recRelationId.toJSONString());
		
	}
	
	//Show conditions
	public static JSONObject showConditions(SessionThread session, String user, String relationId, EnvManager env) throws Exception {
		
		relationId = Command.executeAllCommands(false, session, Misc.splitCommand(relationId), env);
		
		if (relationId.length()==0) {

			throw new Exception("Sorry, the relation id is required.");

		}
		
		//Try to get the relation tab link id
		JSONObject recRelationId = Record.getNode(relationId);

		if (recRelationId==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		return ((JSONObject) recRelationId.get("dco"));
		
	}
	
	//Show conditions
	public static String showAction(SessionThread session, String user, String relationId, EnvManager env) throws Exception {
		
		relationId = Command.executeAllCommands(false, session, Misc.splitCommand(relationId), env);
		
		if (relationId.length()==0) {

			throw new Exception("Sorry, the relation id is required.");

		}
		
		//Try to get the relation tab link id
		JSONObject recRelationId = Record.getNode(relationId);

		if (recRelationId==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		return (String) recRelationId.get("da");
		
	}
	
	//Check a condition exist
	public static boolean existCondition(SessionThread session, String user, String relationId) throws Exception {
		
		try {
			if (getDco(user, relationId, "dco").size()==0) return false;
			else return true;
		} catch (Exception e) {
			return false;
		}
		
	}
	
	//Check conditions from another relation
	public static String checkConditionFromAnotherRelation(String relationId) throws Exception {
		
		return "1";
		
	}
	
	//Generate conditions
	public static String generateConditions(String relationId) throws Exception {
		
		String result = "";
		
		//Try to get the relation tab link id
		JSONObject recRelationId = Record.getNode(relationId);

		if (recRelationId==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		JSONObject dco = ((JSONObject) recRelationId.get("dco"));
		
		for(Object o : dco.keySet()) {
			
			String key = (String) o;
			
			String mql = (String) dco.get(key);
			
			result += "\nrelation set condition "+relationId+" "+key+" {"+mql+"};";
			
		}
		
		if (result.length()>0) result = result.substring(1);
		
		if (result.equals("")) throw new Exception("Sorry, no condition found.");
		else return result;
		
	}
	
	//Generate conditions
	public static String generateAction(String relationId) throws Exception {
		
		String result = "";
		
		//Try to get the relation tab link id
		JSONObject recRelationId = Record.getNode(relationId);

		if (recRelationId==null) {

			throw new Exception("Sorry, the relation id "+relationId+" does not exist.");

		}
		
		String mql = ((String) recRelationId.get("da"));
		
		result = "relation set action "+relationId+" {"+mql+"};";
				
		if (mql==null) throw new Exception("Sorry, no action found.");
		else return result;
		
	}
	
	//Check a condition
	@SuppressWarnings("unchecked")
	public static JSONObject checkCondition(SessionThread session, String user, EnvManager env, JSONArray stimulation_tab, String fromUser) throws Exception {
		
		//Initialization
		JSONArray process = ((JSONArray) stimulation_tab.get(stimulation_tab.size()-1));
		JSONObject firstStimulation = (JSONObject) process.get(0);
		JSONObject lastStimulation = (JSONObject) process.get(process.size()-1);
		String sentence = (String) firstStimulation.get("input");
		JSONArray input = (JSONArray) firstStimulation.get("reader_found_word");
		String relationId = (String) lastStimulation.get("relation");
		
		JSONObject objResult = new JSONObject();
		env.set("[$]", sentence);
		String[] inWords = sentence.split(" ");
		
		JSONArray thoughtNodes = RelationManager.showThoughtNodes(session.user, relationId);
		
		objResult.put("thoughtNodes", thoughtNodes);

		env.set("[R]", relationId);
		env.set("[fromUser]", fromUser);
		
		//Set input variables
		for(int i=1;i<=thoughtNodes.size();i++) {
			env.set("[R"+i+"]", thoughtNodes.get(i-1)+"");
		}
		
		//Set input word
		for(int i=0;i<input.size();i++) {
			
			env.set("[W"+i+"]", input.get(i)+"");
		}
		
		JSONArray thoughts = RelationManager.showThoughtsRecursively(session.user, relationId);
		
		JSONArray groups = new JSONArray();
		
		//Parse all thoughts
		for(int i=1;i<=thoughts.size();i++) {
			
			String th = thoughts.get(i-1)+"";
			
			if (th.equals("TH[]")) {
				
				JSONObject o = new JSONObject();
				o.put("key", i-1);
				o.put("type", "TH");
				o.put("th", new JSONArray());
				groups.add(o);
				
			} else if (groups.size()==0 || (((JSONObject) groups.get(groups.size()-1)).get("type")+"").equals("TH")) {
				
				JSONObject o = new JSONObject();
				o.put("key", i-1);
				o.put("type", "W");
				o.put("th", new JSONArray());
				groups.add(o);
				
			}
			
			if (!th.equals("TH[]")) {
				
				JSONObject o = new JSONObject();
				o.put("thoughtId", th);
				JSONArray a = ThoughtManager.getWords(th);
				JSONObject words = new JSONObject();
				for(int j=0;j<a.size();j++) {
					String curWord = (((JSONObject) a.get(j)).get("word")+"");
					curWord = curWord.substring(2, curWord.length()-1);
					words.put(curWord, j);
				}
				o.put("words", words);
				((JSONArray) ((JSONObject) groups.get(groups.size()-1)).get("th")).add(o);
				
			}
			
		}
		
		JSONObject thoughtWordFound = new JSONObject();
		
		//Cognitive search
		if (groups.size()>1) {

			//Cognitive search by probability
			CognSearch.byProbability(groups, inWords);

			int j = 1;
			for(int i=0;i<groups.size();i++) {
				
				if ((((JSONObject) groups.get(i)).get("type")+"").equals("TH")) {

					String str = CognSearch.get(i);

					env.set("[TH"+j+"]", str);
					thoughtWordFound.put("[TH"+j+"]", str);
					
					j++;
					
				}
				
			}

		}
		
		objResult.put("thoughtWordFound", thoughtWordFound);
		
		JSONObject mqls = getDco(user, relationId, "dco");
		boolean result = true;
		
		JSONObject results = new JSONObject();
		objResult.put("check_result", results);
		
		if (mqls.size()>0) {
			
			for(Object o : mqls.keySet()) {
				
				String key = (String) o;
				
				String mql = (String) mqls.get(key);
				
				String b = "0";
				try {
					b = Command.executeAllCommands(false, session, Misc.splitCommand(mql), env);
				} catch (Exception e) {
					
				}
				
				if (!b.equals("1")) {
					
					result = false;
					results.put(key, false);
					break;
					
				} else {
					
					results.put(key, true);
					
				}
				
			}
			
		}
		
		if (result) objResult.put("result", "1");
		else objResult.put("result", "0");
		
		return objResult;
		
	}
	
	//Check if an action exist
	public static boolean existAction(SessionThread session, String user, String rIdOrTabLinkId) throws Exception {
		
		if (getDa(user, rIdOrTabLinkId, "da").equals("")) return false;
		else return true;
		
	}
	
	//Contains in string
	public static int containsInString(JSONObject group, String in, int lastFound) throws Exception {
		
		JSONArray a = ((JSONArray) group.get("th"));
		
		for(int u=lastFound+1;u<a.size();u++) {

			JSONObject obj = ((JSONObject) ((JSONObject) a.get(u)).get("reader_found_word"));
			
			if (obj.containsKey(in)) {

				return u;
				
			}
			
		}
		
		return -1;
		
	}
	
	//Execute an action
	public static String executeAction(SessionThread session, String user, EnvManager env, String fromUser) throws Exception {
		
		//Initialization
		String result = "";
		JSONArray stimulation_tab = ((JSONArray) Brain.stimulation.get(fromUser));
		
		JSONArray array = (JSONArray) stimulation_tab.get(stimulation_tab.size()-1);
		String relationId = (String) (((JSONObject) array.get(array.size()-1)).get("relation"));
		JSONObject thoughtWordFound = (JSONObject) (((JSONObject) array.get(array.size()-1)).get("thoughtWordFound"));
		String sentence = (String) ((JSONObject) array.get(0)).get("input");
		JSONArray input = (JSONArray) ((JSONObject) array.get(0)).get("reader_found_word");
		env.set("[$]", sentence);
		env.set("[fromUser]", fromUser);
		
		JSONArray thoughtNodes = RelationManager.showThoughtNodes(session.user, relationId);
		env.set("[R]", relationId);
		
		//Set input variables
		for(int i=1;i<=thoughtNodes.size();i++) {
			env.set("[R"+i+"]", thoughtNodes.get(i-1)+"");
		}
		
		//Set input word
		for(int i=0;i<input.size();i++) {
			env.set("[W"+i+"]", input.get(i)+"");
		}
		
		for(Object o : thoughtWordFound.keySet()) {
			env.set((String)o, ""+thoughtWordFound.get((String)o));
		}
		
		String mql = getDa(user, relationId, "da");
		
		if (mql!=null && !mql.equals("")) {
			
			result = Command.executeAllCommands(false, session, Misc.splitCommand(mql), env);
			
		}
		
		return result;
		
	}
	
	//Set an action
	@SuppressWarnings("unchecked")
	public static void setAction(SessionThread session, String user, String rIdOrTabLinkId, String mql, EnvManager env) throws Exception {
		
		rIdOrTabLinkId = Command.executeAllCommands(false, session, Misc.splitCommand(rIdOrTabLinkId), env);
		
		if (rIdOrTabLinkId.length()==0) {

			throw new Exception("Sorry, the relation/tab link id is required.");

		}

		if (mql.length()==0) {

			throw new Exception("Sorry, the MQL action is required.");

		}
		
		//Try to get the relation tab link id
		JSONObject recRelationTabLinkId = Record.getNode(rIdOrTabLinkId);

		if (recRelationTabLinkId==null) {

			throw new Exception("Sorry, the relation/tab link id "+rIdOrTabLinkId+" does not exist.");

		}
		
		recRelationTabLinkId.put("da", mql);

		Record.update(user, rIdOrTabLinkId, recRelationTabLinkId.toJSONString());
		
	}
	
	//Check an else action exist
	public static boolean existElseAction(SessionThread session, String user, String rIdOrTabLinkId) throws Exception {
		
		if (getDa(user, rIdOrTabLinkId, "dea").equals("")) return false;
		else return true;
		
	}
	
	//Execute an else action
	public static String executeElseAction(SessionThread session, String user, String rIdOrTabLinkId, Vector<MQLValue> parameters, EnvManager env) throws Exception {
		
		//Initialization
		String result = "";
		
		//Load variables
		for(int i=0;i<parameters.size();i++) {
			env.set("[$"+(i+1)+"]", parameters.get(i).value);
		}
		
		//Clear variables
		for(int i=parameters.size();i<20;i++) {
			if (env.exist("[$"+(i+1)+"]")) {
				env.remove("[$"+(i+1)+"]");
			}
		}
		
		env.set("[R]", rIdOrTabLinkId);
		
		if (rIdOrTabLinkId.startsWith("R")) {
			
			JSONArray thoughtNodes = RelationManager.showThoughtNodes(session.user, rIdOrTabLinkId);
			for(int i=1;i<=thoughtNodes.size();i++) {
				env.set("[R"+i+"]", thoughtNodes.get(i-1)+"");
			}
			
		}
		
		String mql = getDa(user, rIdOrTabLinkId, "da");
			
		if (!mql.equals("")) {
			
			result = Command.executeAllCommands(false, session, Misc.splitCommand(mql), env);
			
		}
		
		return result;
		
	}
	
	//Set an else action
	@SuppressWarnings("unchecked")
	public static void setElseAction(SessionThread session, String user, String rIdOrTabLinkId, String mql, EnvManager env) throws Exception {
		
		rIdOrTabLinkId = Command.executeAllCommands(false, session, Misc.splitCommand(rIdOrTabLinkId), env);
		
		if (rIdOrTabLinkId.length()==0) {

			throw new Exception("Sorry, the relation/tab link id is required.");

		}

		if (mql.length()==0) {

			throw new Exception("Sorry, the MQL else action is required.");

		}
		
		//Try to get the relation tab link id
		JSONObject recRelationTabLinkId = Record.getNode(rIdOrTabLinkId);

		if (recRelationTabLinkId==null) {

			throw new Exception("Sorry, the relation/tab link id "+rIdOrTabLinkId+" does not exist.");

		}
		
		recRelationTabLinkId.put("dea", mql);

		Record.update(user, rIdOrTabLinkId, recRelationTabLinkId.toJSONString());
		
	}
	
	//Create a strategic relation
	@SuppressWarnings("unchecked")
	public static String createSubWithPosition(String user, String parentRelationId, String thoughtOrRelations, String positions, String lang, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}
		
		if (thoughtOrRelations.length()==0) {

			throw new Exception("Sorry, at least one thought id needed.");

		}
		
		//The default
		String relationId = "R"+type+"["+SequenceManager.incr(user, "relation-"+type)+"]";
		while (Record.getNode(relationId)!=null) {
			
			relationId = "R"+type+"["+SequenceManager.incr(user, "relation-"+type)+"]";
			
		}
		
		//Parse all words in the relation
		int nbThoughtOrRelation = Misc.size(thoughtOrRelations, " ");
		
		//Parse all positions in the relation
		int nbPosition = Misc.size(positions, " ");

		if (nbPosition!=nbThoughtOrRelation) {

			throw new Exception("Sorry, the number of position is not equals to the number of thought.");

		}
		
		int minPosition = 0;
		int maxPosition = 0;
		int p=1;
		try {
			for(p=1;p<=nbPosition;p++) {
				if (p==1) {
					minPosition = Integer.parseInt(Misc.atom(positions, p, " "));
					maxPosition = minPosition;
				} else {
					int curPos = Integer.parseInt(Misc.atom(positions, p, " "));
					if (curPos<minPosition) minPosition = curPos;
					if (curPos>maxPosition) maxPosition = curPos;
				}
			}
		} catch (Exception e) {
			
			throw new Exception("Sorry, the value "+Misc.atom(positions, p, " ")+" at the position "+p+" is not valid (must be an integer).");
			
		}
		
		if (minPosition<=0) {

			throw new Exception("Sorry, the min position "+minPosition+" must be > 0.");

		}
		
		//Get the parent relation
		JSONArray parentRelation = RelationManager.listTabLinks(user, parentRelationId);
		
		JSONArray parentThought = RelationManager.showThoughtNodes(user, parentRelationId);
		
		if (maxPosition>parentRelation.size()) {

			throw new Exception("Sorry, the max position "+maxPosition+" must be <= "+parentRelation.size()+".");

		}
		
		boolean atLeastOnceParent = false;
		
		//Parse all thoughts
		for(int i=1;i<=nbThoughtOrRelation;i++) {
			
			int pos = Integer.parseInt(Misc.atom(positions, i, " "));
			String pthought = ""+parentThought.get(pos-1);
			String pid = ""+parentRelation.get(pos-1);
			String thoughtId = Misc.atom(thoughtOrRelations, i, " ");
			
			if (pos!=0 && thoughtId.equals(pthought)) {
				
				if (Record.getNode(pid)==null) {
					
					throw new Exception("Sorry, "+pid+" does not exist.");

				} else {
					
					atLeastOnceParent = true;
					ExperienceManagerL2.addNodeTabLink(user, pid, "["+relationId+" "+i+"/"+nbThoughtOrRelation+"]", pos, type);
				
				}
				
			} else {
				
				if (Record.getNode(thoughtId)==null) {
					
					throw new Exception("Sorry, "+thoughtId+" does not exist.");

				} else {
					
					ExperienceManagerL2.addNodeTabLink(user, thoughtId, "["+relationId+" "+i+"/"+nbThoughtOrRelation+"]", pos, type);
				
				}
				
			}

		}
		
		if (!atLeastOnceParent) {

			throw new Exception("Sorry, the relation "+relationId+" ("+thoughtOrRelations+") must contains at least one parent "+parentRelationId+" ("+parentThought.toJSONString()+").");

		}

		//Add the record
		BasicNode br = null;
		br = new BasicNode(relationId, BasicNode.RELATION_TYPE, lang);
		
		br.dataNode.put("n", nbThoughtOrRelation);
		br.dataNode.put("st", new JSONArray());
		br.dataNode.put("l", lang);
		br.dataNode.put("p", parentRelationId);
		
		String[] array = thoughtOrRelations.split(" ", -1);
		String lth = "";
		for(int i=0;i<array.length;i++) {
			if (array[i].startsWith("T")) lth+=" "+array[i];
			else {
				JSONArray ar = RelationManager.showThoughtsRecursively(user, array[i]);
				for(int z=0;z<ar.size();z++) lth+=" "+ar.get(z);
			}
		}
		lth = lth.substring(1);
		br.dataNode.put("rlth", lth);
		br.dataNode.put("lth", thoughtOrRelations);
		br.dataNode.put("dco", new JSONObject());
		
		br.dataNode.put("c", null);
		
		Record.add(user, relationId, br.dataNode.toJSONString());
		
		return relationId;
		
	}
	
	//Get the first relation tab link
	public static String firstTabLink(String user, String relationTabLink, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}
		
		if (relationTabLink.length()==0) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" is not valid [required].");

		}

		//Try to get the record
		JSONObject rec = Record.getNode(relationTabLink);

		if (rec==null) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" does not exist.");

		}
		
		return ExperienceManagerL2.getFirstTabLink(user, relationTabLink, type);

	}
	
	//Get the last relation tab link
	public static String lastTabLink(String user, String relationTabLink, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}

		if (relationTabLink.length()==0) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" is not valid [required].");

		}

		//Try to get the record
		JSONObject rec = Record.getNode(relationTabLink);

		if (rec==null) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" does not exist.");

		}
		
		return ExperienceManagerL2.getLastTabLink(user, relationTabLink, type);

	}
	
	//Check if contains tab links
	public static boolean containsTabLink(String user, String relationId, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}

		//Initialization
		boolean result = false;
		
		//list all tab links
		JSONArray list = RelationManager.listTabLinks(user, relationId);
		
		//Parse the list
		for(int i=0;i<list.size();i++) {
			
			if (firstTabLink(user, ""+list.get(i), type)!=null) {
				result = true;
				
				break;
			}
			
		}
		
		return result;
		
	}
	
	//Get relation tab link
	public static String getTabLink(String user, String relationTabLink, int position, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}
		
		if (position>=ConcentrationManager.getConcentrationDepth("C[relation]")) {

			throw new Exception("Sorry, the position > concentration depth.");

		}
		
		JSONArray tab = showTabLinks(user, relationTabLink, type, ConcentrationManager.getConcentrationDepth("C[relation]"));
		
		if (position>=tab.size()) {

			throw new Exception("Sorry, not found tab link at the position "+position+" for the relation tab link "+relationTabLink+".");

		}
		
		return ""+tab.get(position);

	}
	
	//Find relation tab link
	public static String findTabLink(String user, String relationId, int position, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}
		
		JSONArray thoughtNodes = RelationManager.showThoughtNodes(user, relationId);
		
		int concentration = ConcentrationManager.getConcentrationDepth("C[relation]");

		if (position>=concentration) {

			throw new Exception("Sorry, the position > concentration depth.");

		}
		
		int i = 1;
		
		JSONArray tab = showTabLinks(user, "["+relationId+" "+i+"/"+thoughtNodes.size()+"]", type, concentration);

		while (tab.size()==0 && i<=thoughtNodes.size()) {
			
			tab = showTabLinks(user, "["+relationId+" "+i+"/"+thoughtNodes.size()+"]", type, concentration);
			i++;
			
		}

		if (position>=tab.size()) {

			throw new Exception("Sorry, not found tab link at the position "+position+" for the relation "+relationId+".");

		}

		return ""+tab.get(position);

	}
	
	//Show relation tab links
	@SuppressWarnings("unchecked")
	public static JSONArray showTabLinks(String user, String relationTabLink, String type, int depth) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}

		if (relationTabLink.length()==0) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" is not valid [required].");

		}

		//Try to get the record
		JSONObject rec = Record.getNode(relationTabLink);

		if (rec==null) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" does not exist.");

		}
		
		JSONArray result = new JSONArray();
		
		//Get the first
		String tl = ExperienceManagerL2.getFirstTabLink(user, relationTabLink, type);
		
		//Add only if null
		if (tl!=null) {
			
			result.add(tl);
			
			//Get the down node
			tl = ExperienceManagerL2.getDownTabLink(user, relationTabLink, tl, type);
			int i = 0;
			
			while(tl!=null && i+1<depth) {

				result.add(tl);
				
				i++;

				//Get the next down node
				tl = ExperienceManagerL2.getDownTabLink(user, relationTabLink, tl, type);
				
			}

		}
		
		return result;

	}
	
	//Show relation tab links
	@SuppressWarnings("unchecked")
	public static JSONArray showTabLinkPerceptions(String user, String relationTabLink, String type) throws Exception {
		
		if (DeepSearchEngine.allTypeList.indexOf(type)==-1) {
			
			throw new Exception("Sorry, the type is not valid ("+DeepSearchEngine.allTypeList+").");
			
		}

		if (relationTabLink.length()==0) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" is not valid [required].");

		}

		//Try to get the record
		JSONObject rec = Record.getNode(relationTabLink);

		if (rec==null) {

			throw new Exception("Sorry, the relation tab link "+relationTabLink+" does not exist.");

		}
		
		JSONArray result = new JSONArray();
		int maxPerception = 0;
		
		//Get the first
		String tl = ExperienceManagerL2.getFirstTabLink(user, relationTabLink, type);
		
		//Add only if null
		if (tl!=null) {
			
			JSONObject bd = Record.getNode(Misc.atom(tl, 1, " ").substring(1));
			int perception = Integer.parseInt(""+bd.get("w"));
			
			maxPerception += perception;
			JSONObject obj = new JSONObject();
			obj.put("k", tl);
			obj.put("w", perception);
			result.add(obj);
			
			//Get the down node
			tl = ExperienceManagerL2.getDownTabLink(user, relationTabLink, tl, type);
			
			int i = 0;
			
			while(tl!=null && i+1<ConcentrationManager.getConcentrationDepth("C[relation]")) {
				
				bd = Record.getNode(Misc.atom(tl, 1, " ").substring(1));
				
				perception = Integer.parseInt(""+bd.get("w"));
				
				maxPerception += perception;
				obj = new JSONObject();
				obj.put("k", tl);
				obj.put("w", perception);
				result.add(obj);
				
				i++;

				//Get the next down node
				tl = ExperienceManagerL2.getDownTabLink(user, relationTabLink, tl, type);
				
			}
			
			for(i=0;i<result.size();i++) {
				
				//Get the current object
				obj = (JSONObject) result.get(i);
				
				obj.put("mw", maxPerception);

				if (maxPerception==0) obj.put("v", 0D);
				else obj.put("v", Double.parseDouble(""+obj.get("w"))/Double.parseDouble(""+maxPerception)*100);
				
			}

		}
		
		return result;

	}

}
