/**
 * Project: MentDB
 * Description: Mentalese Database Engine
 * Website: http://www.mentdb.org
 * Twitter: https://twitter.com/mentalese_db
 * LinkedIn: https://www.linkedin.com/company/mentdb
 * Facebook: https://www.facebook.com/mentdb
 * Author: Jimmitry Payet
 * Website: http://www.jpayet.re
 * Mail: contact@jpayet.re
 * Locality: Reunion Island (French)
 */

package re.jpayet.mentdb.doc;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.LinkedHashMap;
import java.util.Map.Entry;
import java.util.Vector;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

import re.jpayet.mentdb.core.command.Command;
import re.jpayet.mentdb.core.command.MQLValue;
import re.jpayet.mentdb.core.db.file.transaction.Transaction;
import re.jpayet.mentdb.core.fx.StringFx;
import re.jpayet.mentdb.core.search.DeepSearchEngine;
import re.jpayet.mentdb.core.session.SessionThread;
import re.jpayet.mentdb.server.Start;
import re.jpayet.mentdb.tools.Misc;

//The MQL documentation object
public class MQLDocumentation {

	public static LinkedHashMap<String, Vector<MQLDocumentation>> functions = new LinkedHashMap<String, Vector<MQLDocumentation>>();
	public static LinkedHashMap<String, String> ghost_functions = new LinkedHashMap<String, String>();
	public static LinkedHashMap<String, String> page_description = new LinkedHashMap<String, String>();
	public static LinkedHashMap<String, String> video = new LinkedHashMap<String, String>();
	
	//Function to help to develop JSP pages
	public static void test() {
		
		
		
	}
	
	//Properties
	public boolean valid = false;
	public String FX = "";
	public String functionId = "";
	public String description = "";
	public String example1 = null;
	public String result1 = null;
	public String example2 = null;
	public String result2 = null;
	public String example3 = null;
	public String result3 = null;
	public String synonymous = "";
	public boolean webSocket = false;
	public String warning = "";
	public Vector<MQLParam> parameters = new Vector<MQLParam>();
	
	//Add parameter
	public void addParam(MQLParam param) {
		
		parameters.addElement(param);
		
	}
	
	//Constructor
	public MQLDocumentation(String FX, boolean valid, String functionId, String description, 
			String example1, String result1, 
			String example2, String result2, 
			String example3, String result3, 
			boolean webSocket, String synonymous) {

		this.valid = valid;
		this.FX = FX;
		this.functionId = functionId;
		this.description = description;
		this.example1 = example1;
		this.result1 = result1;
		this.example2 = example2;
		this.result2 = result2;
		this.example3 = example3;
		this.result3 = result3;
		this.webSocket = webSocket;
		this.synonymous = synonymous;
		
	}
	
	//Constructor
	public MQLDocumentation(boolean valid, String functionId, String description, 
			String example1, String result1, 
			String example2, String result2, 
			String example3, String result3, 
			boolean webSocket, String synonymous) {

		this.valid = valid;
		this.functionId = functionId;
		this.description = description;
		this.example1 = example1;
		this.result1 = result1;
		this.example2 = example2;
		this.result2 = result2;
		this.example3 = example3;
		this.result3 = result3;
		this.webSocket = webSocket;
		this.synonymous = synonymous;
		
	}
	
	//To HTML
	public static String menu_to_html() {
		
		//Initialization
		String result = "";
	
		int z = 0;
		for (Entry<String, Vector<MQLDocumentation>> e : MQLDocumentation.functions.entrySet()) {
			
			if (!ghost_functions.containsKey(e.getKey())) {

				if (z>0) result += "<div style='clear:left;height:20px'></div>";
	
				result += "<div style='line-height: 32px;clear:left;font-size: 13px;font-weight:bold;padding:0px 8px 4px 0px;'><a href='mql_"+e.getKey().toLowerCase().replace(" ", "_")+".html' id='mql_"+e.getKey()+"' style='color: #000;padding: 5px 8px;border-bottom: 1px solid #D9DEE4;margin-left: -8px;width: 100%;display: block;font-size:16px'>"+e.getKey()+"</a></div>";
				result += "<div style='display: inline-block;padding-top: 5px;padding-bottom: 10px;'>";
				for(int i=0;i<e.getValue().size();i++) {
				
					//Get the current MQLDocumentation object
					
					MQLDocumentation currentFunction = e.getValue().get(i);
				
					String color = "#000";
					if (!currentFunction.warning.equals("")) color = "rgb(0, 102, 255)";
					
					String valid = "";
					if (currentFunction.valid) valid = ";color:"+color;
					else valid = ";color:#d70000";
					
					if (!currentFunction.FX.equals("")) {
						if (i==0) result += "<h3 style='clear: left;padding-top: 0px;padding-bottom: 0px;margin-bottom: 10px;margin-top: 0px;'>"+currentFunction.FX+"</h3>";
						else result += "<h3 style='clear: left;padding-top: 15px;padding-bottom: 0px;margin-bottom: 10px;'>"+currentFunction.FX+"</h3>";
					}
					
					if (currentFunction.webSocket) result += "<a title='"+currentFunction.warning.replace("'", "&#39;")+"' style='"+valid+";font-size: 12px;padding:2px 8px 2px 0px;float:left;display:inline-block;width:250px' href='mql_"+e.getKey().toLowerCase().replace(" ", "_")+".html#fx_"+currentFunction.functionId.replace(" ", "_")+"'>@"+currentFunction.functionId+(!currentFunction.synonymous.equals("")?" <span style='font-size:8px;'>"+currentFunction.synonymous+"</span>":"")+"</a>";
					else result += "<a title='"+currentFunction.warning.replace("'", "&#39;")+"' style='"+valid+";font-size: 12px;padding:2px 8px 2px 0px;float:left;display:inline-block;width:250px' href='mql_"+e.getKey().toLowerCase().replace(" ", "_")+".html#fx_"+currentFunction.functionId.replace(" ", "_")+"'>"+currentFunction.functionId+(!currentFunction.synonymous.equals("")?" <span style='font-size:8px;'>"+currentFunction.synonymous+"</span>":"")+"</a>";
				
				}
				
				result += "</div>";
				z++;
				
			}
			
		}
		result += "<div style='clear:left;'></div>";

		result += "<div style='border-bottom: 1px solid #D9DEE4;'></div>";
		
		return result;
		
	}
	
	//To HTML
	public static String mqlPage_to_html() throws Exception {
		
		//Initialization
		String result = "";
		
		SessionThread mentdbInternalSession = new SessionThread(null);
		mentdbInternalSession.user = "mentdb";
	
		//Description
		for (Entry<String, Vector<MQLDocumentation>> e : MQLDocumentation.functions.entrySet()) {
			
			result = "";

			result += "<div style='clear:left;height:20px'></div>";

			result += "<div style='line-height: 32px;clear:left;font-size: 13px;font-weight:bold;padding:0px 8px 4px 0px;font-size:24px'><a href='javascript:window.history.back()' style='color:#000'>Back</a> "+(!ghost_functions.containsKey(e.getKey())?"/ <a href='mql.html#mql_"+e.getKey()+"' style='color:#000'>All functions</a> ":"")+"/ "+e.getKey()+"</div>";
			if (page_description.containsKey(e.getKey())) result += "<div style='border-bottom: 1px solid #D9DEE4;'></div><h1 style='margin-top: 20px;margin-bottom: 15px;font-size: 14px;color: #3a3a3a;font-weight: normal;'>"+page_description.get(e.getKey())+"</h1>";
			if (!ghost_functions.containsKey(e.getKey())) {
				result += "<div style='display: inline-block;padding-top: 10px;padding-bottom: 10px;padding-left: 15px;background-color: #F0F0F0;border: 1px #D9D9D9 solid;'>";
				for(int i=0;i<e.getValue().size();i++) {
				
					//Get the current MQLDocumentation object
					
					MQLDocumentation currentFunction = e.getValue().get(i);
				
					String color = "#333";
					if (!currentFunction.warning.equals("")) color = "rgb(0, 102, 255)";
					
					String valid = "";
					if (currentFunction.valid) valid = ";color:"+color;
					else valid = ";color:#d70000";
					
					if (!currentFunction.FX.equals("")) {
						if (i==0) result += "<h3 style='clear: left;padding-top: 0px;padding-bottom: 0px;margin-bottom: 10px;margin-top: 0px;'>"+currentFunction.FX+"</h3>";
						else result += "<h3 style='clear: left;padding-top: 15px;padding-bottom: 0px;margin-bottom: 10px;'>"+currentFunction.FX+"</h3>";
					}
					
					if (currentFunction.webSocket) result += "<a title='"+currentFunction.warning.replace("'", "&#39;")+"' style='"+valid+";font-size: 12px;padding:2px 8px 2px 0px;float:left;display:inline-block;width:250px;' href='mql_"+e.getKey().toLowerCase().replace(" ", "_")+".html#fx_"+currentFunction.functionId.replace(" ", "_")+"'>@"+currentFunction.functionId+(!currentFunction.synonymous.equals("")?" <span style='font-size:8px;'>"+currentFunction.synonymous+"</span>":"")+"</a>";
					else result += "<a title='"+currentFunction.warning.replace("'", "&#39;")+"' style='"+valid+";font-size: 12px;padding:2px 8px 2px 0px;float:left;display:inline-block;width:250px;' href='mql_"+e.getKey().toLowerCase().replace(" ", "_")+".html#fx_"+currentFunction.functionId.replace(" ", "_")+"'>"+currentFunction.functionId+(!currentFunction.synonymous.equals("")?" <span style='font-size:8px;'>"+currentFunction.synonymous+"</span>":"")+"</a>";
				
				}
				
				result += "</div>";
			}
			result += "<div style='clear:left;'></div>";
			
			if (video.containsKey(e.getKey())) {
				
				result += "<iframe style='border:0px;margin-top:10px' width='800' height='400'";
				result += "		src='"+video.get(e.getKey())+"' allowfullscreen='allowfullscreen' mozallowfullscreen='mozallowfullscreen' msallowfullscreen='msallowfullscreen' oallowfullscreen='oallowfullscreen' webkitallowfullscreen='webkitallowfullscreen'>";
				result += "</iframe>";
				
			}
						
			for(int i=0;i<e.getValue().size();i++) {
			
				//Get the current MQLDocumentation object
				
				MQLDocumentation currentFunction = e.getValue().get(i);
				
				String currentDocId = "Doc > "+currentFunction.functionId;
				if (currentDocId.length()<60) Misc.system_out_print("\r"+StringFx.rpad(currentDocId, " ", "61"), false, "", "");
				else Misc.system_out_print("\r"+StringFx.rpad(currentDocId.substring(0, 60), " ", "61"), false, "", "");
				
				result += "<div style='-moz-box-sizing: border-box;box-sizing: border-box;border-bottom:1px #F0F0F0 solid;padding-bottom:20px;font-size:12px'>";
				result += "<div id='fx_"+currentFunction.functionId.replace(" ", "_")+"'><h2 style='color:#000'>"+(currentFunction.webSocket?"@":"")+currentFunction.functionId;
				if (currentFunction.parameters.size()>0) for(int r=0;r<currentFunction.parameters.size();r++) {
					result += " &lt;<span style='color:#808080;font-weight:normal'>"+currentFunction.parameters.get(r).name+"</span>&gt;";
				}
				result += "</h2></div>";
				result += "<div style='margin-left:0px;'>";
				if (!currentFunction.synonymous.equals("")) result += "<div style='color:#000'><b>Synonymous</b><br><br> &nbsp;  &nbsp; "+currentFunction.synonymous+"</div><br>";
				if (!currentFunction.description.equals("")) result += "<div style='color:#000'><b>Description</b><br><br> &nbsp;  &nbsp; "+currentFunction.description+"</div><br>";
				if (currentFunction.parameters.size()>0) {
					result += "<div style='color:#000'><b>Parameters</b><br>";
					
					for(int r=0;r<currentFunction.parameters.size();r++) {
						result += "<br> &nbsp;  &nbsp; <span style='color:#808080'>"+currentFunction.parameters.get(r).name+"</span>: &nbsp; "+currentFunction.parameters.get(r).description+" - <span style='color:#333'>"+currentFunction.parameters.get(r).type+"</span> - "+(currentFunction.parameters.get(r).required?"<span style='color:#333'>required</span>":"<span style='color:#333'>not required</span>");
					}
					result += "</div>";
				}
				if (currentFunction.example1!=null) {
					result += "<div style='overflow:auto;line-height: 20px;margin-top: 15px;padding:10px 15px 0px 15px;background-color: #F8F8F8;border:1px #F0F0F0 solid'>";
					result += "<pre style='color:#000;margin-top: 0px;'><span style='font-size: 14px;'><b>admin</b></span><br>"+currentFunction.example1+"</pre>";
					result += "<pre style='color:#4C9ED9'><span style='font-size: 14px;'><b>mentdb</b></span><br>"+currentFunction.result1+"</pre>";
					if (currentFunction.valid && Start.CHECK_DOCUMENTATION==1) {
						String mqlResult = "";
						try {
							
							Vector<Vector<MQLValue>> inputVectorCmds = Misc.splitCommand(currentFunction.example1.replace("<br>", "\n"));
							for(int ii=0;ii<inputVectorCmds.size();ii++) {
								
								Vector<MQLValue> inputVectorCmd = inputVectorCmds.get(ii);
								mqlResult = Command.execute(mentdbInternalSession, inputVectorCmd, mentdbInternalSession.env);
	
							}
							
							Transaction.commit("mentdb");

							result += "<pre style='color:#090'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult+"</pre>";
							result += "<pre style='color:#009'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult.replace("\n", "&lt;br&gt;").replace("\\", "\\\\").replace("\"", "\\\"")+"</pre>";
							
						} catch (Exception z) {
							
							Transaction.rollback("mentdb");
							mqlResult = z.getMessage()+"";
							result += "<pre style='color:#F00'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult+"</pre>";
							
						}
						
					}
					result += "</div>";
				}
				if (currentFunction.example2!=null) {
					result += "<div style='overflow:auto;line-height: 20px;margin-top: 15px;padding:10px 15px 0px 15px;background-color: #F8F8F8;border:1px #F0F0F0 solid'>";
					result += "<pre style='color:#000;margin-top: 0px;'><span style='font-size: 14px;'><b>admin</b></span><br>"+currentFunction.example2+"</pre>";
					result += "<pre style='color:#4C9ED9'><span style='font-size: 14px;'><b>mentdb</b></span><br>"+currentFunction.result2+"</pre>";
					if (currentFunction.valid && Start.CHECK_DOCUMENTATION==1) {
						String mqlResult = "";
						try {
							
							Vector<Vector<MQLValue>> inputVectorCmds = Misc.splitCommand(currentFunction.example2.replace("<br>", "\n"));
							for(int ii=0;ii<inputVectorCmds.size();ii++) {
								
								Vector<MQLValue> inputVectorCmd = inputVectorCmds.get(ii);
								mqlResult = Command.execute(mentdbInternalSession, inputVectorCmd, mentdbInternalSession.env);
	
							}

							Transaction.commit("mentdb");
							
							result += "<pre style='color:#090'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult+"</pre>";
							result += "<pre style='color:#009'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult.replace("\n", "&lt;br&gt;").replace("\\", "\\\\").replace("\"", "\\\"")+"</pre>";
							
						} catch (Exception z) {
							
							Transaction.rollback("mentdb");
							mqlResult = z.getMessage()+"";
							result += "<pre style='color:#F00'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult+"</pre>";
							
						}
						
					}
					result += "</div>";
				}
				if (currentFunction.example3!=null) {
					result += "<div style='overflow:auto;line-height: 20px;margin-top: 15px;padding:10px 15px 0px 15px;background-color: #F8F8F8;border:1px #F0F0F0 solid'>";
					result += "<pre style='color:#000;margin-top: 0px;'><span style='font-size: 14px;'><b>admin</b></span><br>"+currentFunction.example3+"</pre>";
					result += "<pre style='color:#4C9ED9'><span style='font-size: 14px;'><b>mentdb</b></span><br>"+currentFunction.result3+"</pre>";
					if (currentFunction.valid && Start.CHECK_DOCUMENTATION==1) {
						String mqlResult = "";
						try {
							
							Vector<Vector<MQLValue>> inputVectorCmds = Misc.splitCommand(currentFunction.example3.replace("<br>", "\n"));
							for(int ii=0;ii<inputVectorCmds.size();ii++) {
								
								Vector<MQLValue> inputVectorCmd = inputVectorCmds.get(ii);
								mqlResult = Command.execute(mentdbInternalSession, inputVectorCmd, mentdbInternalSession.env);
	
							}
							
							Transaction.commit("mentdb");
							
							result += "<pre style='color:#090'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult+"</pre>";
							result += "<pre style='color:#009'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult.replace("\n", "&lt;br&gt;").replace("\\", "\\\\").replace("\"", "\\\"")+"</pre>";
							
						} catch (Exception z) {
							
							Transaction.rollback("mentdb");
							mqlResult = z.getMessage()+"";
							result += "<pre style='color:#F00'><span style='font-size: 14px;'><b>checker</b></span><br>"+mqlResult+"</pre>";
							
						}
						
					}
					result += "</div>";
				}
				result += "</div>";
				result += "</div>";
				
			}
			
			Misc.create("docs/mql_"+e.getKey().toLowerCase().replace(" ", "_")+".html", Misc.load("docs/mql_fx_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ""+e.getKey()).replace("[MQL_FX]", result).replace("[MQL_VERSION]", Start.version));
			
		}
		
		return result;
		
	}
	
	//To HTML
	public static String mqlFuntion_to_html(String idClass, String idFunction) throws Exception {
		
		//Initialization
		String result = "";
		
		Vector<MQLDocumentation> e = MQLDocumentation.functions.get(idClass);
						
		for(int i=0;i<e.size();i++) {
		
			//Get the current MQLDocumentation object
			
			MQLDocumentation currentFunction = e.get(i);
			
			if (currentFunction.functionId.equals(idFunction)) {
			
				result += "<div style='-moz-box-sizing: border-box;box-sizing: border-box;padding-bottom:20px;font-size:12px'>";
				result += "<div id='fx_"+currentFunction.functionId.replace(" ", "_")+"'><h2 style='color:#000'><a href='#top'><img src='images/top.png' style='width:16px'></a> "+(currentFunction.webSocket?"@":"")+currentFunction.functionId;
				if (currentFunction.parameters.size()>0) for(int r=0;r<currentFunction.parameters.size();r++) {
					result += " &lt;<span style='color:#d70000;font-weight:normal'>"+currentFunction.parameters.get(r).name+"</span>&gt;";
				}
				result += "</h2></div>";
				result += "<div style='margin-left:20px;'>";
				if (!currentFunction.synonymous.equals("")) result += "<div style='color:#000'><b>Synonymous</b><br> &nbsp;  &nbsp; "+currentFunction.synonymous+"</div>";
				if (!currentFunction.description.equals("")) result += "<div style='color:#000'><b>Description</b><br> &nbsp;  &nbsp; "+currentFunction.description+"</div>";
				if (currentFunction.parameters.size()>0) {
					result += "<div style='color:#000'><b>Parameters</b>";
					
					for(int r=0;r<currentFunction.parameters.size();r++) {
						result += "<br> &nbsp;  &nbsp; <span style='color:#d70000'>"+currentFunction.parameters.get(r).name+"</span>: &nbsp; "+currentFunction.parameters.get(r).description+" - <span style='color:#4C9ED9'>"+currentFunction.parameters.get(r).type+"</span> - "+(currentFunction.parameters.get(r).required?"<span style='color:#d70000'>required</span>":"<span style='color:#0d0'>not required</span>");
					}
					result += "</div>";
				}
				if (currentFunction.example1!=null) {
					result += "<div style='overflow:auto;line-height: 20px;margin-top: 15px;padding:10px 15px 0px 15px;background-color: #F8F8F8;border:1px #F0F0F0 solid'>";
					result += "<div style='color:#000'><span style='font-size: 14px;'><b>admin</b></span><br>"+currentFunction.example1+"</div>";
					result += "<pre style='color:#4C9ED9'><span style='font-size: 14px;'><b>mentdb</b></span><br>"+currentFunction.result1+"</pre>";
					result += "</div>";
				}
				if (currentFunction.example2!=null) {
					result += "<div style='overflow:auto;line-height: 20px;margin-top: 15px;padding:10px 15px 0px 15px;background-color: #F8F8F8;border:1px #F0F0F0 solid'>";
					result += "<div style='color:#000'><span style='font-size: 14px;'><b>admin</b></span><br>"+currentFunction.example2+"</div>";
					result += "<pre style='color:#4C9ED9'><span style='font-size: 14px;'><b>mentdb</b></span><br>"+currentFunction.result2+"</pre>";
					result += "</div>";
				}
				if (currentFunction.example3!=null) {
					result += "<div style='overflow:auto;line-height: 20px;margin-top: 15px;padding:10px 15px 0px 15px;background-color: #F8F8F8;border:1px #F0F0F0 solid'>";
					result += "<div style='color:#000'><span style='font-size: 14px;'><b>admin</b></span><br>"+currentFunction.example3+"</div>";
					result += "<pre style='color:#4C9ED9'><span style='font-size: 14px;'><b>mentdb</b></span><br>"+currentFunction.result3+"</pre>";
					result += "</div>";
				}
				result += "</div>";
				result += "</div>";
				
			}
			
		}
		
		return result;
		
	}
	
	//Initialization
	@SuppressWarnings("unchecked")
	public static void init() throws Exception {
		
		MQLDocumentation mql = null;
		
		functions.put("Language", new Vector<MQLDocumentation>());
		page_description.put("Language", "<img src='images/p.png' style='vertical-align: middle;'>Here you can find the functions needed to create new languages in the MentDB brain."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A language can be stimulated, and recalled more quickly if needed."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A created language can not be deleted.");
		mql = new MQLDocumentation(true, "language create", "To insert a new language", "language create \"en\";", "Language en created with successful.", null, null, null, null, false, "language create|insert|add");
		mql.addParam(new MQLParam("lang", "The language id", "string", true));
		functions.get("Language").add(mql);
		functions.get("Language").add(new MQLDocumentation(true, "language show", "To show all languages", "language show", "[<br>  \"en\",<br>  \"fr\",<br>  \"io\"<br>]", "@[r_fr_?_est_une_langue]", "RG[b]", "graph relation (@[r_fr_?_est_une_langue]);", "j23i88m90m76i39t04r09y35p14a96y09e57t39{\"e\":[{\"n1\":\"V0_RG[b]\",\"n2\":\"T1_TH[]\",\"k\":\"$V0_RG[b]_T1_TH[]\"},{\"n1\":\"T1_TH[]\",\"n2\":\"T2_TH[2]\",\"k\":\"$T1_TH[]_T2_TH[2]\"},{\"n1\":\"T2_TH[2]\",\"n2\":\"T3_TH[8]\",\"k\":\"$T2_TH[2]_T3_TH[8]\"},{\"n1\":\"T3_TH[8]\",\"n2\":\"T4_TH[e]\",\"k\":\"$T3_TH[8]_T4_TH[e]\"},{\"n1\":\"V0_RG[b]\",\"n2\":\"M5_[RG[b] 1\\/3]\",\"k\":\"$V0_RG[b]_5\"},{\"n1\":\"M5_[RG[b] 1\\/3]\",\"n2\":\"M6_[RG[b] 2\\/3]\",\"k\":\"$M5_[RG[b] 1\\/3]_6\"},{\"n1\":\"M6_[RG[b] 2\\/3]\",\"n2\":\"B7_L\",\"k\":\"M6_[RG[b] 2\\/3]_B7_L\"},{\"n1\":\"B7_L\",\"n2\":\"R8_RL[4x]\",\"k\":\"B7_L_R8_RL[4x]\"},{\"n1\":\"R8_RL[4x]\",\"n2\":\"R9_RL[4w]\",\"k\":\"R8_RL[4x]_R9_RL[4w]\"},{\"n1\":\"R9_RL[4w]\",\"n2\":\"R10_RL[4v]\",\"k\":\"R9_RL[4w]_R10_RL[4v]\"},{\"n1\":\"M6_[RG[b] 2\\/3]\",\"n2\":\"M11_[RG[b] 3\\/3]\",\"k\":\"$M6_[RG[b] 2\\/3]_11\"},{\"n1\":\"M11_[RG[b] 3\\/3]\",\"n2\":\"B12_L\",\"k\":\"M11_[RG[b] 3\\/3]_B12_L\"},{\"n1\":\"B12_L\",\"n2\":\"R13_RL[4x]\",\"k\":\"B12_L_R13_RL[4x]\"},{\"n1\":\"R13_RL[4x]\",\"n2\":\"R14_RL[4w]\",\"k\":\"R13_RL[4x]_R14_RL[4w]\"},{\"n1\":\"R14_RL[4w]\",\"n2\":\"R15_RL[4v]\",\"k\":\"R14_RL[4w]_R15_RL[4v]\"}],\"n\":{\"M5_[RG[b] 1\\/3]\":\"1\",\"T4_TH[e]\":\"TH[e]\",\"R10_RL[4v]\":\"RL[4v]\",\"M6_[RG[b] 2\\/3]\":\"2\",\"T2_TH[2]\":\"TH[2]\",\"B12_L\":\"L\",\"V0_RG[b]\":\"RG[b]\",\"R14_RL[4w]\":\"RL[4w]\",\"T1_TH[]\":\"TH[]\",\"T3_TH[8]\":\"TH[8]\",\"R8_RL[4x]\":\"RL[4x]\",\"R13_RL[4x]\":\"RL[4x]\",\"R9_RL[4w]\":\"RL[4w]\",\"M11_[RG[b] 3\\/3]\":\"3\",\"R15_RL[4v]\":\"RL[4v]\",\"B7_L\":\"L\"}}", false, ""));
		mql = new MQLDocumentation(true, "language stimulate", "To stimulate a language", "language stimulate fr", "Language stimulated with successful.", "language show", "[<br>  \"fr\",<br>  \"en\",<br>  \"io\"<br>]", "relation show perceptions (concat \"[\" @[r_fr_?_est_une_langue] \" 2/3]\") L;", "[<br>  {<br>    \"v\": 100.0,<br>    \"w\": 1,<br>    \"mw\": 1,<br>    \"k\": \"[RL[4w] 2/3]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 1,<br>    \"k\": \"[RL[4x] 2/3]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 1,<br>    \"k\": \"[RL[4v] 2/3]\"<br>  }<br>]", false, "");
		mql.addParam(new MQLParam("language", "The language", "string", true));
		functions.get("Language").add(mql);
		mql = new MQLDocumentation(true, "language first", "To show the first language", "language first", "fr", "relation show (concat \"[\" @[r_fr_?_est_une_langue] \" 2/3]\") L;", "[<br>  \"[RL[4w] 2/3]\",<br>  \"[RL[4x] 2/3]\",<br>  \"[RL[4v] 2/3]\"<br>]", "relation show sentence RL[4x];", "en est une langue", false, "");
		functions.get("Language").add(mql);
		
		functions.put("Symbol", new Vector<MQLDocumentation>());
		page_description.put("Symbol", "<img src='images/p.png' style='vertical-align: middle;'>Words are based on symbols linked together congnitively."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>When you stimulate a word, you also stimulate all the symbols making up the word."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can not delete a symbol.");
		mql = new MQLDocumentation(true, "symbol create", "To insert a new symbol", "#Add the symbols A-Z;<br>for (-> \"[i]\" 65) (<= [i] 90) (++ \"[i]\") {<br>	symbol create (string char [i];) \"en\"; <br>};<br>#Add the symbols 0-9;<br>for (-> \"[i]\" 48) (<= [i] 57) (++ \"[i]\") {<br>	symbol create (string char [i];) \"en\"; <br>};<br>#Add the symbols a-z;<br>for (-> \"[i]\" 97) (<= [i] 122) (++ \"[i]\") {<br>	symbol create (string char [i];) \"en\"; <br>};", "Symbol S[z] created with successful in the language 'en'.", "symbol create \"-\" en", "Symbol S[-] created with successful in the language 'en'.", null, null, false, "symbol create|insert|add");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol delete", "To delete a symbol", "symbol delete x en", "Symbol S[x] deleted with successful in the language 'en'.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol first", "To get the first symbol tab link", "symbol first s fr", "TL[657370616365 1 fr]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol get", "To get the symbol tab link", "symbol get s 0 fr", "TL[657370616365 1 fr]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("position", "The position", "integer>=0", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol last", "To get the last symbol tab link", "symbol last s fr", "TL[63686f7365 3 fr]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol show", "To show symbol tab links", "symbol show s fr", "[<br>  \"TL[657370616365 1 fr]\",<br>  \"TL[73e9706172e9 0 fr]\",<br>  \"TL[706872617365 4 fr]\",<br>  \"TL[646976697365 4 fr]\",<br>  \"TL[7375707072696d65 0 fr]\",<br>  \"TL[6c69737465 2 fr]\",<br>  \"TL[706173 2 fr]\",<br>  \"TL[73616c7574 0 fr]\",<br>  \"TL[73696e6f6e 0 fr]\",<br>  \"TL[616c6f7273 4 fr]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol perception", "To show symbol tab link perception", "symbol perception s fr", "[<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[657370616365 1 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[73e9706172e9 0 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[706872617365 4 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[646976697365 4 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[7375707072696d65 0 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[6c69737465 2 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[706173 2 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[73616c7574 0 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[73696e6f6e 0 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"nw\": 19,<br>    \"k\": \"TL[616c6f7273 4 fr]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol show words", "To show symbol words", "symbol show words s fr", "[<br>  \"espace\",<br>  \"séparé\",<br>  \"phrase\",<br>  \"divise\",<br>  \"supprime\",<br>  \"liste\",<br>  \"pas\",<br>  \"salut\",<br>  \"sinon\",<br>  \"alors\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol show languages", "To show symbol languages", "symbol show languages s", "[<br>  \"io\",<br>  \"en\",<br>  \"fr\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbol", "The symbol", "string", true));
		functions.get("Symbol").add(mql);
		mql = new MQLDocumentation(true, "symbol stimulate", "To stimulate a symbol tab link", "symbol stimulate \"TL[63686f7365 3 fr]\"", "Symbol tab link stimulated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("tabLinkId", "The symbol tab link id", "string", true));
		functions.get("Symbol").add(mql);
		
		functions.put("Word", new Vector<MQLDocumentation>());
		page_description.put("Word", "<img src='images/p.png' style='vertical-align: middle;'>Words and symbols form a gigantic index that points to thoughts."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can easily search for a word with just a few symbols."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A word can exist in several languages, or sometimes in the same language.");
		mql = new MQLDocumentation(true, "word create", "To insert a new word in a language", "word create dog en", "TH[6q]", "word create chien fr", "TH[6r]", "word create cat en", "TH[6s]", false, "word create|insert|add");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word create", "To insert a new word in a language with separator", "word create \"New-York\" en \"-\"", "TH[6v]", null, null, null, null, false, "word create|insert|add");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("separator", "The separator", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word import from text", "To insert words in a language from a text", "word import from text \"132 candy street, Bigtown BG23 4YZ, England.<br>133 big candy street, Kingston, New-York 12401, United-States.\" en \"-\" \",.;:?!\"", "Words imported with successful.", "word import from text \"134 rue des bonbons, La Réunion, France.\" fr \"-\" \",.;:?!\"", "Words imported with successful.", null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("separator", "The separator", "string", true));
		mql.addParam(new MQLParam("excludedChars", "Excluded chars", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word exist", "To check if a word already exist in a language", "word exist dog en", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word first", "To get the first word tab link", "word first dog en", "TH[6q]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word last", "To get the last word tab link", "word last dog en", "TH[6q]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word get", "To get the word tab link", "word get dog 0 en", "TH[6q]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("position", "The position", "integer>=0", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word lang probability", "To get the language probability", "word lang probability \"Jimmitry est un human\"", "[<br>  {<br>    \"v\": 100.0,<br>    \"k\": \"fr\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("sentence", "The sentence", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word search", "To search words", "word search ad \"ad.*\" top fr", "[{\"admin\":100.0}]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("symbols", "Symbols", "string", true));
		mql.addParam(new MQLParam("regex", "The regex", "string", true));
		mql.addParam(new MQLParam("order", "The order", "string (asc|desc|top)", true));
		mql.addParam(new MQLParam("lang", "The lang", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word levenshtein", "To search levenshtein distance between words", "word levenshtein admn top fr", "[{\"France\":83.33333333333334},{\"La\":200.0},{\"admin\":20.0},{\"ai\":150.0},{\"orange\":83.33333333333334},{\"jaune\":80.0},{\"blanc\":80.0},{\"espace\":100.0},{\"par\":133.33333333333331},{\"séparé\":100.0},{\"phrase\":100.0},{\"ajout\":80.0},{\"pas\":133.33333333333331},{\"salut\":80.0},{\"alors\":80.0},{\"ça\":200.0},{\"faire\":80.0},{\"a\":300.0},{\"faux\":100.0},{\"vrai\":100.0},{\"langue\":83.33333333333334},{\"utilisateur\":90.9090909090909},{\"article\":85.71428571428571},{\"la\":200.0},{\"base\":100.0},{\"mentdb\":100.0},{\"divise\":100.0},{\"du\":150.0},{\"comprend\":75.0},{\"dire\":100.0},{\"répondre\":87.5},{\"dois\":100.0},{\"dit\":100.0},{\"seconde\":85.71428571428571},{\"donnée\":83.33333333333334},{\"des\":100.0},{\"de\":150.0},{\"multiplie\":100.0},{\"supprime\":87.5},{\"première\":87.5},{\"mot\":133.33333333333331},{\"merci\":100.0},{\"me\":150.0},{\"minute\":100.0},{\"nombre\":83.33333333333334},{\"extrême\":85.71428571428571},{\"émotion\":85.71428571428571},{\"Réunion\":85.71428571428571},{\"bonbons\":85.71428571428571},{\"chien\":80.0},{\"en\":150.0},{\"noire\":100.0},{\"technique\":88.88888888888889},{\"ne\":200.0},{\"bonjour\":100.0},{\"rien\":75.0},{\"sinon\":80.0},{\"quelqu'un\":88.88888888888889},{\"booléen\":85.71428571428571},{\"une\":133.33333333333331},{\"un\":150.0}]", "word levenshtein admn asc fr", "[{\"admin\":20.0},{\"comprend\":75.0},{\"rien\":75.0},{\"jaune\":80.0},{\"blanc\":80.0},{\"ajout\":80.0},{\"salut\":80.0},{\"alors\":80.0},{\"faire\":80.0},{\"chien\":80.0},{\"sinon\":80.0},{\"France\":83.33333333333334},{\"orange\":83.33333333333334},{\"langue\":83.33333333333334},{\"donnée\":83.33333333333334},{\"nombre\":83.33333333333334},{\"article\":85.71428571428571},{\"seconde\":85.71428571428571},{\"extrême\":85.71428571428571},{\"émotion\":85.71428571428571},{\"Réunion\":85.71428571428571},{\"bonbons\":85.71428571428571},{\"booléen\":85.71428571428571},{\"répondre\":87.5},{\"supprime\":87.5},{\"première\":87.5},{\"technique\":88.88888888888889},{\"quelqu'un\":88.88888888888889},{\"utilisateur\":90.9090909090909},{\"espace\":100.0},{\"séparé\":100.0},{\"phrase\":100.0},{\"faux\":100.0},{\"vrai\":100.0},{\"base\":100.0},{\"mentdb\":100.0},{\"divise\":100.0},{\"dire\":100.0},{\"dois\":100.0},{\"dit\":100.0},{\"des\":100.0},{\"multiplie\":100.0},{\"merci\":100.0},{\"minute\":100.0},{\"noire\":100.0},{\"bonjour\":100.0},{\"par\":133.33333333333331},{\"pas\":133.33333333333331},{\"mot\":133.33333333333331},{\"une\":133.33333333333331},{\"ai\":150.0},{\"du\":150.0},{\"de\":150.0},{\"me\":150.0},{\"en\":150.0},{\"un\":150.0},{\"La\":200.0},{\"ça\":200.0},{\"la\":200.0},{\"ne\":200.0},{\"a\":300.0}]", null, null, false, "");
		mql.addParam(new MQLParam("symbols", "Symbols", "string", true));
		mql.addParam(new MQLParam("order", "The order", "string (asc|desc|top)", true));
		mql.addParam(new MQLParam("lang", "The lang", "string", false));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word show languages", "To show all languages for a specific word", "word show languages est", "[<br>  \"fr\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word show", "To show word tab links in all languages", "word show dog", "[<br>  {<br>    \"thoughtId\": \"TH[6q]\",<br>    \"lang\": \"en\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word show", "To show word tab links", "word show dog en", "[<br>  \"TH[6q]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word perception", "To show symbol tab link perceptions for a specific word (average % tab link)", "word perception est fr", "[<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 0,<br>    \"k\": \"TL[657374 0 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 0,<br>    \"k\": \"TL[657374 1 fr]\"<br>  },<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 0,<br>    \"k\": \"TL[657374 2 fr]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word perception symbol", "To show symbol perception for a specific word (average % symbol)", "word perception symbol est fr", "[<br>  {<br>    \"v\": 60.909090909090914,<br>    \"w\": 67,<br>    \"mw\": 110,<br>    \"k\": \"S[e]\"<br>  },<br>  {<br>    \"v\": 18.181818181818183,<br>    \"w\": 20,<br>    \"mw\": 110,<br>    \"k\": \"S[s]\"<br>  },<br>  {<br>    \"v\": 20.909090909090907,<br>    \"w\": 23,<br>    \"mw\": 110,<br>    \"k\": \"S[t]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word perception thought", "To show thought perception for a specific word", "word perception thought est fr", "[<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 0,<br>    \"k\": \"TH[2]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word stimulate", "To stimulate a word", "word stimulate est fr", "Word W[est] stimulated with successful in the language 'fr'.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word delete", "To delete a word in a language", "word delete chien fr", "Word W[chien] deleted with successful in the language 'fr'.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Word").add(mql);
		mql = new MQLDocumentation(true, "word delete", "To delete a word", "word delete dog", "Word W[dog] deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		functions.get("Word").add(mql);
		
		functions.put("Thought", new Vector<MQLDocumentation>());
		page_description.put("Thought", "<img src='images/p.png' style='vertical-align: middle;'>Depending on the language, a word may have identical or different meanings."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>The same word can have many different thoughts, eg: the hand, to hand ..."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A thought can be stimulated and all these relationships can go up.");
		mql = new MQLDocumentation(true, "thought create", "insert thought", "thought insert cat \"\" en", "TH[6t]", null, null, null, null, false, "thought create|insert|add");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("separator", "The separator", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought first", "To get the first thought", "thought first cat en", "TH[6t]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought get", "To get a thought", "thought get cat 0 en", "TH[6t]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("position", "The position", "integer>=0", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought last", "To get the last thought", "thought last cat en", "TH[6t]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought merge", "To merge thoughts", "thought merge TH[6t] TH[6s];", "Thoughts merged with successful.", "thought merge (<br>	word create snake en;<br>) (<br>	word create serpent fr;<br>);", "Thoughts merged with successful.", null, null, false, "");
		mql.addParam(new MQLParam("thoughtId1", "The thought id 1", "string", true));
		mql.addParam(new MQLParam("thoughtId2", "The thought id 2", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought show", "To show all thoughts linked with a word in a language", "thought show cat en", "[<br>  \"TH[6t]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought show", "To show all thoughts linked with a word in all languages", "thought show cat", "[<br>  {<br>    \"thoughtId\": \"TH[6t]\",<br>    \"lang\": \"en\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought show words", "To show all words in a thought", "thought show words TH[]", "[<br>  {<br>    \"lang\": \"fr\",<br>    \"word\": \"W[quelque chose]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought show words", "To show all words in a thought for a specific language", "thought show words TH[] fr", "[<br>  {<br>    \"lang\": \"fr\",<br>    \"word\": \"W[quelque chose]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought stimulate", "To stimulate a thought", "thought stimulate TH[6t]", "Thought TH[6t] stimulated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought delete", "To delete a specific thought", "thought delete (word create chicken en)", "Thought TH[115] deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought delete by lang", "To delete a specific thought in a language", "thought delete by lang en (word create chicken en)", "Thought TH[115] deleted with successful in the language 'en'.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought delete by word", "To delete a specific thought in a word", "thought delete by word chicken (word create chicken en)", "Thought TH[115] deleted with successful in the word 'chicken'.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Thought").add(mql);
		mql = new MQLDocumentation(true, "thought delete by word lang", "To delete a specific thought in a word and in a language", "thought delete by word lang chicken en (word create chicken en)", "Thought TH[115] deleted with successful in the word 'chicken' and the language en.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Thought").add(mql);
		
		functions.put("Relation", new Vector<MQLDocumentation>());
		page_description.put("Relation", "<img src='images/p.png' style='vertical-align: middle;'>Here you can create relationships between thoughts, or other relationships."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A relation can be stimulated and all these sub relationships can go up."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A relation can contain a condition and an action in MQL.");
		mql = new MQLDocumentation(true, "relation create", "To insert a new relation", "#Create and merge 'something/is' words;<br>thought merge TH[] (word create \"something\" en);<br>thought merge TH[2] (word create \"is\" en);<br>relation insert \"TH[] TH[2] TH[]\" G en;", "RG[v]", "relation show sentence RG[v]", "something is something", null, null, false, "relation create|insert|add");
		mql.addParam(new MQLParam("thoughtOrRelations", "Thoughts or relations", "string", true));
		mql.addParam(new MQLParam("type", "The type ("+DeepSearchEngine.allTypeList+")", "string", true));
		mql.addParam(new MQLParam("lang", "The language id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation create sub", "To insert a new sub relation with parent positions", "relation insert sub RG[v] <br>	(\"TH[] TH[2] \" (relation insert ((word create \"an\" en) \" \" (word create \"animal\" en)) U en))<br>	\"1 2 3\" G en", "RG[w]", null, null, null, null, false, "relation create|insert|add sub");
		mql.addParam(new MQLParam("parentRelationId", "The parent relation id", "string", true));
		mql.addParam(new MQLParam("thoughtOrRelations", "The thoughts", "string", true));
		mql.addParam(new MQLParam("positions", "The positions (ex: 1 2 3 3)", "string", true));
		mql.addParam(new MQLParam("type", "The type ("+DeepSearchEngine.allTypeList+")", "string", true));
		mql.addParam(new MQLParam("lang", "The language id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show sentence", "To show the sentence of a relation", "relation show sentence RG[w];", "something is an animal", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation translate", "Try to get a translation of a relation", "relation translate RG[w] en;", "something is an animal", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation translate", "Try to get a translation of a relation (by the first language)", "relation translate RG[w];", "something is an animal", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation new", "Generate a new 'relation create' with thoughts (thoughts are automatically created if does not exist).", "relation new \"a dog is an animal\" L en", "RL[4y]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("sentence", "The sentence", "string", true));
		mql.addParam(new MQLParam("type", "The type of the relation", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation has sub", "To search if a relation contains sub relation", "relation has sub RG[v] G", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		mql.addParam(new MQLParam("type", "The branch type  ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show info", "To show information about a relation", "relation show info RG[v];", "RG[v]: something is something <br>thoughts: TH[] TH[2] TH[] <br>nodes: TH[] TH[2] TH[] <br>", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation first", "To get the first relation tab link", "relation first \"[RG[v] 2/3]\" G", "[RG[w] 2/3]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationTabLinkId", "The relation tab link id", "string", true));
		mql.addParam(new MQLParam("type", "The branch type ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation get", "To get a relation tab link", "relation get \"[RG[v] 2/3]\" 0 G", "[RG[w] 2/3]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationTabLinkId", "The relation tab link id", "string", true));
		mql.addParam(new MQLParam("position", "The position", "integer>=0", true));
		mql.addParam(new MQLParam("type", "The branch type ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation last", "To get the last relation tab link", "relation last \"[RG[v] 2/3]\" G", "[RG[w] 2/3]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationTabLinkId", "The relation tab link id", "string", true));
		mql.addParam(new MQLParam("type", "The branch type ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation find", "To find a relation tab link", "relation find \"RG[v]\" 0 G", "[RG[w] 1/3]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		mql.addParam(new MQLParam("position", "The position", "integer>=0", true));
		mql.addParam(new MQLParam("type", "The branch type ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation network", "To get the network relation from a specific relation", "relation network RG[w]", "{<br>  \"th\": {<br>    \"RU[n]\": 1<br>  },<br>  \"tl\": {}<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation list ids", "To list all tab link ids in a relation", "relation list ids RG[w]", "[<br>  \"[RG[w] 1/3]\",<br>  \"[RG[w] 2/3]\",<br>  \"[RG[w] 3/3]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation search basic", "To make a basic search on thoughts", "relation search basic \"couleur noire\" \"*\"", "[<br>  \"RL[4n] 2\",<br>  \"RU[c] 1\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("text", "The text (separate by a space)", "string", true));
		mql.addParam(new MQLParam("typeList", "The type liste (*|A&S&L&G&I&O&U&D&R)", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation search deep", "To search relations after a basic search", "relation search deep (\n	relation search basic \"couleur noire\" \"*\"\n) \"*\" \"ASLGIOUDR\" 5 10;", "[<br>  \"RL[4n] 10\",<br>  \"RG[i] 10\",<br>  \"RG[1] 9\",<br>  \"RL[4m] 1\",<br>  \"RL[4l] 1\"<br>]", "relation search deep (\n	relation search basic \"couleur noire\" \"*\"\n) \"UL\" \"ASLGIOUDR\" 5 10", "[<br>  \"RL[4n] 10\",<br>  \"RL[4m] 1\",<br>  \"RL[4l] 1\",<br>  \"RL[4k] 1\",<br>  \"RL[4j] 1\"<br>]", null, null, false, "");
		mql.addParam(new MQLParam("jsonRelations", "The relations tab (from a basic search)", "string", true));
		mql.addParam(new MQLParam("typeFilter", "The type liste (*|A&S&L&G&I&O&U&D&R)", "string", true));
		mql.addParam(new MQLParam("typeOrder", "The type liste (*|A&S&L&G&I&O&U&D&R)", "string", true));
		mql.addParam(new MQLParam("nbResult", "The number max of result into the tab", "string", true));
		mql.addParam(new MQLParam("depth", "The depth to search", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation set condition", "To set a condition into a relation", "relation set condition RG[w] TH1 {if {or (string equal [TH1] \"dog\") (string equal [TH1] \"cat\")} {true} {false}}", "Condition inserted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		mql.addParam(new MQLParam("target", "The target (R|TH1|THN)", "string", true));
		mql.addParam(new MQLParam("mql", "The MQL source code", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation set action", "To set the action into a relation", "relation set action RG[w] {log trace \"OK\";}", "Action inserted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationTabLinkId", "The id", "string", true));
		mql.addParam(new MQLParam("mql", "The MQL source code", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation set else action", "To set the else action into a relation", "relation set else action RG[w] {log trace \"OK\";}", "Else action inserted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationTabLinkId", "The relation id", "string", true));
		mql.addParam(new MQLParam("mql", "The MQL source code", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show conditions", "To show all conditions from a relation", "relation show conditions RG[w]", "{<br>  \"TH1\": \"if {or (string equal [TH1] \\\"dog\\\") (string equal [TH1] \\\"cat\\\")} {true} {false};\"<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show action", "To show the action from a relation", "relation show action RG[w]", "log trace \"OK\";", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation exist condition", "To check if a relation has a condition", "relation exist condition RG[w]", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation generate condition", "To generate all conditions from a relation", "relation generate condition RG[w]", "relation set condition RG[w] TH1 {if {or (string equal [TH1] \"dog\") (string equal [TH1] \"cat\")} {true} {false};};", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation generate action", "To generate action from a relation", "relation generate action RG[w]", "relation set action RG[w] {log trace \"OK\";};", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show", "To show all tab links in a thought or in a relation tab link", "relation show \"[RG[v] 2/3]\" G", "[<br>  \"[RG[w] 2/3]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtIdOrRelationIdOrRelationTabLinkId", "The thought id or the relation tab link id", "string", true));
		mql.addParam(new MQLParam("type", "The branch type ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show perceptions", "To show all tab link perceptions in a thought or in a relation tab link", "relation show perceptions \"[RG[v] 2/3]\" G", "[<br>  {<br>    \"v\": 0.0,<br>    \"w\": 0,<br>    \"mw\": 0,<br>    \"k\": \"[RG[w] 2/3]\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtIdOrRelationIdOrRelationTabLinkId", "The thought id or the relation tab link id", "string", true));
		mql.addParam(new MQLParam("type", "The branch type ("+DeepSearchEngine.allTypeList+")", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show thoughts", "To show all thoughts in a relation", "relation show thoughts RG[w]", "[<br>  \"TH[]\",<br>  \"TH[2]\",<br>  \"TH[72]\",<br>  \"TH[73]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show thought nodes", "To show all thought nodes in a relation", "relation show thought nodes RG[w]", "[<br>  \"TH[]\",<br>  \"TH[2]\",<br>  \"RU[n]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation show words", "To show all words in a relation", "relation show words RG[w]", "[<br>  [<br>    {<br>      \"lang\": \"en\",<br>      \"word\": \"W[something]\"<br>    }<br>  ],<br>  [<br>    {<br>      \"lang\": \"en\",<br>      \"word\": \"W[is]\"<br>    }<br>  ],<br>  [<br>    {<br>      \"lang\": \"en\",<br>      \"word\": \"W[an animal]\"<br>    }<br>  ]<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation pos no something", "To show the position of the first not something thought in a relation", "relation pos no something RG[w]", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation thought stimulate", "To stimulate all relations in a thought", "relation thought stimulate \"TH[73]\"", "Relations stimulated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation stimulate", "To stimulate a relation", "relation stimulate RG[w]", "Relation RG[w] stimulated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation check condition", "Do not use! This function is used by mental process to check a condition into a relation", "relation check condition \"admin\"", "Do not use!", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation condition", "Do not use! This function is used by mental process to check conditions from another relation", "relation condition RL[2]", "Do not use!", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation execute action", "Do not use! This function is used by mental process to execute an action into a relation", "relation execute action \"admin\"", "Do not use!", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Relation").add(mql);
		mql = new MQLDocumentation(true, "relation delete", "To a relation", "relation delete RG[w]", "Relation 'RG[w]' deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Relation").add(mql);
		
		functions.put("Circle", new Vector<MQLDocumentation>());
		page_description.put("Circle", "<img src='images/p.png' style='vertical-align: middle;'>Here you can add a relation into a circle of relations."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>All relations in a circle have the same meaning (with the same or different languages)."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Warning: All relations within a circle must be combined together."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Warning: Position blocks are separate with a space (and each block separate with | ).");
		mql = new MQLDocumentation(true, "circle merge", "To merge two circles (Warning: All relations within a circle must be combined together.)", "-> \"[R1]\" (relation new \"le soleil brille\" L fr);<br>-> \"[R2]\" (relation new \"the sun is shining\" L en);<br>circle merge [R1] \"1 2 3|4\" [R2] \"1 2 3 3\"", "CI[0]", "circle show [R1]", "{<br>  \"en RL[4z] RL[50]\": \"1 2 3 3\",<br>  \"en RL[50] RL[4z]\": \"1 2 3|4\"<br>}", "circle show [R2]", "{<br>  \"en RL[4z] RL[50]\": \"1 2 3 3\",<br>  \"en RL[50] RL[4z]\": \"1 2 3|4\"<br>}", false, "");
		mql.addParam(new MQLParam("relationId1", "The relation id", "string", true));
		mql.addParam(new MQLParam("positions1", "The positions of each node (ex: 1 2 - 3)", "string", true));
		mql.addParam(new MQLParam("relationId2", "The relation id", "string", true));
		mql.addParam(new MQLParam("positions2", "The positions of each node (ex: 1 2 - 3)", "string", true));
		functions.get("Circle").add(mql);
		mql = new MQLDocumentation(true, "circle exist", "To chek if a link already existe into a circle", "circle exist [R1]", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationIdToFind", "The relation id to find", "string", true));
		functions.get("Circle").add(mql);
		mql = new MQLDocumentation(true, "circle show", "To show all relations from a circle", "circle show [R1]", "{<br>  \"en RL[4z] RL[50]\": \"1 2 3 3\",<br>  \"en RL[50] RL[4z]\": \"1 2 3|4\"<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Circle").add(mql);
		mql = new MQLDocumentation(true, "circle ids", "To show all relations ids from a circle", "circle ids [R1]", "[<br>  \"RL[4z]\",<br>  \"RL[50]\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Circle").add(mql);
		mql = new MQLDocumentation(true, "circle delete", "To delete all links by relation from circle", "circle delete [R1]", "The relation is out of the circle.", "circle show [R2]", "{}", "circle show [R1]", "{}", false, "");
		mql.addParam(new MQLParam("relationIdToDelete", "The relation id to delete", "string", true));
		functions.get("Circle").add(mql);
		
		functions.put("MQL Syntax", new Vector<MQLDocumentation>());
		page_description.put("MQL Syntax", "<img src='images/p.png' style='vertical-align: middle;'>You can can see the basic syntax of the MQL language.");
		ghost_functions.put("MQL Syntax", null);
		mql = new MQLDocumentation(true, "category1 categoryN action param1 paramN", "All MQL function respect the format \"category1 categoryN action param1 paramN\"", "who", "mentdb", "concat \"Hello \" \"Adam!\"", "Hello Adam!", "stimulation initialize \"bonjour\"", "[<br>  {<br>    \"thought_list\": \"TH[70]\",<br>    \"reader_found_word\": [<br>      \"bonjour\"<br>    ],<br>    \"first_checked_relation\": null,<br>    \"validated_relation\": null,<br>    \"fromUser\": \"mentdb\",<br>    \"reader_exclusion\": {},<br>    \"io_relation\": \"RI[d]\",<br>    \"_current_action_\": \"INITIALIZATION\",<br>    \"reader_found_type\": [],<br>    \"i_basic_search\": -1.0,<br>    \"output\": [],<br>    \"input\": \"bonjour\",<br>    \"i_last_strategies\": 0.0,<br>    \"_next_action_\": \"basic_search\",<br>    \"last_strategies\": [],<br>    \"has_a_condition\": \"0\"<br>  }<br>]", false, "");
		mql.addParam(new MQLParam("category1", "The category 1", "string", false));
		mql.addParam(new MQLParam("categoryN", "The category N", "string", false));
		mql.addParam(new MQLParam("action", "The category 1", "string", true));
		mql.addParam(new MQLParam("param1", "The parameter 1", "string", false));
		mql.addParam(new MQLParam("paramN", "The parameter 1", "string", false));
		functions.get("MQL Syntax").add(mql);
		
		functions.put("Data Type", new Vector<MQLDocumentation>());
		page_description.put("Data Type", "<img src='images/p.png' style='vertical-align: middle;'>You can can see all data type in the MQL language.");
		ghost_functions.put("Data Type", null);
		mql = new MQLDocumentation(true, "numeric", "Numeric type", "123", "123", "+ 3 4", "7", null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "string", "String type", "\"Hello!\"", "Hello!", "concat \"Hello \" \"Adam\" \"!\"", "Hello Adam!", null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "null", "Null type", "null", "null", "is null (null);", "1", null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "variable", "Variable type", "-> \"[var1]\" 16", "16", "env show;", "{<br>  \"[var2]\": \"16\"<br>}", null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "boolean", "Boolean type", "true", "1", "false", "0", null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "json", "JSON type", "json load \"keyId\" \"{}\";", "1", "json doc \"keyId\";", "{}", null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "comment", "Comment type", "#This is a comment;", "This is a comment ...", null, null, null, null, false, "");
		functions.get("Data Type").add(mql);
		mql = new MQLDocumentation(true, "error", "Error type", "exception (1) (\"your message ...\");", "1: your message ...", null, null, null, null, false, "");
		functions.get("Data Type").add(mql);
		
		functions.put("Link", new Vector<MQLDocumentation>());
		page_description.put("Link", "<img src='images/p.png' style='vertical-align: middle;'>You can create links to certain thoughts and relationships."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Like @[th_en_lion] or @[r_en_lion]."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Access to these thoughts and relationships is faster.");
		mql = new MQLDocumentation(true, "link create thought", "To create a thought link", "link create thought en \"lion\" TH[34]", "@[th_en_lion]", null, null, null, null, false, "link create|insert|add thought");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("thoughtId", "The thought id", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link create relation", "To create a relation link", "link create relation en \"lion\" R[34]", "@[r_en_lion]", null, null, null, null, false, "link create|insert|add relation");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("relationId", "The relation id", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link get value", "To get a value link", "link get value \"@[th_en_lion]\"", "TH[34]", "@[th_en_lion]", "TH[34]", "@[r_en_lion]", "R[34]", false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link get thought", "To get a thought link", "link get thought en \"lion\"", "TH[34]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link get relation", "To get a relation link", "link get relation en \"lion\"", "R[34]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link exist thought", "To check if a thought link already exist", "link exist thought en \"lion\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link exist relation", "To check if a relation link already exist", "link exist relation en \"lion\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link delete thought", "To delete a thought link", "link delete thought en \"lion\"", "Link deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		mql = new MQLDocumentation(true, "link delete relation", "To delete a relation link", "link delete relation en \"lion\"", "Link deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Link").add(mql);
		
		functions.put("Stimulation", new Vector<MQLDocumentation>());
		page_description.put("Stimulation", "<img src='images/p.png' style='vertical-align: middle;'>Mental processes are based on a stimulation table."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>This table allows to memorize all the actions executed by the AI."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can execute clicks on the brain after initialize the stimulation.");
		mql = new MQLDocumentation(true, "stimulation initialize", "To initialize a stimulation into the brain", "stimulation initialize \"bonjour\"", "[<br>  {<br>    \"thought_list\": \"TH[70]\",<br>    \"reader_found_word\": [<br>      \"bonjour\"<br>    ],<br>    \"first_checked_relation\": null,<br>    \"validated_relation\": null,<br>    \"fromUser\": \"mentdb\",<br>    \"reader_exclusion\": {},<br>    \"io_relation\": \"RI[1]\",<br>    \"_current_action_\": \"INITIALIZATION\",<br>    \"reader_found_type\": [],<br>    \"i_basic_search\": -1.0,<br>    \"output\": [],<br>    \"input\": \"bonjour\",<br>    \"i_last_strategies\": 0.0,<br>    \"_next_action_\": \"basic_search\",<br>    \"last_strategies\": [],<br>    \"has_a_condition\": \"0\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation click", "To click (execute) a stimulation into the brain", "stimulation click \"mentdb\"", "[<br>  {<br>    \"thought_list\": \"TH[70]\",<br>    \"reader_found_word\": [<br>      \"bonjour\"<br>    ],<br>    \"first_checked_relation\": null,<br>    \"validated_relation\": null,<br>    \"fromUser\": \"mentdb\",<br>    \"reader_exclusion\": {},<br>    \"io_relation\": \"RI[1]\",<br>    \"_current_action_\": \"INITIALIZATION\",<br>    \"reader_found_type\": [],<br>    \"i_basic_search\": 1.0,<br>    \"output\": [],<br>    \"input\": \"bonjour\",<br>    \"i_last_strategies\": 0.0,<br>    \"_next_action_\": \"basic_search\",<br>    \"last_strategies\": [],<br>    \"has_a_condition\": \"0\"<br>  },<br>  {<br>    \"_next_action_\": \"search_strategy\",<br>    \"exclusion_strategic\": {},<br>    \"exclusion_basic_search_logic\": {},<br>    \"exclusion_logic\": {},<br>    \"_current_action_\": \"BASIC_SEARCH\",<br>    \"exclusion_basic_search_strategic\": {},<br>    \"ordered_basic_search\": [<br>      \"RI[0] 51\",<br>      \"RR[2] 51\"<br>    ]<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation click", "To click (execute) a stimulation into the brain", "stimulation click", "[<br>  {<br>    \"thought_list\": \"TH[70]\",<br>    \"reader_found_word\": [<br>      \"bonjour\"<br>    ],<br>    \"first_checked_relation\": null,<br>    \"validated_relation\": null,<br>    \"fromUser\": \"mentdb\",<br>    \"reader_exclusion\": {},<br>    \"io_relation\": \"RI[1]\",<br>    \"_current_action_\": \"INITIALIZATION\",<br>    \"reader_found_type\": [],<br>    \"i_basic_search\": 1.0,<br>    \"output\": [],<br>    \"input\": \"bonjour\",<br>    \"i_last_strategies\": 0.0,<br>    \"_next_action_\": \"basic_search\",<br>    \"last_strategies\": [],<br>    \"has_a_condition\": \"0\"<br>  },<br>  {<br>    \"_next_action_\": \"search_strategy\",<br>    \"exclusion_strategic\": {<br>      \"RS[1]\": 0<br>    },<br>    \"exclusion_basic_search_logic\": {},<br>    \"exclusion_logic\": {},<br>    \"_current_action_\": \"BASIC_SEARCH\",<br>    \"exclusion_basic_search_strategic\": {<br>      \"RI[0]\": 0<br>    },<br>    \"ordered_basic_search\": [<br>      \"RI[0] 51\",<br>      \"RR[2] 51\"<br>    ]<br>  },<br>  {<br>    \"_next_action_\": \"check_relation\",<br>    \"_current_action_\": \"SEARCH_STRATEGY\",<br>    \"relation\": \"RS[1]\"<br>  }<br>]", null, null, null, null, false, "");
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation show", "To show the stimulation tab", "stimulation show \"mentdb\"", "[<br>  [<br>    {<br>      \"thought_list\": \"TH[70]\",<br>      \"reader_found_word\": [<br>        \"bonjour\"<br>      ],<br>      \"first_checked_relation\": null,<br>      \"validated_relation\": null,<br>      \"fromUser\": \"mentdb\",<br>      \"reader_exclusion\": {},<br>      \"io_relation\": \"RI[1]\",<br>      \"_current_action_\": \"INITIALIZATION\",<br>      \"reader_found_type\": [],<br>      \"i_basic_search\": 1.0,<br>      \"output\": [],<br>      \"input\": \"bonjour\",<br>      \"i_last_strategies\": 0.0,<br>      \"_next_action_\": \"basic_search\",<br>      \"last_strategies\": [],<br>      \"has_a_condition\": \"0\"<br>    },<br>    {<br>      \"_next_action_\": \"search_strategy\",<br>      \"exclusion_strategic\": {<br>        \"RS[1]\": 0<br>      },<br>      \"exclusion_basic_search_logic\": {},<br>      \"exclusion_logic\": {},<br>      \"_current_action_\": \"BASIC\",<br>      \"exclusion_basic_search_strategic\": {<br>        \"RI[0]\": 0<br>      },<br>      \"ordered_basic_search\": [<br>        \"RI[0] 51\",<br>        \"RR[2] 51\"<br>      ]<br>    },<br>    {<br>      \"_next_action_\": \"check_relation\",<br>      \"_current_action_\": \"SEARCH_STRATEGY\",<br>      \"relation\": \"RS[1]\"<br>    }<br>  ]<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation size", "To get the size of the stimulation tab", "stimulation size \"mentdb\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation initialize", "To initialize a stimulation into the brain", "stimulation initialize \"admin\" \"bonjour\"", "[<br>  {<br>    \"thought_list\": \"TH[70]\",<br>    \"reader_found_word\": [<br>      \"bonjour\"<br>    ],<br>    \"first_checked_relation\": null,<br>    \"validated_relation\": null,<br>    \"fromUser\": \"admin\",<br>    \"reader_exclusion\": {},<br>    \"io_relation\": \"RI[3]\",<br>    \"_current_action_\": \"INITIALIZATION\",<br>    \"reader_found_type\": [],<br>    \"i_basic_search\": -1.0,<br>    \"output\": [],<br>    \"input\": \"bonjour\",<br>    \"i_last_strategies\": 0.0,<br>    \"_next_action_\": \"basic_search\",<br>    \"last_strategies\": [],<br>    \"has_a_condition\": \"0\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("str", "The string", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation get", "To get a specific stimulation", "stimulation get \"admin\" 0;", "[<br>  {<br>    \"thought_list\": \"TH[70]\",<br>    \"reader_found_word\": [<br>      \"bonjour\"<br>    ],<br>    \"first_checked_relation\": null,<br>    \"validated_relation\": null,<br>    \"fromUser\": \"admin\",<br>    \"reader_exclusion\": {},<br>    \"io_relation\": \"RI[3]\",<br>    \"_current_action_\": \"INITIALIZATION\",<br>    \"reader_found_type\": [],<br>    \"i_basic_search\": -1.0,<br>    \"output\": [],<br>    \"input\": \"bonjour\",<br>    \"i_last_strategies\": 0.0,<br>    \"_next_action_\": \"basic_search\",<br>    \"last_strategies\": [],<br>    \"has_a_condition\": \"0\"<br>  }<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("index", "The index", "number", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation replace", "To replace a specific stimulation", "stimulation replace \"admin\" \"[]\";", "Stimulation replace with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("json", "The json", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation search basic", "Do not use! This function is used by mental process to make a basic search into the brain", "stimulation search basic \"admin\" \"*\"", "Do not use!", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("typeList", "The type liste (*|A&S&L&G&I&O&U&D&R)", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation search ra_rs", "Do not use! This function is used by mental process to search a RA or RS strategy into the brain", "stimulation search ra_rs \"admin\"", "Do not use!", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation search rl", "Do not use! This function is used by mental process to search a RL relation into the brain", "stimulation search rl \"admin\"", "Do not use!", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation commit", "Do not use! This function is used by mental process to commit a stimulation into the brain by the ai user", "stimulation commit", "Do not use!", null, null, null, null, false, "");
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation create tab", "To create a new tab into a stimulation in the brain", "stimulation create tab \"admin\"", "Tab created with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation remove", "To remove a specific stimulation", "stimulation remove \"mentdb\" 0;", "Stimulation removed with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("index", "The index", "number", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation clear", "To clear the stimulation tab", "stimulation clear \"mentdb\"", "Stimulation cleared with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		functions.get("Stimulation").add(mql);
		mql = new MQLDocumentation(true, "stimulation redirect", "To redirect a relation into the stimulation tab", "stimulation redirect \"admin\" RL[2]", "Redirection executed with successfully.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("data", "The data", "string", true));
		functions.get("Stimulation").add(mql);
		
		/*functions.put("Emotion", new Vector<MQLDocumentation>());
		mql = new MQLDocumentation(false, "emotion create", "To insert a new emotion", "emotion insert peur 1 negative fr", "Emotion peur created with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("emotion", "The emotion", "string", true));
		mql.addParam(new MQLParam("theType", "The (1:male,2:neutre,0:female)", "string", true));
		mql.addParam(new MQLParam("type", "Is positive, negative or neutral", "string", true));
		mql.addParam(new MQLParam("lang", "The language", "string", true));
		functions.get("Emotion").add(mql);
		functions.get("Emotion").add(new MQLDocumentation(false, "emotion show", "To show all emotions", "emotion show", "[<br>  [<br>    {<br>      \"lang\": \"fr\",<br>      \"word\": \"W[la vigilance]\",<br>      \"percent\": 67.0<br>    }<br>  ]<br>]", null, null, null, null, false, ""));
		*/
		
		functions.put("Concentration", new Vector<MQLDocumentation>());
		page_description.put("Concentration", "<img src='images/p.png' style='vertical-align: middle;'>All research is done by a degree of concentration."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>If the concentration is large, the brain will do extensive research."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>If the concentration is low, the brain will search quickly.");
		mql = new MQLDocumentation(true, "concentration show", "To show all the concentrations depth", "concentration show", "{<br>  \"C[basic-search]\": \"10\",<br>  \"C[deep-search]\": \"10\",<br>  \"C[symbol]\": \"200\",<br>  \"C[relation]\": \"200\",<br>  \"C[language]\": \"25\",<br>  \"C[user]\": \"25\",<br>  \"C[emotion]\": \"25\",<br>  \"C[stimulation-tab]\": \"25\"<br>}", null, null, null, null, false, "");
		functions.get("Concentration").add(mql);
		mql = new MQLDocumentation(true, "concentration set depth", "To set the concentration depth", "concentration set depth C[symbol] 200", "Concentration depth saved with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The concentration key", "string", true));
		mql.addParam(new MQLParam("depth", "The depth value", "integer > 0", true));
		functions.get("Concentration").add(mql);
		mql = new MQLDocumentation(true, "concentration depth", "To show the concentration depth", "concentration depth C[symbol]", "200", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The concentration key", "string", true));
		functions.get("Concentration").add(mql);
		
		functions.put("Brain", new Vector<MQLDocumentation>());
		page_description.put("Brain", "<img src='images/p.png' style='vertical-align: middle;'>The brain can be stopped at any time."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can execute queries to manipulate the structure of the brain."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>And then relaunch the brain.");
		page_description.put("Brain", "<img src='images/p.png' style='vertical-align: middle;'>You can stop the execution of mental processes at any time."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Make changes to the storage structure ..."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>And restart the brain.");
		functions.get("Brain").add(new MQLDocumentation(true, "brain stop", "To stop the brain", "brain stop", "The brain has stopped successfully.", null, null, null, null, false, ""));
		functions.get("Brain").add(new MQLDocumentation(true, "brain start", "To start the brain", "brain start", "The brain has started successfully.", null, null, null, null, false, ""));
		functions.get("Brain").add(new MQLDocumentation(true, "brain status", "To show the status of the brain", "brain status", "Brain is running ...", null, null, null, null, false, ""));
		mql = new MQLDocumentation(true, "brain send message", "To send a message to a user", "brain send message \"admin\" 1 \"hello\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("fromUser", "The user", "string", true));
		mql.addParam(new MQLParam("type", "The type", "string", true));
		mql.addParam(new MQLParam("msg", "The message", "string", true));
		functions.get("Brain").add(mql);

		functions.put("External REST API", new Vector<MQLDocumentation>());
		page_description.put("External REST API", "<img src='images/p.png' style='vertical-align: middle;'>You can call remote web services from MentDB."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>And trigger your REST web services in mental process executions."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>HTTP or HTTPS.");
		mql = new MQLDocumentation(true, "rest http", "To execute a REST request through HTTP protocol", "rest http GET \"http://services.groupkt.com/country/get/all\" \"\";", "{  \"RestResponse\" : {    \"messages\" : [ \"More webservices are available at http://www.groupkt.com/post/f2129b88/services.htm\", \"Total [249] records found.\" ],    \"result\" : [ {      \"name\" : \"Afghanistan\",      \"alpha2_code\" : \"AF\",      \"alpha3_code\" : \"AFG\"    }, {      \"name\" : \"Åland Islands\",      \"alpha2_code\" : \"AX\",      \"alpha3_code\" : \"ALA\"    }, {      \"name\" : \"Albania\",      \"alpha2_code\" : \"AL\",      \"alpha3_code\" : \"ALB\"    }, {      \"name\" : \"Algeria\",      \"alpha2_code\" : \"DZ\",      \"alpha3_code\" : \"DZA\"    }, {      \"name\" : \"American Samoa\",      \"alpha2_code\" : \"AS\",      \"alpha3_code\" : \"ASM\"    }, {      \"name\" : \"Andorra\",      \"alpha2_code\" : \"AD\",      \"alpha3_code\" : \"AND\"    }, {      \"name\" : \"Angola\",      \"alpha2_code\" : \"AO\",      \"alpha3_code\" : \"AGO\"    }, {      \"name\" : \"Anguilla\",      \"alpha2_code\" : \"AI\",      \"alpha3_code\" : \"AIA\"    }, {      \"name\" : \"Antarctica\",      \"alpha2_code\" : \"AQ\",      \"alpha3_code\" : \"ATA\"    }, {      \"name\" : \"Antigua and Barbuda\",      \"alpha2_code\" : \"AG\",      \"alpha3_code\" : \"ATG\"    }, {      \"name\" : \"Argentina\",      \"alpha2_code\" : \"AR\",      \"alpha3_code\" : \"ARG\"    }, {      \"name\" : \"Armenia\",      \"alpha2_code\" : \"AM\",      \"alpha3_code\" : \"ARM\"    }, {      \"name\" : \"Aruba\",      \"alpha2_code\" : \"AW\",      \"alpha3_code\" : \"ABW\"    }, {      \"name\" : \"Australia\",      \"alpha2_code\" : \"AU\",      \"alpha3_code\" : \"AUS\"    }, {      \"name\" : \"Austria\",      \"alpha2_code\" : \"AT\",      \"alpha3_code\" : \"AUT\"    }, {      \"name\" : \"Azerbaijan\",      \"alpha2_code\" : \"AZ\",      \"alpha3_code\" : \"AZE\"    }, {      \"name\" : \"Bahamas\",      \"alpha2_code\" : \"BS\",      \"alpha3_code\" : \"BHS\"    }, {      \"name\" : \"Bahrain\",      \"alpha2_code\" : \"BH\",      \"alpha3_code\" : \"BHR\"    }, {      \"name\" : \"Bangladesh\",      \"alpha2_code\" : \"BD\",      \"alpha3_code\" : \"BGD\"    }, {      \"name\" : \"Barbados\",      \"alpha2_code\" : \"BB\",      \"alpha3_code\" : \"BRB\"    }, {      \"name\" : \"Belarus\",      \"alpha2_code\" : \"BY\",      \"alpha3_code\" : \"BLR\"    }, {      \"name\" : \"Belgium\",      \"alpha2_code\" : \"BE\",      \"alpha3_code\" : \"BEL\"    }, {      \"name\" : \"Belize\",      \"alpha2_code\" : \"BZ\",      \"alpha3_code\" : \"BLZ\"    }, {      \"name\" : \"Benin\",      \"alpha2_code\" : \"BJ\",      \"alpha3_code\" : \"BEN\"    }, {      \"name\" : \"Bermuda\",      \"alpha2_code\" : \"BM\",      \"alpha3_code\" : \"BMU\"    }, {      \"name\" : \"Bhutan\",      \"alpha2_code\" : \"BT\",      \"alpha3_code\" : \"BTN\"    }, {      \"name\" : \"Bolivia (Plurinational State of)\",      \"alpha2_code\" : \"BO\",      \"alpha3_code\" : \"BOL\"    }, {      \"name\" : \"Bonaire, Sint Eustatius and Saba\",      \"alpha2_code\" : \"BQ\",      \"alpha3_code\" : \"BES\"    }, {      \"name\" : \"Bosnia and Herzegovina\",      \"alpha2_code\" : \"BA\",      \"alpha3_code\" : \"BIH\"    }, {      \"name\" : \"Botswana\",      \"alpha2_code\" : \"BW\",      \"alpha3_code\" : \"BWA\"    }, {      \"name\" : \"Bouvet Island\",      \"alpha2_code\" : \"BV\",      \"alpha3_code\" : \"BVT\"    }, {      \"name\" : \"Brazil\",      \"alpha2_code\" : \"BR\",      \"alpha3_code\" : \"BRA\"    }, {      \"name\" : \"British Indian Ocean Territory\",      \"alpha2_code\" : \"IO\",      \"alpha3_code\" : \"IOT\"    }, {      \"name\" : \"Brunei Darussalam\",      \"alpha2_code\" : \"BN\",      \"alpha3_code\" : \"BRN\"    }, {      \"name\" : \"Bulgaria\",      \"alpha2_code\" : \"BG\",      \"alpha3_code\" : \"BGR\"    }, {      \"name\" : \"Burkina Faso\",      \"alpha2_code\" : \"BF\",      \"alpha3_code\" : \"BFA\"    }, {      \"name\" : \"Burundi\",      \"alpha2_code\" : \"BI\",      \"alpha3_code\" : \"BDI\"    }, {      \"name\" : \"Cambodia\",      \"alpha2_code\" : \"KH\",      \"alpha3_code\" : \"KHM\"    }, {      \"name\" : \"Cameroon\",      \"alpha2_code\" : \"CM\",      \"alpha3_code\" : \"CMR\"    }, {      \"name\" : \"Canada\",      \"alpha2_code\" : \"CA\",      \"alpha3_code\" : \"CAN\"    }, {      \"name\" : \"Cabo Verde\",      \"alpha2_code\" : \"CV\",      \"alpha3_code\" : \"CPV\"    }, {      \"name\" : \"Cayman Islands\",      \"alpha2_code\" : \"KY\",      \"alpha3_code\" : \"CYM\"    }, {      \"name\" : \"Central African Republic\",      \"alpha2_code\" : \"CF\",      \"alpha3_code\" : \"CAF\"    }, {      \"name\" : \"Chad\",      \"alpha2_code\" : \"TD\",      \"alpha3_code\" : \"TCD\"    }, {      \"name\" : \"Chile\",      \"alpha2_code\" : \"CL\",      \"alpha3_code\" : \"CHL\"    }, {      \"name\" : \"China\",      \"alpha2_code\" : \"CN\",      \"alpha3_code\" : \"CHN\"    }, {      \"name\" : \"Christmas Island\",      \"alpha2_code\" : \"CX\",      \"alpha3_code\" : \"CXR\"    }, {      \"name\" : \"Cocos (Keeling) Islands\",      \"alpha2_code\" : \"CC\",      \"alpha3_code\" : \"CCK\"    }, {      \"name\" : \"Colombia\",      \"alpha2_code\" : \"CO\",      \"alpha3_code\" : \"COL\"    }, {      \"name\" : \"Comoros\",      \"alpha2_code\" : \"KM\",      \"alpha3_code\" : \"COM\"    }, {      \"name\" : \"Congo\",      \"alpha2_code\" : \"CG\",      \"alpha3_code\" : \"COG\"    }, {      \"name\" : \"Congo (Democratic Republic of the)\",      \"alpha2_code\" : \"CD\",      \"alpha3_code\" : \"COD\"    }, {      \"name\" : \"Cook Islands\",      \"alpha2_code\" : \"CK\",      \"alpha3_code\" : \"COK\"    }, {      \"name\" : \"Costa Rica\",      \"alpha2_code\" : \"CR\",      \"alpha3_code\" : \"CRI\"    }, {      \"name\" : \"Côte d'Ivoire\",      \"alpha2_code\" : \"CI\",      \"alpha3_code\" : \"CIV\"    }, {      \"name\" : \"Croatia\",      \"alpha2_code\" : \"HR\",      \"alpha3_code\" : \"HRV\"    }, {      \"name\" : \"Cuba\",      \"alpha2_code\" : \"CU\",      \"alpha3_code\" : \"CUB\"    }, {      \"name\" : \"Curaçao\",      \"alpha2_code\" : \"CW\",      \"alpha3_code\" : \"CUW\"    }, {      \"name\" : \"Cyprus\",      \"alpha2_code\" : \"CY\",      \"alpha3_code\" : \"CYP\"    }, {      \"name\" : \"Czech Republic\",      \"alpha2_code\" : \"CZ\",      \"alpha3_code\" : \"CZE\"    }, {      \"name\" : \"Denmark\",      \"alpha2_code\" : \"DK\",      \"alpha3_code\" : \"DNK\"    }, {      \"name\" : \"Djibouti\",      \"alpha2_code\" : \"DJ\",      \"alpha3_code\" : \"DJI\"    }, {      \"name\" : \"Dominica\",      \"alpha2_code\" : \"DM\",      \"alpha3_code\" : \"DMA\"    }, {      \"name\" : \"Dominican Republic\",      \"alpha2_code\" : \"DO\",      \"alpha3_code\" : \"DOM\"    }, {      \"name\" : \"Ecuador\",      \"alpha2_code\" : \"EC\",      \"alpha3_code\" : \"ECU\"    }, {      \"name\" : \"Egypt\",      \"alpha2_code\" : \"EG\",      \"alpha3_code\" : \"EGY\"    }, {      \"name\" : \"El Salvador\",      \"alpha2_code\" : \"SV\",      \"alpha3_code\" : \"SLV\"    }, {      \"name\" : \"Equatorial Guinea\",      \"alpha2_code\" : \"GQ\",      \"alpha3_code\" : \"GNQ\"    }, {      \"name\" : \"Eritrea\",      \"alpha2_code\" : \"ER\",      \"alpha3_code\" : \"ERI\"    }, {      \"name\" : \"Estonia\",      \"alpha2_code\" : \"EE\",      \"alpha3_code\" : \"EST\"    }, {      \"name\" : \"Ethiopia\",      \"alpha2_code\" : \"ET\",      \"alpha3_code\" : \"ETH\"    }, {      \"name\" : \"Falkland Islands (Malvinas)\",      \"alpha2_code\" : \"FK\",      \"alpha3_code\" : \"FLK\"    }, {      \"name\" : \"Faroe Islands\",      \"alpha2_code\" : \"FO\",      \"alpha3_code\" : \"FRO\"    }, {      \"name\" : \"Fiji\",      \"alpha2_code\" : \"FJ\",      \"alpha3_code\" : \"FJI\"    }, {      \"name\" : \"Finland\",      \"alpha2_code\" : \"FI\",      \"alpha3_code\" : \"FIN\"    }, {      \"name\" : \"France\",      \"alpha2_code\" : \"FR\",      \"alpha3_code\" : \"FRA\"    }, {      \"name\" : \"French Guiana\",      \"alpha2_code\" : \"GF\",      \"alpha3_code\" : \"GUF\"    }, {      \"name\" : \"French Polynesia\",      \"alpha2_code\" : \"PF\",      \"alpha3_code\" : \"PYF\"    }, {      \"name\" : \"French Southern Territories\",      \"alpha2_code\" : \"TF\",      \"alpha3_code\" : \"ATF\"    }, {      \"name\" : \"Gabon\",      \"alpha2_code\" : \"GA\",      \"alpha3_code\" : \"GAB\"    }, {      \"name\" : \"Gambia\",      \"alpha2_code\" : \"GM\",      \"alpha3_code\" : \"GMB\"    }, {      \"name\" : \"Georgia\",      \"alpha2_code\" : \"GE\",      \"alpha3_code\" : \"GEO\"    }, {      \"name\" : \"Germany\",      \"alpha2_code\" : \"DE\",      \"alpha3_code\" : \"DEU\"    }, {      \"name\" : \"Ghana\",      \"alpha2_code\" : \"GH\",      \"alpha3_code\" : \"GHA\"    }, {      \"name\" : \"Gibraltar\",      \"alpha2_code\" : \"GI\",      \"alpha3_code\" : \"GIB\"    }, {      \"name\" : \"Greece\",      \"alpha2_code\" : \"GR\",      \"alpha3_code\" : \"GRC\"    }, {      \"name\" : \"Greenland\",      \"alpha2_code\" : \"GL\",      \"alpha3_code\" : \"GRL\"    }, {      \"name\" : \"Grenada\",      \"alpha2_code\" : \"GD\",      \"alpha3_code\" : \"GRD\"    }, {      \"name\" : \"Guadeloupe\",      \"alpha2_code\" : \"GP\",      \"alpha3_code\" : \"GLP\"    }, {      \"name\" : \"Guam\",      \"alpha2_code\" : \"GU\",      \"alpha3_code\" : \"GUM\"    }, {      \"name\" : \"Guatemala\",      \"alpha2_code\" : \"GT\",      \"alpha3_code\" : \"GTM\"    }, {      \"name\" : \"Guernsey\",      \"alpha2_code\" : \"GG\",      \"alpha3_code\" : \"GGY\"    }, {      \"name\" : \"Guinea\",      \"alpha2_code\" : \"GN\",      \"alpha3_code\" : \"GIN\"    }, {      \"name\" : \"Guinea-Bissau\",      \"alpha2_code\" : \"GW\",      \"alpha3_code\" : \"GNB\"    }, {      \"name\" : \"Guyana\",      \"alpha2_code\" : \"GY\",      \"alpha3_code\" : \"GUY\"    }, {      \"name\" : \"Haiti\",      \"alpha2_code\" : \"HT\",      \"alpha3_code\" : \"HTI\"    }, {      \"name\" : \"Heard Island and McDonald Islands\",      \"alpha2_code\" : \"HM\",      \"alpha3_code\" : \"HMD\"    }, {      \"name\" : \"Holy See\",      \"alpha2_code\" : \"VA\",      \"alpha3_code\" : \"VAT\"    }, {      \"name\" : \"Honduras\",      \"alpha2_code\" : \"HN\",      \"alpha3_code\" : \"HND\"    }, {      \"name\" : \"Hong Kong\",      \"alpha2_code\" : \"HK\",      \"alpha3_code\" : \"HKG\"    }, {      \"name\" : \"Hungary\",      \"alpha2_code\" : \"HU\",      \"alpha3_code\" : \"HUN\"    }, {      \"name\" : \"Iceland\",      \"alpha2_code\" : \"IS\",      \"alpha3_code\" : \"ISL\"    }, {      \"name\" : \"India\",      \"alpha2_code\" : \"IN\",      \"alpha3_code\" : \"IND\"    }, {      \"name\" : \"Indonesia\",      \"alpha2_code\" : \"ID\",      \"alpha3_code\" : \"IDN\"    }, {      \"name\" : \"Iran (Islamic Republic of)\",      \"alpha2_code\" : \"IR\",      \"alpha3_code\" : \"IRN\"    }, {      \"name\" : \"Iraq\",      \"alpha2_code\" : \"IQ\",      \"alpha3_code\" : \"IRQ\"    }, {      \"name\" : \"Ireland\",      \"alpha2_code\" : \"IE\",      \"alpha3_code\" : \"IRL\"    }, {      \"name\" : \"Isle of Man\",      \"alpha2_code\" : \"IM\",      \"alpha3_code\" : \"IMN\"    }, {      \"name\" : \"Israel\",      \"alpha2_code\" : \"IL\",      \"alpha3_code\" : \"ISR\"    }, {      \"name\" : \"Italy\",      \"alpha2_code\" : \"IT\",      \"alpha3_code\" : \"ITA\"    }, {      \"name\" : \"Jamaica\",      \"alpha2_code\" : \"JM\",      \"alpha3_code\" : \"JAM\"    }, {      \"name\" : \"Japan\",      \"alpha2_code\" : \"JP\",      \"alpha3_code\" : \"JPN\"    }, {      \"name\" : \"Jersey\",      \"alpha2_code\" : \"JE\",      \"alpha3_code\" : \"JEY\"    }, {      \"name\" : \"Jordan\",      \"alpha2_code\" : \"JO\",      \"alpha3_code\" : \"JOR\"    }, {      \"name\" : \"Kazakhstan\",      \"alpha2_code\" : \"KZ\",      \"alpha3_code\" : \"KAZ\"    }, {      \"name\" : \"Kenya\",      \"alpha2_code\" : \"KE\",      \"alpha3_code\" : \"KEN\"    }, {      \"name\" : \"Kiribati\",      \"alpha2_code\" : \"KI\",      \"alpha3_code\" : \"KIR\"    }, {      \"name\" : \"Korea (Democratic People's Republic of)\",      \"alpha2_code\" : \"KP\",      \"alpha3_code\" : \"PRK\"    }, {      \"name\" : \"Korea (Republic of)\",      \"alpha2_code\" : \"KR\",      \"alpha3_code\" : \"KOR\"    }, {      \"name\" : \"Kuwait\",      \"alpha2_code\" : \"KW\",      \"alpha3_code\" : \"KWT\"    }, {      \"name\" : \"Kyrgyzstan\",      \"alpha2_code\" : \"KG\",      \"alpha3_code\" : \"KGZ\"    }, {      \"name\" : \"Lao People's Democratic Republic\",      \"alpha2_code\" : \"LA\",      \"alpha3_code\" : \"LAO\"    }, {      \"name\" : \"Latvia\",      \"alpha2_code\" : \"LV\",      \"alpha3_code\" : \"LVA\"    }, {      \"name\" : \"Lebanon\",      \"alpha2_code\" : \"LB\",      \"alpha3_code\" : \"LBN\"    }, {      \"name\" : \"Lesotho\",      \"alpha2_code\" : \"LS\",      \"alpha3_code\" : \"LSO\"    }, {      \"name\" : \"Liberia\",      \"alpha2_code\" : \"LR\",      \"alpha3_code\" : \"LBR\"    }, {      \"name\" : \"Libya\",      \"alpha2_code\" : \"LY\",      \"alpha3_code\" : \"LBY\"    }, {      \"name\" : \"Liechtenstein\",      \"alpha2_code\" : \"LI\",      \"alpha3_code\" : \"LIE\"    }, {      \"name\" : \"Lithuania\",      \"alpha2_code\" : \"LT\",      \"alpha3_code\" : \"LTU\"    }, {      \"name\" : \"Luxembourg\",      \"alpha2_code\" : \"LU\",      \"alpha3_code\" : \"LUX\"    }, {      \"name\" : \"Macao\",      \"alpha2_code\" : \"MO\",      \"alpha3_code\" : \"MAC\"    }, {      \"name\" : \"Macedonia (the former Yugoslav Republic of)\",      \"alpha2_code\" : \"MK\",      \"alpha3_code\" : \"MKD\"    }, {      \"name\" : \"Madagascar\",      \"alpha2_code\" : \"MG\",      \"alpha3_code\" : \"MDG\"    }, {      \"name\" : \"Malawi\",      \"alpha2_code\" : \"MW\",      \"alpha3_code\" : \"MWI\"    }, {      \"name\" : \"Malaysia\",      \"alpha2_code\" : \"MY\",      \"alpha3_code\" : \"MYS\"    }, {      \"name\" : \"Maldives\",      \"alpha2_code\" : \"MV\",      \"alpha3_code\" : \"MDV\"    }, {      \"name\" : \"Mali\",      \"alpha2_code\" : \"ML\",      \"alpha3_code\" : \"MLI\"    }, {      \"name\" : \"Malta\",      \"alpha2_code\" : \"MT\",      \"alpha3_code\" : \"MLT\"    }, {      \"name\" : \"Marshall Islands\",      \"alpha2_code\" : \"MH\",      \"alpha3_code\" : \"MHL\"    }, {      \"name\" : \"Martinique\",      \"alpha2_code\" : \"MQ\",      \"alpha3_code\" : \"MTQ\"    }, {      \"name\" : \"Mauritania\",      \"alpha2_code\" : \"MR\",      \"alpha3_code\" : \"MRT\"    }, {      \"name\" : \"Mauritius\",      \"alpha2_code\" : \"MU\",      \"alpha3_code\" : \"MUS\"    }, {      \"name\" : \"Mayotte\",      \"alpha2_code\" : \"YT\",      \"alpha3_code\" : \"MYT\"    }, {      \"name\" : \"Mexico\",      \"alpha2_code\" : \"MX\",      \"alpha3_code\" : \"MEX\"    }, {      \"name\" : \"Micronesia (Federated States of)\",      \"alpha2_code\" : \"FM\",      \"alpha3_code\" : \"FSM\"    }, {      \"name\" : \"Moldova (Republic of)\",      \"alpha2_code\" : \"MD\",      \"alpha3_code\" : \"MDA\"    }, {      \"name\" : \"Monaco\",      \"alpha2_code\" : \"MC\",      \"alpha3_code\" : \"MCO\"    }, {      \"name\" : \"Mongolia\",      \"alpha2_code\" : \"MN\",      \"alpha3_code\" : \"MNG\"    }, {      \"name\" : \"Montenegro\",      \"alpha2_code\" : \"ME\",      \"alpha3_code\" : \"MNE\"    }, {      \"name\" : \"Montserrat\",      \"alpha2_code\" : \"MS\",      \"alpha3_code\" : \"MSR\"    }, {      \"name\" : \"Morocco\",      \"alpha2_code\" : \"MA\",      \"alpha3_code\" : \"MAR\"    }, {      \"name\" : \"Mozambique\",      \"alpha2_code\" : \"MZ\",      \"alpha3_code\" : \"MOZ\"    }, {      \"name\" : \"Myanmar\",      \"alpha2_code\" : \"MM\",      \"alpha3_code\" : \"MMR\"    }, {      \"name\" : \"Namibia\",      \"alpha2_code\" : \"NA\",      \"alpha3_code\" : \"NAM\"    }, {      \"name\" : \"Nauru\",      \"alpha2_code\" : \"NR\",      \"alpha3_code\" : \"NRU\"    }, {      \"name\" : \"Nepal\",      \"alpha2_code\" : \"NP\",      \"alpha3_code\" : \"NPL\"    }, {      \"name\" : \"Netherlands\",      \"alpha2_code\" : \"NL\",      \"alpha3_code\" : \"NLD\"    }, {      \"name\" : \"New Caledonia\",      \"alpha2_code\" : \"NC\",      \"alpha3_code\" : \"NCL\"    }, {      \"name\" : \"New Zealand\",      \"alpha2_code\" : \"NZ\",      \"alpha3_code\" : \"NZL\"    }, {      \"name\" : \"Nicaragua\",      \"alpha2_code\" : \"NI\",      \"alpha3_code\" : \"NIC\"    }, {      \"name\" : \"Niger\",      \"alpha2_code\" : \"NE\",      \"alpha3_code\" : \"NER\"    }, {      \"name\" : \"Nigeria\",      \"alpha2_code\" : \"NG\",      \"alpha3_code\" : \"NGA\"    }, {      \"name\" : \"Niue\",      \"alpha2_code\" : \"NU\",      \"alpha3_code\" : \"NIU\"    }, {      \"name\" : \"Norfolk Island\",      \"alpha2_code\" : \"NF\",      \"alpha3_code\" : \"NFK\"    }, {      \"name\" : \"Northern Mariana Islands\",      \"alpha2_code\" : \"MP\",      \"alpha3_code\" : \"MNP\"    }, {      \"name\" : \"Norway\",      \"alpha2_code\" : \"NO\",      \"alpha3_code\" : \"NOR\"    }, {      \"name\" : \"Oman\",      \"alpha2_code\" : \"OM\",      \"alpha3_code\" : \"OMN\"    }, {      \"name\" : \"Pakistan\",      \"alpha2_code\" : \"PK\",      \"alpha3_code\" : \"PAK\"    }, {      \"name\" : \"Palau\",      \"alpha2_code\" : \"PW\",      \"alpha3_code\" : \"PLW\"    }, {      \"name\" : \"Palestine, State of\",      \"alpha2_code\" : \"PS\",      \"alpha3_code\" : \"PSE\"    }, {      \"name\" : \"Panama\",      \"alpha2_code\" : \"PA\",      \"alpha3_code\" : \"PAN\"    }, {      \"name\" : \"Papua New Guinea\",      \"alpha2_code\" : \"PG\",      \"alpha3_code\" : \"PNG\"    }, {      \"name\" : \"Paraguay\",      \"alpha2_code\" : \"PY\",      \"alpha3_code\" : \"PRY\"    }, {      \"name\" : \"Peru\",      \"alpha2_code\" : \"PE\",      \"alpha3_code\" : \"PER\"    }, {      \"name\" : \"Philippines\",      \"alpha2_code\" : \"PH\",      \"alpha3_code\" : \"PHL\"    }, {      \"name\" : \"Pitcairn\",      \"alpha2_code\" : \"PN\",      \"alpha3_code\" : \"PCN\"    }, {      \"name\" : \"Poland\",      \"alpha2_code\" : \"PL\",      \"alpha3_code\" : \"POL\"    }, {      \"name\" : \"Portugal\",      \"alpha2_code\" : \"PT\",      \"alpha3_code\" : \"PRT\"    }, {      \"name\" : \"Puerto Rico\",      \"alpha2_code\" : \"PR\",      \"alpha3_code\" : \"PRI\"    }, {      \"name\" : \"Qatar\",      \"alpha2_code\" : \"QA\",      \"alpha3_code\" : \"QAT\"    }, {      \"name\" : \"Réunion\",      \"alpha2_code\" : \"RE\",      \"alpha3_code\" : \"REU\"    }, {      \"name\" : \"Romania\",      \"alpha2_code\" : \"RO\",      \"alpha3_code\" : \"ROU\"    }, {      \"name\" : \"Russian Federation\",      \"alpha2_code\" : \"RU\",      \"alpha3_code\" : \"RUS\"    }, {      \"name\" : \"Rwanda\",      \"alpha2_code\" : \"RW\",      \"alpha3_code\" : \"RWA\"    }, {      \"name\" : \"Saint Barthélemy\",      \"alpha2_code\" : \"BL\",      \"alpha3_code\" : \"BLM\"    }, {      \"name\" : \"Saint Helena, Ascension and Tristan da Cunha\",      \"alpha2_code\" : \"SH\",      \"alpha3_code\" : \"SHN\"    }, {      \"name\" : \"Saint Kitts and Nevis\",      \"alpha2_code\" : \"KN\",      \"alpha3_code\" : \"KNA\"    }, {      \"name\" : \"Saint Lucia\",      \"alpha2_code\" : \"LC\",      \"alpha3_code\" : \"LCA\"    }, {      \"name\" : \"Saint Martin (French part)\",      \"alpha2_code\" : \"MF\",      \"alpha3_code\" : \"MAF\"    }, {      \"name\" : \"Saint Pierre and Miquelon\",      \"alpha2_code\" : \"PM\",      \"alpha3_code\" : \"SPM\"    }, {      \"name\" : \"Saint Vincent and the Grenadines\",      \"alpha2_code\" : \"VC\",      \"alpha3_code\" : \"VCT\"    }, {      \"name\" : \"Samoa\",      \"alpha2_code\" : \"WS\",      \"alpha3_code\" : \"WSM\"    }, {      \"name\" : \"San Marino\",      \"alpha2_code\" : \"SM\",      \"alpha3_code\" : \"SMR\"    }, {      \"name\" : \"Sao Tome and Principe\",      \"alpha2_code\" : \"ST\",      \"alpha3_code\" : \"STP\"    }, {      \"name\" : \"Saudi Arabia\",      \"alpha2_code\" : \"SA\",      \"alpha3_code\" : \"SAU\"    }, {      \"name\" : \"Senegal\",      \"alpha2_code\" : \"SN\",      \"alpha3_code\" : \"SEN\"    }, {      \"name\" : \"Serbia\",      \"alpha2_code\" : \"RS\",      \"alpha3_code\" : \"SRB\"    }, {      \"name\" : \"Seychelles\",      \"alpha2_code\" : \"SC\",      \"alpha3_code\" : \"SYC\"    }, {      \"name\" : \"Sierra Leone\",      \"alpha2_code\" : \"SL\",      \"alpha3_code\" : \"SLE\"    }, {      \"name\" : \"Singapore\",      \"alpha2_code\" : \"SG\",      \"alpha3_code\" : \"SGP\"    }, {      \"name\" : \"Sint Maarten (Dutch part)\",      \"alpha2_code\" : \"SX\",      \"alpha3_code\" : \"SXM\"    }, {      \"name\" : \"Slovakia\",      \"alpha2_code\" : \"SK\",      \"alpha3_code\" : \"SVK\"    }, {      \"name\" : \"Slovenia\",      \"alpha2_code\" : \"SI\",      \"alpha3_code\" : \"SVN\"    }, {      \"name\" : \"Solomon Islands\",      \"alpha2_code\" : \"SB\",      \"alpha3_code\" : \"SLB\"    }, {      \"name\" : \"Somalia\",      \"alpha2_code\" : \"SO\",      \"alpha3_code\" : \"SOM\"    }, {      \"name\" : \"South Africa\",      \"alpha2_code\" : \"ZA\",      \"alpha3_code\" : \"ZAF\"    }, {      \"name\" : \"South Georgia and the South Sandwich Islands\",      \"alpha2_code\" : \"GS\",      \"alpha3_code\" : \"SGS\"    }, {      \"name\" : \"South Sudan\",      \"alpha2_code\" : \"SS\",      \"alpha3_code\" : \"SSD\"    }, {      \"name\" : \"Spain\",      \"alpha2_code\" : \"ES\",      \"alpha3_code\" : \"ESP\"    }, {      \"name\" : \"Sri Lanka\",      \"alpha2_code\" : \"LK\",      \"alpha3_code\" : \"LKA\"    }, {      \"name\" : \"Sudan\",      \"alpha2_code\" : \"SD\",      \"alpha3_code\" : \"SDN\"    }, {      \"name\" : \"Suriname\",      \"alpha2_code\" : \"SR\",      \"alpha3_code\" : \"SUR\"    }, {      \"name\" : \"Svalbard and Jan Mayen\",      \"alpha2_code\" : \"SJ\",      \"alpha3_code\" : \"SJM\"    }, {      \"name\" : \"Swaziland\",      \"alpha2_code\" : \"SZ\",      \"alpha3_code\" : \"SWZ\"    }, {      \"name\" : \"Sweden\",      \"alpha2_code\" : \"SE\",      \"alpha3_code\" : \"SWE\"    }, {      \"name\" : \"Switzerland\",      \"alpha2_code\" : \"CH\",      \"alpha3_code\" : \"CHE\"    }, {      \"name\" : \"Syrian Arab Republic\",      \"alpha2_code\" : \"SY\",      \"alpha3_code\" : \"SYR\"    }, {      \"name\" : \"Taiwan, Province of China\",      \"alpha2_code\" : \"TW\",      \"alpha3_code\" : \"TWN\"    }, {      \"name\" : \"Tajikistan\",      \"alpha2_code\" : \"TJ\",      \"alpha3_code\" : \"TJK\"    }, {      \"name\" : \"Tanzania, United Republic of\",      \"alpha2_code\" : \"TZ\",      \"alpha3_code\" : \"TZA\"    }, {      \"name\" : \"Thailand\",      \"alpha2_code\" : \"TH\",      \"alpha3_code\" : \"THA\"    }, {      \"name\" : \"Timor-Leste\",      \"alpha2_code\" : \"TL\",      \"alpha3_code\" : \"TLS\"    }, {      \"name\" : \"Togo\",      \"alpha2_code\" : \"TG\",      \"alpha3_code\" : \"TGO\"    }, {      \"name\" : \"Tokelau\",      \"alpha2_code\" : \"TK\",      \"alpha3_code\" : \"TKL\"    }, {      \"name\" : \"Tonga\",      \"alpha2_code\" : \"TO\",      \"alpha3_code\" : \"TON\"    }, {      \"name\" : \"Trinidad and Tobago\",      \"alpha2_code\" : \"TT\",      \"alpha3_code\" : \"TTO\"    }, {      \"name\" : \"Tunisia\",      \"alpha2_code\" : \"TN\",      \"alpha3_code\" : \"TUN\"    }, {      \"name\" : \"Turkey\",      \"alpha2_code\" : \"TR\",      \"alpha3_code\" : \"TUR\"    }, {      \"name\" : \"Turkmenistan\",      \"alpha2_code\" : \"TM\",      \"alpha3_code\" : \"TKM\"    }, {      \"name\" : \"Turks and Caicos Islands\",      \"alpha2_code\" : \"TC\",      \"alpha3_code\" : \"TCA\"    }, {      \"name\" : \"Tuvalu\",      \"alpha2_code\" : \"TV\",      \"alpha3_code\" : \"TUV\"    }, {      \"name\" : \"Uganda\",      \"alpha2_code\" : \"UG\",      \"alpha3_code\" : \"UGA\"    }, {      \"name\" : \"Ukraine\",      \"alpha2_code\" : \"UA\",      \"alpha3_code\" : \"UKR\"    }, {      \"name\" : \"United Arab Emirates\",      \"alpha2_code\" : \"AE\",      \"alpha3_code\" : \"ARE\"    }, {      \"name\" : \"United Kingdom of Great Britain and Northern Ireland\",      \"alpha2_code\" : \"GB\",      \"alpha3_code\" : \"GBR\"    }, {      \"name\" : \"United States of America\",      \"alpha2_code\" : \"US\",      \"alpha3_code\" : \"USA\"    }, {      \"name\" : \"United States Minor Outlying Islands\",      \"alpha2_code\" : \"UM\",      \"alpha3_code\" : \"UMI\"    }, {      \"name\" : \"Uruguay\",      \"alpha2_code\" : \"UY\",      \"alpha3_code\" : \"URY\"    }, {      \"name\" : \"Uzbekistan\",      \"alpha2_code\" : \"UZ\",      \"alpha3_code\" : \"UZB\"    }, {      \"name\" : \"Vanuatu\",      \"alpha2_code\" : \"VU\",      \"alpha3_code\" : \"VUT\"    }, {      \"name\" : \"Venezuela (Bolivarian Republic of)\",      \"alpha2_code\" : \"VE\",      \"alpha3_code\" : \"VEN\"    }, {      \"name\" : \"Viet Nam\",      \"alpha2_code\" : \"VN\",      \"alpha3_code\" : \"VNM\"    }, {      \"name\" : \"Virgin Islands (British)\",      \"alpha2_code\" : \"VG\",      \"alpha3_code\" : \"VGB\"    }, {      \"name\" : \"Virgin Islands (U.S.)\",      \"alpha2_code\" : \"VI\",      \"alpha3_code\" : \"VIR\"    }, {      \"name\" : \"Wallis and Futuna\",      \"alpha2_code\" : \"WF\",      \"alpha3_code\" : \"WLF\"    }, {      \"name\" : \"Western Sahara\",      \"alpha2_code\" : \"EH\",      \"alpha3_code\" : \"ESH\"    }, {      \"name\" : \"Yemen\",      \"alpha2_code\" : \"YE\",      \"alpha3_code\" : \"YEM\"    }, {      \"name\" : \"Zambia\",      \"alpha2_code\" : \"ZM\",      \"alpha3_code\" : \"ZMB\"    }, {      \"name\" : \"Zimbabwe\",      \"alpha2_code\" : \"ZW\",      \"alpha3_code\" : \"ZWE\"    } ]  }}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("method", "The method (GET|POST|PUT|DELETE|PATCH)", "string", true));
		mql.addParam(new MQLParam("beginUrl", "The begin url", "string", true));
		mql.addParam(new MQLParam("endUrl", "The end url", "string", true));
		functions.get("External REST API").add(mql);
		mql = new MQLDocumentation(true, "rest https", "To execute a REST request through HTTPS protocol", "rest https GET \"https://jsonplaceholder.typicode.com/posts/1\" \"\";", "{  \"userId\": 1,  \"id\": 1,  \"title\": \"sunt aut facere repellat provident occaecati excepturi optio reprehenderit\",  \"body\": \"quia et suscipit\nsuscipit recusandae consequuntur expedita et cum\nreprehenderit molestiae ut ut quas totam\nnostrum rerum est autem sunt rem eveniet architecto\"}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("method", "The method (GET|POST|PUT|DELETE|PATCH)", "string", true));
		mql.addParam(new MQLParam("beginUrl", "The begin url", "string", true));
		mql.addParam(new MQLParam("endUrl", "The end url", "string", true));
		functions.get("External REST API").add(mql);

		functions.put("Script", new Vector<MQLDocumentation>());
		page_description.put("Script", "<img src='images/p.png' style='vertical-align: middle;'>MQL functions can be saved in a script."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can call these scripts inside the mental processes."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Or directly in web-service.");
		mql = new MQLDocumentation(true, "script add", "To add a new script", "script add \"division\" 0 \"[v1],[v2]\" {\n  / [v1] [v2];\n};", "Script added with successful.", null, null, null, null, false, "script create|insert|add");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("isInMemory", "Is the script in memory ?", "boolean", true));
		mql.addParam(new MQLParam("variables", "The variables separate by ','", "string", true));
		mql.addParam(new MQLParam("script", "The script", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script update", "To update a script", "script update \"division\" 0 \"[v1],[v2]\" {/ [v1] [v2];};", "Script updated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("isInMemory", "Is the script in memory ?", "boolean", true));
		mql.addParam(new MQLParam("variables", "The variables separate by ','", "string", true));
		mql.addParam(new MQLParam("script", "The script", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script get", "To get a script object", "script get \"division\"", "{<br>  \"mem\": \"0\",<br>  \"mql\": \"/ [v1] [v2];\",<br>  \"k\": \"division\",<br>  \"vars\": \"[v1],[v2]\"<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script generate update", "To generate an update script request", "script generate update \"division\"", "script update \"division\" 0 \"[v1],[v2]\" {/ [v1] [v2];};", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script exist", "To check if a script already exist", "script exist \"division\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script show", "To show all scripts", "script show", "{<br>  \"action.execute_relation_rl\": 0,<br>  \"brain.internal_stimulation\": 0,<br>  \"action.execute_relation_ra\": 0,<br>  \"perf.symbol\": 0,<br>  \"division\": 0,<br>  \"action.search_logical\": 0,<br>  \"action.select_reader\": 0,<br>  \"server.version\": 0,<br>  \"action.closed\": 0,<br>  \"stimulation.last\": 0,<br>  \"action.get_last_relation\": 0,<br>  \"action.redirection_relation\": 0,<br>  \"action.check_relation\": 0,<br>  \"reset_server_id\": 0,<br>  \"perf.relation\": 0,<br>  \"process\\u003cclick\\u003e\": 0,<br>  \"action.relation_response\": 0,<br>  \"reader.space_separator\": 0,<br>  \"server.info\": 0,<br>  \"perf.thought\": 0,<br>  \"brain.stimulate\": 0,<br>  \"create_word_fr_en\": 0,<br>  \"perf.node\": 0,<br>  \"process\\u003cinitialize\\u003e\": 0,<br>  \"action.basic_search\": 0,<br>  \"perf.word\": 0,<br>  \"action.execute_relation_rs\": 0,<br>  \"action.relation_redirection\": 0,<br>  \"action.search_strategy\": 0<br>}", null, null, null, null, false, "");
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script unload accelerator", "To unload script accelerator", "script unload accelerator", "Script accelerator unloaded with successful.", null, null, null, null, false, "");
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "execute", "To execute a script in a new environment", "execute \"division\" \"[v1]\" 10 \"[v2]\" 5", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("varName1", "The variable name 1", "string", false));
		mql.addParam(new MQLParam("varlue1", "The value 1", "string", false));
		mql.addParam(new MQLParam("varNameN", "The variable name N", "string", false));
		mql.addParam(new MQLParam("varlueN", "The value N", "string", false));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "include", "To execute a script in the parent environment", "include \"division\" \"[v1]\" 10 \"[v2]\" 5", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("varName1", "The variable name 1", "string", false));
		mql.addParam(new MQLParam("varlue1", "The value 1", "string", false));
		mql.addParam(new MQLParam("varNameN", "The variable name N", "string", false));
		mql.addParam(new MQLParam("varlueN", "The value N", "string", false));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script is granted", "To check if a script is granted to a specific user", "script is granted \"division\" \"admin\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script generate url", "To generate an URL to execute a mental script", "script generate url \"division\"", "http://localhost:9999/api-rest/?token=ZcbtW2arBcZsANZUrtrBMCa6rdIDIRHFIia5kTGntYzMrAcl7vj2AKkaZcikoQdGVq3wnT&script=division&v1=&v2=", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script generate execute", "To generate a MQL execute script", "script generate execute \"division\"", "execute \"division\"<br>	\"[v1]\" \"\"<br>	\"[v2]\" \"\"<br>;", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script generate include", "To generate a MQL include script", "script generate include \"division\"", "include \"division\"<br>	\"[v1]\" \"\"<br>	\"[v2]\" \"\"<br>;", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script show groups", "To show all groups for a specific script", "script show groups \"division\"", "[<br>  \"all\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script show users", "To show all granted users for a specific script", "script show users \"division\"", "[<br>  \"mentdb\",<br>  \"ai\",<br>  \"admin\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		mql = new MQLDocumentation(true, "script delete", "To delete a script", "script delete \"division\"", "Script deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		functions.get("Script").add(mql);
		
		functions.put("Statement", new Vector<MQLDocumentation>());
		page_description.put("Statement", "<img src='images/p.png' style='vertical-align: middle;'>Like all languages, you have the classic statements functions.");
		mql = new MQLDocumentation(true, "eval", "Eval a MQL expression", "eval \"log trace OK!;\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("mql", "The MQL source code", "string", true));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "if", "IF statement", "if (true) {<br><br>	log trace OK!<br><br>} {<br><br>	log trace KO!<br><br>};", "1", "if (true) {<br><br>	log trace hello!<br><br>}", "1", null, null, false, "");
		mql.addParam(new MQLParam("condition", "The condition", "string", true));
		mql.addParam(new MQLParam("trueAction", "The true action", "boolean", true));
		mql.addParam(new MQLParam("falseAction", "The false action", "boolean", false));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "if force", "IF statement (force true and false execution)", "if force (true) {<br><br>	log trace OK!<br><br>} {<br><br>	log trace KO!<br><br>};", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("condition", "The condition", "string", true));
		mql.addParam(new MQLParam("trueAction", "The true action", "boolean", true));
		mql.addParam(new MQLParam("falseAction", "The false action", "boolean", true));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "case", "CASE statement", "case<br>	(true) {A}<br>	(false) {B}<br>	{C}<br>;", "A", null, null, null, null, false, "");
		mql.addParam(new MQLParam("condition1", "The condition 1", "boolean", true));
		mql.addParam(new MQLParam("action1", "The action 1", "string", true));
		mql.addParam(new MQLParam("conditionN", "The condition N", "boolean", true));
		mql.addParam(new MQLParam("actionN", "The action N", "string", true));
		mql.addParam(new MQLParam("elseAction", "The else action", "string", false));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "for", "FOR statement", "for (-> \"[i]\" 0) (< [i] 45) (++ \"[i]\") {<br><br>	log trace [i];<br><br>};", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("init", "The init mql source code", "string", true));
		mql.addParam(new MQLParam("condition", "The condition", "boolean", true));
		mql.addParam(new MQLParam("increment", "The increment", "string", true));
		mql.addParam(new MQLParam("action", "The action", "string", true));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "while", "WHILE statement", "-> \"[i]\" 0;<br>while (< [i] 45) {<br><br>	log trace [i];<br>	++ \"[i]\";<br><br>};", "45", null, null, null, null, false, "");
		mql.addParam(new MQLParam("condition", "The condition", "boolean", true));
		mql.addParam(new MQLParam("action", "The action", "string", true));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "repeat", "REPEAT statement", "-> \"[i]\" 0;<br>repeat (< [i] 45) {<br><br>	log trace [i];<br>	++ \"[i]\";<br><br>};", "45", null, null, null, null, false, "");
		mql.addParam(new MQLParam("condition", "The condition", "boolean", true));
		mql.addParam(new MQLParam("action", "The action", "string", true));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "exception", "EXCEPTION statement", "exception (1) (\"your message ...\")", "1: your message ...", null, null, null, null, false, "");
		mql.addParam(new MQLParam("id", "The error message id", "string", true));
		mql.addParam(new MQLParam("message", "The error message", "string", true));
		functions.get("Statement").add(mql);
		mql = new MQLDocumentation(true, "try", "TRY statement", "try {<br><br>	exception (1) (\"your messffage ...\");<br><br>} {<br><br>	log trace [err]; <br><br>} \"[err]\";", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("action", "The main action", "string", true));
		mql.addParam(new MQLParam("errorAction", "The error action if an error was generated", "string", true));
		mql.addParam(new MQLParam("idException", "The variable to save the error message", "string", true));
		functions.get("Statement").add(mql);
		
		functions.put("Operator", new Vector<MQLDocumentation>());
		page_description.put("Operator", "<img src='images/p.png' style='vertical-align: middle;'>Like all languages, you have the basic operators.");
		mql = new MQLDocumentation(true, "+", "addition", "+ 23 42", "65", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "-", "substration", "- 50 26", "24", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "*", "multiplication", "* 2 5", "10", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "/", "division", "/ 30 2", "15", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "and", "and boolean", "and 1 1", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("boolean1", "The boolean 1", "boolean", true));
		mql.addParam(new MQLParam("boolean2", "The boolean 2", "boolean", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "or", "or boolean", "or 1 0", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("boolean1", "The boolean 1", "boolean", true));
		mql.addParam(new MQLParam("boolean2", "The boolean 2", "boolean", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "xor", "xor boolean", "xor 1 0", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("boolean1", "The boolean 1", "boolean", true));
		mql.addParam(new MQLParam("boolean2", "The boolean 2", "boolean", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "not", "not boolean", "not 0", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("boolean", "The boolean", "boolean", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "is null", "Check if the value is null", "is null (null)", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "is not null", "Check if the value is not null", "is not null (null)", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "is null or empty", "Check if the value is null or empty", "is null or empty (null)", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "is empty", "Check if the string is empty", "is empty \"\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "is not empty", "Check if the string is not empty", "is not empty \"\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "equal", "Check if 2 strings are equal", "equal \"str1\" \"str1\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string 1", "string", true));
		mql.addParam(new MQLParam("str2", "The string 2", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "not equal", "Check if 2 strings are not equal", "not equal \"str1\" \"str1\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string 1", "string", true));
		mql.addParam(new MQLParam("str2", "The string 2", "string", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "==", "Check if 2 numbers are equal", "== \"25\" \"25.0\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "!=", "Check if 2 numbers are not equal", "!= \"25\" \"25.0\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, ">", "Check if a number is greater than another", "> \"27\" \"26.0\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, ">=", "Check if a number is greater or equal than another", ">= \"26\" \"26.0\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "<", "Check if a number is less than another", "< \"25\" \"26.0\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		mql = new MQLDocumentation(true, "<=", "Check if a number is less or equal than another", "<= \"26\" \"26.0\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "The number 1", "number", true));
		mql.addParam(new MQLParam("number2", "The number 2", "number", true));
		functions.get("Operator").add(mql);
		
		functions.put("FX", new Vector<MQLDocumentation>());
		page_description.put("FX", "<img src='images/p.png' style='vertical-align: middle;'>Like all languages, you have a few data manipulation functions.");
		mql = new MQLDocumentation("Atom", true, "atom before_exclud", "get the atom list before the first occurrence of an atom in an atom list (excluded atom).<br>", "atom before_exclud \" A, B,A ,A ,C, D   \" \"3\" \",\"", " A, B", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("index", "The index atom", "Number", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom before_includ", "get the atom list before the first occurrence of an atom in an atom list (included atom).<br>", "atom before_includ \" A, B,A ,A ,C, D   \" \"3\" \",\"", " A, B,A ", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("index", "The index atom", "Number", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom count", "count the occurrence number of an atom in an atom list.<br>", "atom count \" A, B,A ,A ,C, D   \" \"A \" \",\"", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToCount", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom count_distinct", "count the occurrence number of a distinct atom in an atom list.<br>", "atom count_distinct \" A, B,A ,A ,C, D   \" \"A \" \",\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToCount", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom count_lrtrim", "count the occurrence number of a lrtrim atom in an atom list.<br>", "atom count_lrtrim \" A, B,A ,A ,C, D   \" \"A \" \",\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToCount", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom count_lrtrim_distinct", "count the occurrence number of a lrtrim distinct atom in an atom list.<br>", "atom count_lrtrim_distinct \" A, B,A ,A ,C, D   \" \"A \" \",\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToCount", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom distinct", "get distinct atom for each atom in a list<br>", "atom distinct \"A,B,A,A ,C,D\" \",\"", "A,B,A ,C,D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom distinct_lrtrim", "get lrtrim distinct atom for each atom in a list<br>", "atom distinct_lrtrim \" A, B,A ,A ,C, D   \" \",\"", "A,B,C,D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom distinct_lrtrim_1sbefore", "get lrtrim distinct atom for each atom in a list. add one space before all atoms, except the first.<br>", "atom distinct_lrtrim_1sbefore \" A, B,A ,A ,C, D   \" \",\"", "A, B, C, D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom find", "find the position of atom in an atom list.<br>", "atom find \" A, B,A ,A ,C, D   \" \"A \" \",\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToFind", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom find_lrtrim", "find the position of atom in an atom list. lrtrim on each atom before.<br>", "atom find_lrtrim \" A, B,A ,A ,C, D   \" \"A \" \",\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToFind", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom get", "get atom at the specific index in an atom list<br>", "atom get \"A,B,C,D\" 2 \",\"", "B", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("index", "Position of the atom (start to 1)", "Number", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom get_first", "get the first atom in an atom list<br>", "atom get_first \"A, B  ,C,D\" \",\"", "A", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom get_first_lrtrim", "get the first lrtrim atom in an atom list<br>", "atom get_first_lrtrim \"  A, B  ,C,D\" \",\"", "A", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom get_last", "get the last atom in an atom list<br>", "atom get_last \"A, B  ,C,D\" \",\"", "D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom get_last_lrtrim", "get the last lrtrim atom in an atom list<br>", "atom get_last_lrtrim \"  A, B  ,C,  D\" \",\"", "D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom get_lrtrim", "get lrtrim atom at the specific index in an atom list<br>", "atom get_lrtrim \"A, B  ,C,D\" 2 \",\"", "B", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("index", "Position of the atom (start to 1)", "Number", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom lrtrim", "get lrtrim atom for each atom in a list<br>", "atom lrtrim \" A, B,A ,A ,C, D   \" \",\"", "A,B,A,A,C,D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom position", "find the position of atom in an atom list.<br>", "atom position \" A, B,A ,A ,C, D   \" \"A \" \",\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToFind", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom position_lrtrim", "find the position of atom in an atom list. lrtrim on each atom before.<br>", "atom position_lrtrim \" A, B,A ,A ,C, D   \" \"A \" \",\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("atomToFind", "The atom to search", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "atom size", "get size atom of an atom list (size >=1)<br>", "atom size \"A,B,C,D\" \",\"", "4", null, null, null, null, false, "");
		mql.addParam(new MQLParam("atomList", "Atom list (example A,B,C)", "String", true));
		mql.addParam(new MQLParam("separator", "List Separator (in this example ','), 1 CHAR MAX", "Number", true));
		functions.get("FX").add(mql);

		mql = new MQLDocumentation("File", true, "file create", "Create a new text file", "file create \"data/.id\" (string generate_random_str 12;)", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("filePath", "The path to the file", "String", true));
		mql.addParam(new MQLParam("data", "String data", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "file load", "To get text from a file", "file load \"data/.id\"", "2pvrPdmfuTrz", null, null, null, null, false, "");
		mql.addParam(new MQLParam("filePath", "The path to the file", "String", true));
		functions.get("FX").add(mql);
		
		mql = new MQLDocumentation("Date", true, "date add", "add a number to a date.<br>", "date add \"1980-06-18\" \"DAY\" 1", "1980-06-19", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date addt", "add a number to a time stamp.<br>", "date addt \"1980-06-18 00:00:00\" \"DAY\" 1", "1980-06-19 00:00:00", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestamp", "The timestamp", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (SEC|MIN|HOUR|DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date current_ms", "returns the current time in milli second.<br>", "date current_ms", "1496164604097", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date current_ns", "returns the current time in nano second.<br>", "date current_ns", "823603146305304", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date curdate", "returns the current date.<br>", "date curdate", "2015-09-13", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date current_date", "returns the current date.<br>", "date current_date", "2015-09-13", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date current_time", "returns the current time.<br>", "date current_time", "14:01:22", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date current_timestamp", "returns the current timestamp.<br>", "date current_timestamp", "2015-09-13 14:01:42", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date curtime", "returns the current time.<br>", "date curtime", "14:01:57", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date curtimestamp", "returns the current timestamp.<br>", "date curtimestamp", "2015-09-13 14:02:13", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date dateadd", "add a number to a date.<br>", "date dateadd \"1980-06-18\" \"DAY\" 1", "1980-06-19", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date dateaddt", "add a number to a timestamp.<br>", "date dateaddt \"1980-06-18 00:00:00\" \"DAY\" 1", "1980-06-19 00:00:00", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestamp", "The timestamp", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (SEC|MIN|HOUR|DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date datediff", "sub a number to a date.<br>", "date datediff \"1980-06-18\" \"DAY\" 1", "1980-06-17", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date datedifft", "sub a number to a date.<br>", "date datedifft \"1980-06-18 00:00:00\" \"DAY\" 1", "1980-06-17 00:00:00", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestamp", "The timestamp", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (SEC|MIN|HOUR|DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date ts_to_long", "timestamp to long.<br>", "date ts_to_long \"1980-06-18 00:00:00\"", "330120000000", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestamp", "The timestamp", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date dt_to_long", "date to long.<br>", "date dt_to_long \"1980-06-18\"", "330120000000", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date day_of_month", "get the day of month of a date.<br>", "date day_of_month \"1980-06-18\"", "18", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date day_of_week", "get the day of week of a date.<br>", "date day_of_week \"1980-06-18\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date day_of_year", "get the day of year of a date.<br>", "date day_of_year \"1980-06-18\"", "170", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date dayname", "get the day name of a date.<br>", "date dayname \"1980-06-18\"", "mercredi", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date diff", "sub a number to a date.<br>", "date diff \"1980-06-18\" \"DAY\" 1", "1980-06-17", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date difft", "sub a number to a time stamp.<br>", "date difft \"1980-06-18 00:00:00\" \"DAY\" 1", "1980-06-17 00:00:00", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestamp", "The timestamp", "String", true));
		mql.addParam(new MQLParam("field", "The field to add (SEC|MIN|HOUR|DAY|MONTH|YEAR)", "String", true));
		mql.addParam(new MQLParam("number", "The number to add", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date full_systimestamp", "returns the current full timestamp.<br>", "date full_systimestamp ", "2015-09-13 14:03:20.395+0400", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date hour", "get the hour of a time.<br>", "date hour \"12:15:56\"", "12", null, null, null, null, false, "");
		mql.addParam(new MQLParam("time", "The time", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date is_valid_date", "check if a date is valid.<br>", "date is_valid_date \"1980-06-18\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dateToValidate", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date is_valid_date", "check if a date is valid.<br>", "date is_valid_date \"1980-06-18\" \"yyyy-MM-dd\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dateToValidate", "The date", "String", true));
		mql.addParam(new MQLParam("format", "The date format", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date is_valid_time", "check if a time is valid.<br>", "date is_valid_time \"12:00:01\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timeToValidate", "The time (12:00:01)", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date is_valid_timestamp", "check if a timestamp is valid.<br>", "date is_valid_timestamp \"1980-06-18 12:00:01\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestampToValidate", "The timestamp", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date is_valid_timestamp", "check if a timestamp is valid.<br>", "date is_valid_timestamp \"1980-06-18 12:00:01\" \"yyyy-MM-dd HH:mm:ss\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestampToValidate", "The timestamp", "String", true));
		mql.addParam(new MQLParam("format", "The format (ex: yyyy MM dd HH:mm:ss)", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date format", "format a timestamp to another format.<br>", "date format \"19800618 12:00\" \"yyyyMMdd HH:mm\" \"yyyy-MM-dd HH:mm:ss\"", "1980-06-18 12:00:01", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestampToFormat", "The timestamp", "String", true));
		mql.addParam(new MQLParam("formatIn", "The in format (ex: yyyyMMdd HH:mm)", "String", true));
		mql.addParam(new MQLParam("formatOut", "The out format (ex: yyyy MM dd HH:mm:ss)", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date minute", "get the minute of a time.<br>", "date minute \"12:15:56\"", "15", null, null, null, null, false, "");
		mql.addParam(new MQLParam("time", "The time", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date month", "get the month of a date.<br>", "date month \"1980-06-18\"", "6", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date monthname", "get the month name of a date.<br>", "date monthname \"1980-06-18\"", "juin", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date seconde", "get the seconde of a time.<br>", "date seconde \"12:15:56\"", "56", null, null, null, null, false, "");
		mql.addParam(new MQLParam("time", "The time", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date sysdate", "returns the current date.<br>", "date sysdate", "2015-09-13", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date systime", "returns the current time.<br>", "date systime", "14:04:18", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date systimestamp", "returns the current timestamp.<br>", "date systimestamp", "2015-09-13 14:04:32", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date systimestamp_min", "returns the current timestamp in minimum format.<br>", "date systimestamp_min", "20150913140446", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date time", "get the time of a timestamp.<br>", "date time \"1980-06-18 12:25:56\"", "12:25:56", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestamp", "The timestamp", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "date year", "get the year of a date.<br>", "date year \"1980-06-18\"", "1980", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);

		mql = new MQLDocumentation("Math", true, "math abs", "get the absolute value of a number<br>", "math abs -56", "56", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math acos", "get the arc cosine of a number<br>", "math acos 42", "NaN", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math asin", "get the arc sine of a number<br>", "math asin 0.5", "0.5235987755982989", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math atan", "get the arc tangent of a number<br>", "math atan 42", "1.5469913006098266", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math atan2", "get the angle theta from the conversion of rectangular coordinates (x, y) to polar coordinates (r, theta)<br>", "math atan2 4 2", "1.1071487177940904", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math avg", "get the avg of a json array of vales<br>", "math avg \"[10 20]\"", "15.0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("json_array", "A json array (of double values)", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math bit_and", "the bitwise and operation<br>", "math bit_and 4456 234", "104", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math bit_or", "the bitwise or operation<br>", "math bit_or 4456 234", "4586", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math bit_xor", "the bitwise xor operation<br>", "math bit_xor 4456 234", "4482", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math cbrt", "returns the cube root of a double value<br>", "math cbrt 42", "3.4760266448864496", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math ceil", "get the smallest integer value not less than the number specified as an argument<br>", "math ceil 1.2", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math ceiling", "get the smallest integer value not less than the number specified as an argument<br>", "math ceiling 1.2", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math cos", "get the cosine of a number<br>", "math cos 42", "-0.39998531498835127", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math cosh", "get the hyperbolic cosine of a number<br>", "math cosh 42", "8.696374707602505E17", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math cot", "get the cotangent of a number<br>", "math cot 42", "0.4364167060752729", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math decimal_format", "convert a number to another formated decimal number<br>", "math decimal_format 42.456 \"##.##\" \".\" \" \"", "42.46", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		mql.addParam(new MQLParam("pattern", "The pattern", "String", true));
		mql.addParam(new MQLParam("decimalSeparator", "The decimal separator", "String", true));
		mql.addParam(new MQLParam("groupingSeparator", "The grouping separator", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math deg_to_rad", "converts an angle measured in degrees to an approximately equivalent angle measured in radians<br>", "math deg_to_rad 42", "0.7330382858376184", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math e", "get e value<br>", "math e", "2.718281828459045", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math exp", "returns euler's number e raised to the power of a double value<br>", "math exp 42", "1.73927494152050099E18", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math expm1", "returns e^x -1<br>", "math expm1 42", "1.73927494152050099E18", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math floor", "get the largest integer value not greater than a number specified as argument<br>", "math floor 1.2", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math hypot", "returns sqrt(x^2 +y^2) without intermediate overflow or underflow<br>", "math hypot 4456 234", "4462.139845410496", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math log", "get the natural logarithm of the argument<br>", "math log 42", "3.7376696182833684", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math log10", "get the base 10 logarithm of the argument<br>", "math log10 42", "1.6232492903979006", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math log1p", "returns the natural logarithm of the sum of the argument and 1<br>", "math log1p 42", "3.7612001156935624", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math max", "returns the greater<br>", "math max 4456 234", "4456", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math min", "returns the smaller of two int values<br>", "math min 4456 234", "234", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math mod", "the modulo operation<br>", "math mod 4456 234", "10", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math pi", "get pi value<br>", "math pi", "3.141592653589793", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math pow", " get the value of a number raised to the power of another number<br>", "math pow 0.5 7", "0.0078125", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math power", " get the value of a number raised to the power of another number<br>", "math power 0.5 7", "0.0078125", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math rad_to_deg", "converts an angle measured in radians to an approximately equivalent angle measured in degrees<br>", "math rad_to_deg 42", "2406.4227395494577", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math random", "get the random number between 0 and the argument<br>", "math random 42", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math rint", "returns the double value that is closest in value to the argument and is equal to a mathematical integer<br>", "math rint 42", "42", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math round", "rounds a number specified as an argument up to a number specified as another argument<br>", "math round 1.23654 2", "1.24", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number1", "A number", "Number", true));
		mql.addParam(new MQLParam("number2", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math sign", "get the sign of a number<br>", "math sign 42", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math signum", "get the sign of a number<br>", "math signum 42", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math sin", "get the sine of a number<br>", "math sin 42", "-0.9165215479156338", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math sinh", "get the hyperbolic sine of a number<br>", "math sinh 42", "8.696374707602505E17", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math sqrt", "get the square root of a non-negative number of the argument<br>", "math sqrt 42", "6.48074069840786", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math tan", "get the tangent of a number<br>", "math tan 42", "2.2913879924374863", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math tanh", "get the hyperbolic tangent of a number<br>", "math tanh 42", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "math ulp", "returns the size of an ulp of the argument. an ulp of a double value is the positive distance between this floating-point value and the double value next larger in magnitude<br>", "math ulp 42", "7.105427357601002E-15", null, null, null, null, false, "");
		mql.addParam(new MQLParam("number", "A number", "Number", true));
		functions.get("FX").add(mql);

		mql = new MQLDocumentation("String", true, "string ascii", "convert a character to integer<br>", "string ascii \"a\"", "97", null, null, null, null, false, "");
		mql.addParam(new MQLParam("chr", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string bin", "returns a string representation of the binary value of n, where n is a long (bigint) number<br>", "string bin 97", "1100001", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string bit_length", "returns the length of the string str in bits<br>", "string bit_length 61", "16", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string char", "convert an integer to character<br>", "string char 97", "a", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string char_length", "returns the length of the string. the length is equal to the number of unicode code units in the string.<br>", "string char_length \"azerty\"", "6", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string char_to_int", "convert a character to integer<br>", "string char_to_int \"a\"", "97", null, null, null, null, false, "");
		mql.addParam(new MQLParam("chr", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "concat", "concat all parameters (do not use this function with string. before)<br>", "concat 97 \"a\" \"r\"", "97ar", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("strN", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string levenshtein_distance", "To get the levenshtein distance between two words", "string levenshtein_distance \"admn\" \"admin\"", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word1", "The word 1", "String", true));
		mql.addParam(new MQLParam("word2", "The word 2", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string count", "count the occurrence number of a string in another string<br>", "string count \"azertyaze\" \"a\"", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("string", "The string", "String", true));
		mql.addParam(new MQLParam("find", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string encode_b64", "encode a string into b64<br>", "string encode_b64 \"azerty\"", "YXplcnR5", null, null, null, null, false, "");
		mql.addParam(new MQLParam("string", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string decode_b64", "decode a string from b64<br>", "string decode_b64 \"YXplcnR5\"", "azerty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("string", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string del_char_before_each_line", "delete a number of char on each lines<br>", "string del_char_before_each_line \"sdlfkjdf<br>dfgdfgdfgdfgd\" 4", "kjdf<br>fgdfgdfgd<br>", null, null, null, null, false, "");
		mql.addParam(new MQLParam("data", "The data", "String", true));
		mql.addParam(new MQLParam("nbChar", "The number of chars", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string ends_with", "check if a value ends with another string<br>", "string ends_with \"azertyaze\" 3", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("stringValue", "The string", "String", true));
		mql.addParam(new MQLParam("stringToEnd", "The end string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string first_letter_upper", "returns the string in lower case. the first letter in upper case.<br>", "string first_letter_upper \"azerty\"", "Azerty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string first_letter", "returns the first letter from a string.<br>", "string first_letter \"azerty\"", "a", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string generate_random_str", "generate a random string<br>", "string generate_random_str 1", "D", null, null, null, null, false, "");
		mql.addParam(new MQLParam("size", "The size of the string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string hex", "returns a string representation of the hexadecimal value of n, where n is a long (bigint) number<br>", "string hex 97", "61", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string hex_to_int", "convert a hexadecimal number to (bigint) number<br>", "string hex_to_int 97", "151", null, null, null, null, false, "");
		mql.addParam(new MQLParam("hex", "The hex number", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string hex_to_str", "get a string from a hexadecimal representation (ascii).<br>", "string hex_to_str 61", "a", null, null, null, null, false, "");
		mql.addParam(new MQLParam("hex", "The hexadecimal number", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string indent", "returns the indented string<br>", "string indent \"azerty<br>a<br>b<br>c\" 3", "azerty<br>   a<br>   b<br>   c", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("nbSpaceBefore", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string instr", "returns the index within this string of the first occurrence of the specified substring<br>", "string instr \"azerty\" \"r\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string instr", "returns the index within this string of the first occurrence of the specified substring, starting at the specified index<br>", "string instr \"azerty\" \"r\" 4", "-1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		mql.addParam(new MQLParam("fromIndex", "starting to index", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string int_to_char", "convert an integer to character<br>", "string int_to_char 97", "a", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string int_to_hex", "returns a string representation of the hexadecimal value of n, where n is a long (bigint) number<br>", "string int_to_hex 97", "61", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string int_to_oct", "returns a string representation of the octal value of n, where n is a long (bigint) number<br>", "string int_to_oct 97", "141", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string is_letter", "check if a value contains only letters<br>", "string is_letter \"abcd\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string is_alpha_num_uds", "check if a value contains only letters, underscore char or numbers<br>", "string is_alpha_num_uds \"abc_12d\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string is_alpha_num", "check if a value contains only letters or numbers<br>", "string is_alpha_num \"abc12d\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string is_number_char", "check if a value contains only number char<br>", "string is_number_char \"-456.45\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string itrim", "returns the string with one space character between other chars<br>", "string itrim \"aze   rty\"", "aze rty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string lcase", "converts all of the characters in this string to lower case using the rules of the default locale<br>", "string lcase \"AZERTY\"", "azerty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string left", "get the leftmost len characters from the string str.<br>", "string left \"AZERTY\" 3", "AZE", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("len", "The length", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string length", "returns the length of this string. the length is equal to the number of unicode code units in the string.<br>", "string length \"azerty\"", "6", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string like", "performs a pattern match of a string expression expr against a pattern pat<br>", "string like \"azerty\" \".*ze.*\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "the string", "String", true));
		mql.addParam(new MQLParam("pat", "the paterne", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string locate", "returns the index within this string of the first occurrence of the specified substring<br>", "string locate \"azerty\" \"r\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string locate", "returns the index within this string of the first occurrence of the specified substring, starting at the specified index<br>", "string locate \"azerty\" \"r\" 4", "-1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		mql.addParam(new MQLParam("fromIndex", "starting to index", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string lower", "converts all of the characters in this string to lower case using the rules of the default locale<br>", "string lower \"AZERTY\"", "azerty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string lpad", "pads the left-side of a string with a specific set of characters<br>", "string lpad \"azertyaze\" \"#\" 10", "#azertyaze", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("padString", "The pad string", "String", true));
		mql.addParam(new MQLParam("paddedLength", "The end of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string lrtrim", "returns the string without space character on the left and on the right<br>", "string lrtrim \"  azerty   \"", "azerty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string lrtrim0", "returns the string without 0 character on the left and on the right<br>", "string lrtrim0 \"  123   \"", "  123   ", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string ltrim", "returns the string without space character on the left<br>", "string ltrim \"  azerty   \"", "azerty&nbsp;&nbsp;&nbsp;", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string matches", "performs a pattern match of a string expression expr against a pattern pat<br>", "string matches \"azerty\" \".*ze.*\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "the string", "String", true));
		mql.addParam(new MQLParam("pat", "the paterne", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string mid", "returns a new string that is a substring of this string. the substring begins with the character at the specified index and extends to the end of this string.<br>", "string mid \"azertyaze\" 3", "rtyaze", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("index", "The begin of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string mid", "returns a new string that is a substring of this string. the substring begins at the specified beginindex and extends to the character at index endindex - 1. thus the length of the substring is endindex-beginindex<br>", "string mid \"azertyaze\" 3 5", "rt", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("beginIndex", "The begin of the sub string", "Number", true));
		mql.addParam(new MQLParam("endIndex", "The end of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string not_like", "not performs a pattern match of a string expression expr against a pattern pat<br>", "string not_like \"azerty\" \".*ze.*\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "the string", "String", true));
		mql.addParam(new MQLParam("pat", "the paterne", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string not_regexp", "not performs a pattern match of a string expression expr against a pattern pat<br>", "string not_regexp \"azerty\" \".*ze.*\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "the string", "String", true));
		mql.addParam(new MQLParam("pat", "the paterne", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string oct", "returns a string representation of the octal value of n, where n is a long (bigint) number<br>", "string oct 97", "141", null, null, null, null, false, "");
		mql.addParam(new MQLParam("num", "The number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string oct_to_int", "convert a octal number to (bigint) number<br>", "string oct_to_int \"15\"", "13", null, null, null, null, false, "");
		mql.addParam(new MQLParam("oct", "The octal number", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string position", "returns the index within this string of the first occurrence of the specified substring<br>", "string position \"azerty\" \"r\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string position", "returns the index within this string of the first occurrence of the specified substring, starting at the specified index<br>", "string position \"azerty\" \"r\" 4", "-1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		mql.addParam(new MQLParam("fromIndex", "starting to index", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string regexp", "performs a pattern match of a string expression expr against a pattern pat<br>", "string regexp \"azerty\" \".*ze.*\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "the string", "String", true));
		mql.addParam(new MQLParam("pat", "the paterne", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string repeat", "returns a string consisting of the string str repeated count times<br>", "string repeat \"AZERTY\" 3", "AZERTYAZERTYAZERTY", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("count", "The count number", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string repeat_insert_str", "returns a string with a string inserted all n char<br>", "string repeat_insert_str \"azertyuiop\" \"-\" 3", "aze-rty-uio-p-", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("strToInsert", "The string to insert", "String", true));
		mql.addParam(new MQLParam("incr", "The increment", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string replace", "replaces each substring of this string that matches the literal target sequence with the specified literal replacement sequence<br>", "string replace \"azerty\" \"z\" 9", "a9erty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("target", "The string target", "String", true));
		mql.addParam(new MQLParam("replacement", "The replacement", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string reverse", "returns the string str with the order of the characters reversed.<br>", "string reverse \"AZERTY\"", "YTREZA", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string right", "get the rightmost len characters from the string str.<br>", "string right \"AZERTY\" 3", "RTY", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("len", "The length", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string rpad", "pads the right-side of a string with a specific set of characters<br>", "string rpad \"azertyaze\" \"#\" 10", "azertyaze#", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("padString", "The pad string", "String", true));
		mql.addParam(new MQLParam("paddedLength", "The end of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string rtrim", "returns the string without space character on the right<br>", "string rtrim \"  azerty   \"", "&nbsp;&nbsp;azerty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string space", "returns a string consisting of n space characters<br>", "string space \"5\"", "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;", null, null, null, null, false, "");
		mql.addParam(new MQLParam("count", "The number of space", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string split", "Split a string", "string split \"a b c\" \" \" -1", "[\"a\",\"b\",\"c\"]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("regex", "The regex", "String", true));
		mql.addParam(new MQLParam("limit", "The limit", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string starts_with", "check if a value starts with another string<br>", "string starts_with \"azertyaze\" 3", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("stringValue", "The string", "String", true));
		mql.addParam(new MQLParam("stringToStart", "The start string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string str_to_hex", "get a hexadecimal string representation of str where each character in str is converted to two hexadecimal digits.<br>", "string str_to_hex 97", "3937", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string strcmp", "compare two string<br>", "string strcmp \"AZERTY\" \"iop\"", "-40", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string 1", "String", true));
		mql.addParam(new MQLParam("str2", "The string 2", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string strpos", "returns the index within this string of the first occurrence of the specified substring<br>", "string strpos \"azerty\" \"r\"", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string strpos", "returns the index within this string of the first occurrence of the specified substring, starting at the specified index<br>", "string strpos \"azerty\" \"r\" 4", "-1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str1", "The string", "String", true));
		mql.addParam(new MQLParam("str2", "The string", "String", true));
		mql.addParam(new MQLParam("fromIndex", "starting to index", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string sublrchar", "delete a number of char at the start and at the end of the string<br>", "string sublrchar \"azerty\" 1", "zert", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("numberDeleteChar", "The number of char to delete", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string substr", "returns a new string that is a substring of this string. the substring begins with the character at the specified index and extends to the end of this string.<br>", "string substr \"azertyaze\" 3", "rtyaze", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("index", "The begin of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string substr", "returns a new string that is a substring of this string. the substring begins at the specified beginindex and extends to the character at index endindex - 1. thus the length of the substring is endindex-beginindex<br>", "string substr \"azertyaze\" 3 5", "rt", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("beginIndex", "The begin of the sub string", "Number", true));
		mql.addParam(new MQLParam("endIndex", "The end of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string substring", "returns a new string that is a substring of this string. the substring begins with the character at the specified index and extends to the end of this string.<br>", "string substring \"azertyaze\" 3", "rtyaze", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("index", "The begin of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string substring", "returns a new string that is a substring of this string. the substring begins at the specified beginindex and extends to the character at index endindex - 1. thus the length of the substring is endindex-beginindex<br>", "string substring \"azertyaze\" 3 5", "rt", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("beginIndex", "The begin of the sub string", "Number", true));
		mql.addParam(new MQLParam("endIndex", "The end of the sub string", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string to_string", "return a valid sql string data (example: 'i\'am a man')<br>", "string to_string \"az\\\"erty\"", "'az\"erty'", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string encode", "encode a string to another encoding format<br>", "string encode \"az\\\"erty\" \"ISO-8859-1\" \"UTF-8\"", "az\"erty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		mql.addParam(new MQLParam("sourceEnc", "The source encoding format", "String", true));
		mql.addParam(new MQLParam("destinationEnc", "The destination encoding format", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string empty_if_null", "set the string to empty if the string is null<br>", "string empty_if_null \"a b@mail.com\"", "a b@mail.com", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string null_if_empty", "set the string to null if the string is empty<br>", "string null_if_empty \"a b@mail.com\"", "a b@mail.com", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string trim", "returns the string without space character on the right, left and with one space inside<br>", "string trim \" aze   rty \"", "aze rty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string txt", "return a valid sql string data (example: 'i\'am a man')<br>", "string txt \"az'erty\"", "'az\\'erty'", null, null, null, null, false, "");
		mql.addParam(new MQLParam("data", "The data", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string txt2", "return a valid sql string data (example: 'i''am a man')<br>", "string txt2 \"az'erty\"", "'az''erty'", null, null, null, null, false, "");
		mql.addParam(new MQLParam("data", "The data", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string ucase", "converts of all the characters in this string to upper case using the rules of the default locale<br>", "string ucase \"azerty\"", "AZERTY", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string unhex", "convert a hexadecimal number to (bigint) number<br>", "string unhex 97", "151", null, null, null, null, false, "");
		mql.addParam(new MQLParam("hex", "The hex number", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string upper", "converts of all the characters in this string to upper case using the rules of the default locale<br>", "string upper \"azerty\"", "AZERTY", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "string zero", "return a 0 if empty string else return the string<br>", "string zero \"az'erty\"", "az'erty", null, null, null, null, false, "");
		mql.addParam(new MQLParam("str", "The string", "String", true));
		functions.get("FX").add(mql);

		mql = new MQLDocumentation("Type", true, "type is_bool", "check if a value is equal than two booleans<br>", "type is_bool 1 1 0", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		mql.addParam(new MQLParam("bool1", "The first boolean", "String", true));
		mql.addParam(new MQLParam("bool2", "The second boolean", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_char", "check if a value is char type (with size)<br>", "type is_char 1 1", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		mql.addParam(new MQLParam("size", "The char size authorized", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_date", "check if a value is a valid date (example: 1980-06-18) in english format<br>", "type is_date \"1980-06-18\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("date", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_decimal", "check if a value is a decimal number or not<br>", "type is_decimal 12.23 4 5", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("stringDecimal", "The string in decimal format", "String", true));
		mql.addParam(new MQLParam("digitBeforeTheDecimalPoint", "The number of digit before the decimal point", "Number", true));
		mql.addParam(new MQLParam("digitAfterTheDecimalPoint", "The number of digit after the decimal point", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_email", "check if a value is an email address<br>", "type is_email \"contact@data-prisme.com\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("emailAddress", "The email", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_hour", "check if a value is in hour format (example: 12:35:56)<br>", "type is_hour \"12:35:56\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("stringHour", "The hour", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_time", "check if a value is in hour format (example: 12:35:56)<br>", "type is_time \"12:35:56\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("stringHour", "The hour", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_hour_without_sec", "check if a value is in hour format (example: 12:35) without seconds<br>", "type is_hour_without_sec \"12:35:56\"", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("stringHour", "The hour", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_integer", "check if a value is an integer or not (with size)<br>", "type is_integer 1456 5", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		mql.addParam(new MQLParam("size", "The char size authorized", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_number", "check if a value is a number<br>", "type is_number 1456", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_byte", "check if a value is a byte<br>", "type is_byte 13", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_small_int", "check if a value is a small integer<br>", "type is_small_int 13000", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_medium_int", "check if a value is a medium integer<br>", "type is_medium_int 130000", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_int", "check if a value is an integer<br>", "type is_int 1300000", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_big_int", "check if a value is a big integer<br>", "type is_big_int 1300000", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_float", "check if a value is a float number<br>", "type is_float 1300000.4", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_double", "check if a value is a double number<br>", "type is_double 1300000.4", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_timestamp", "check if a value is a timestamp format (example: '2009-06-18 12:00:00')<br>", "type is_timestamp \"2009-06-18 12:00:00\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_valid_date", "check if a date is valid.<br>", "type is_valid_date \"1980-06-18\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dateToValidate", "The date", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_valid_date", "check if a date is valid.<br>", "type is_valid_date \"1980-06-18\" \"yyyy-MM-dd\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dateToValidate", "The date", "String", true));
		mql.addParam(new MQLParam("format", "The date format", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_valid_time", "check if a time is valid.<br>", "type is_valid_time \"12:00:01\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timeToValidate", "The time (12:00:01)", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_valid_timestamp", "check if a timestamp is valid.<br>", "type is_valid_timestamp \"1980-06-18 12:00:01\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestampToValidate", "The timestamp", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_valid_timestamp", "check if a timestamp is valid.<br>", "type is_valid_timestamp \"1980-06-18 12:00:01\" \"yyyy-MM-dd HH:mm:ss\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("timestampToValidate", "The timestamp", "String", true));
		mql.addParam(new MQLParam("format", "The format (ex: yyyy MM dd HH:mm:ss)", "String", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type is_varchar", "check if a value is varchar type (with size)<br>", "type is_varchar 1 1", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("value", "The value", "String", true));
		mql.addParam(new MQLParam("size", "The char size authorized", "Number", true));
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type data show", "To show all data type", "type data show", "[<br>  {<br>    \"condition\": \"word exist [TH1] fr;\",<br>    \"relation_id\": \"4q\"<br>  },<br>  {<br>    \"condition\": \"or (type is_bool [TH1] \\\"true\\\" \\\"false\\\";) (type is_bool [TH1] 1 0;);\",<br>    \"relation_id\": \"4p\"<br>  },<br>  {<br>    \"condition\": \"type is_number [TH1];\",<br>    \"relation_id\": \"4o\"<br>  }<br>]", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		mql = new MQLDocumentation(true, "type data reload", "To reload all data type", "type data reload", "Data type reloaded with successful", null, null, null, null, false, "");
		functions.get("FX").add(mql);
		
		functions.put("Reader", new Vector<MQLDocumentation>());
		page_description.put("Reader", "<img src='images/p.png' style='vertical-align: middle;'>Readers are the entry points of the mental process executions."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>They allow the AI to choose a reading method to understand stimulations."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A mental process can understand that it is wrong reader.");
		mql = new MQLDocumentation(true, "reader show", "To show all reader type", "reader show", "[<br>  {<br>    \"action\": \"execute \\\"reader.space_separator\\\";\",<br>    \"relation_id\": \"4r\"<br>  }<br>]", null, null, null, null, false, "");
		functions.get("Reader").add(mql);
		mql = new MQLDocumentation(true, "reader reload", "To reload all reader type", "reader reload", "Reader type reloaded with successful", null, null, null, null, false, "");
		functions.get("Reader").add(mql);
		
		functions.put("User", new Vector<MQLDocumentation>());
		page_description.put("User", "<img src='images/p.png' style='vertical-align: middle;'>A user can connect to the AI, or directly to the storage structure."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A user acts directly on the stimulation table."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A user can not be deleted. It can be disabled.");
		mql = new MQLDocumentation(true, "user create", "To create a new user", "user create \"bob\" \"pwd\"", "User bob created with successful.", "user create \"luc-yann\" \"pwd\" \"-\"", "User luc-yann created with successful.", null, null, false, "user create|insert|add");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		mql.addParam(new MQLParam("password", "The password", "string", true));
		mql.addParam(new MQLParam("separator", "The separator for word creation", "string", false));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user force create", "To create a new user (force if the group already exist)", "user force create \"bob2\" \"pwd\"", "User bob2 created with successful.", "user force create \"luc-yann2\" \"pwd\" \"-\"", "User luc-yann2 created with successful.", null, null, false, "user force create|insert|add");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		mql.addParam(new MQLParam("password", "The password", "string", true));
		mql.addParam(new MQLParam("separator", "The separator for word creation", "string", false));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user exist", "To a user already exist", "user exist \"admin\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user first tab link", "To show the first tab link", "user first tab link", "admin", null, null, null, null, false, "");
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user get tab link", "To get a specific tab link", "user get tab link 0", "admin", null, null, null, null, false, "");
		mql.addParam(new MQLParam("position", "The position", "integer>=0", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user set password", "To set a user password", "user set password \"bob\" \"pwd\"", "Password updated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		mql.addParam(new MQLParam("password", "The password", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user show", "To show all users", "user show", "[<br>  \"bob2\",<br>  \"luc-yann\",<br>  \"bob\",<br>  \"admin\",<br>  \"ai\",<br>  \"mentdb\"<br>]", null, null, null, null, false, "");
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user get thought", "To get the user's thought", "user get thought \"admin\"", "TH[25]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user stimulate", "To stimulate a user", "user stimulate \"admin\"", "User stimulated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user token", "To get a token from user", "user token \"admin\"", "lTwGn2QYskaXFF7CLsdAjSaBJdXZHMmU43hWDH91mJGg9TXsavXHBBY4nWiC6zBea3BQqG", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user from token", "To get a user from token", "user from token \"lTwGn2QYskaXFF7CLsdAjSaBJdXZHMmU43hWDH91mJGg9TXsavXHBBY4nWiC6zBea3BQqG\"", "admin", null, null, null, null, false, "");
		mql.addParam(new MQLParam("token", "The token", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user reset token", "To reset the token for a specific user", "user reset token \"admin\"", "Token generated with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user reset all tokens", "To reset token for all users", "user reset all tokens", "Tokens generated with successful.", null, null, null, null, false, "");
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user show groups", "To show all groups for a specific user", "user show groups \"admin\"", "[<br>  \"action\",<br>  \"admin\",<br>  \"all\",<br>  \"api-ai\",<br>  \"api-mql\",<br>  \"brain\",<br>  \"library\",<br>  \"process\",<br>  \"reader\",<br>  \"system\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user show scripts", "To show all granted scripts for a specific user", "user show scripts \"admin\"", "[<br>  \"action.select_reader\",<br>  \"action.basic_search\",<br>  \"action.search_strategy\",<br>  \"action.search_logical\",<br>  \"action.check_relation\",<br>  \"action.execute_relation_rs\",<br>  \"action.execute_relation_ra\",<br>  \"action.execute_relation_rl\",<br>  \"action.get_last_relation\",<br>  \"action.redirection_relation\",<br>  \"action.relation_response\",<br>  \"action.relation_redirection\",<br>  \"action.closed\",<br>  \"brain.stimulate\",<br>  \"brain.internal_stimulation\",<br>  \"process\\u003cinitialize\\u003e\",<br>  \"process\\u003cclick\\u003e\",<br>  \"reader.space_separator\",<br>  \"reset_server_id\",<br>  \"stimulation.last\",<br>  \"server.info\",<br>  \"server.version\",<br>  \"create_word_fr_en\",<br>  \"perf.symbol\",<br>  \"perf.word\",<br>  \"perf.thought\",<br>  \"perf.relation\",<br>  \"perf.node\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		mql = new MQLDocumentation(true, "user disable", "To disable a user", "user disable \"bob\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		functions.get("User").add(mql);
		
		functions.put("Group", new Vector<MQLDocumentation>());
		page_description.put("Group", "<img src='images/p.png' style='vertical-align: middle;'>A user or script can belong to groups."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>A group has rights.");
		mql = new MQLDocumentation(true, "group add", "To add a new group", "group add \"test\"", "Group added with successful.", null, null, null, null, false, "group create|insert|add");
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group exist", "To check if a group exist", "group exist \"test\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group grant user", "To grant a user to a group", "group grant user \"admin\" \"test\"", "User granted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group grant script", "To grant a script to a group", "script create \"substraction\" 0 \"[v1],[v2]\" {<br>	<br>	- [v1] [v2];<br>	<br>};<br>group grant script \"substraction\" \"test\"", "Script granted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group is granted user", "To check if a user granted to a group", "group is granted user \"admin\" \"test\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group is granted script", "To check if a script granted to a group", "group is granted script \"substraction\" \"test\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group get", "To get the group", "group get \"test\"", "{<br>  \"k\": \"test\",<br>  \"scripts\": [<br>    \"substraction\"<br>  ],<br>  \"users\": [<br>    \"admin\"<br>  ]<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group get user", "To get users from a group", "group get user \"test\"", "[<br>  \"admin\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group get script", "To get scripts from a group", "group get script \"test\"", "[<br>  \"substraction\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group show", "To show all groups", "group show", "{<br>  \"all\": \"G[all]\",<br>  \"luc-yann\": \"G[luc-yann]\",<br>  \"process\": \"G[process]\",<br>  \"test\": \"G[test]\",<br>  \"luc-yann2\": \"G[luc-yann2]\",<br>  \"reader\": \"G[reader]\",<br>  \"ai\": \"G[ai]\",<br>  \"admin\": \"G[admin]\",<br>  \"brain\": \"G[brain]\",<br>  \"api-mql\": \"G[api-mql]\",<br>  \"api-rest\": \"G[api-rest]\",<br>  \"system\": \"G[system]\",<br>  \"library\": \"G[library]\",<br>  \"bob2\": \"G[bob2]\",<br>  \"bob\": \"G[bob]\",<br>  \"action\": \"G[action]\",<br>  \"api-ai\": \"G[api-ai]\",<br>  \"perf\": \"G[perf]\",<br>  \"mentdb\": \"G[mentdb]\"<br>}", null, null, null, null, false, "");
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group ungrant user", "To ungrant a user to a group", "group ungrant user \"admin\" \"test\"", "User ungranted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("login", "The login", "string", true));
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group ungrant script", "To ungrant a script to a group", "group ungrant script \"substraction\" \"test\"", "Script ungranted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		mql = new MQLDocumentation(true, "group remove", "To remove a group", "group remove \"test\"", "Group removed with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("groupName", "The group name", "string", true));
		functions.get("Group").add(mql);
		
		functions.put("Job", new Vector<MQLDocumentation>());
		page_description.put("Job", "<img src='images/p.png' style='vertical-align: middle;'>You can run MQL scripts automatically at specific times.");
		mql = new MQLDocumentation(true, "job add", "To add a new job", "script create \"addition\" 0 \"[v1],[v2]\" {<br>	<br>	+ [v1] [v2];<br>	<br>};<br>job add \"myJobId\" \"addition\" \"0 0 * * * ?\"", "Job added with successful.", null, null, null, null, false, "job create|insert|add");
		mql.addParam(new MQLParam("jobId", "The job id", "string", true));
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("pattern", "The pattern", "string", true));
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job show", "To show all jobs", "job show", "{<br>  \"myJobId\": {<br>    \"pattern\": \"0 0 * * * ?\",<br>    \"scriptName\": \"addition\",<br>    \"id\": \"myJobId\"<br>  },<br>  \"reset_server_id\": {<br>    \"pattern\": \"0 0 0 * * ?\",<br>    \"scriptName\": \"reset_server_id\",<br>    \"id\": \"reset_server_id\"<br>  }<br>}", null, null, null, null, false, "");
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job exist", "To check if a job already exist", "job exist \"myJobId\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("jobId", "The job id", "string", true));
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job generate update", "To generate an update job update request", "job generate update \"myJobId\"", "job update \"myJobId\" \"addition\" \"0 0 * * * ?\";", null, null, null, null, false, "");
		mql.addParam(new MQLParam("jobId", "The job id", "string", true));
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job update", "To update a job", "job update \"myJobId\" \"addition\" \"0 1 * * * ?\"", "Job updated with successful.", "job show", "{<br>  \"reset_server_id\": {<br>    \"pattern\": \"0 0 0 * * ?\",<br>    \"scriptName\": \"reset_server_id\",<br>    \"id\": \"reset_server_id\"<br>  },<br>  \"myJobId\": {<br>    \"pattern\": \"0 1 * * * ?\",<br>    \"scriptName\": \"addition\",<br>    \"id\": \"myJobId\"<br>  }<br>}", null, null, false, "");
		mql.addParam(new MQLParam("jobId", "The job id", "string", true));
		mql.addParam(new MQLParam("scriptName", "The script name", "string", true));
		mql.addParam(new MQLParam("pattern", "The pattern", "string", true));
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job remove", "To remove a job", "job remove \"myJobId\"", "Job removed with successful.", "job show", "{<br>  \"reset_server_id\": {<br>    \"pattern\": \"0 0 0 * * ?\",<br>    \"scriptName\": \"reset_server_id\",<br>    \"id\": \"reset_server_id\"<br>  }<br>}", null, null, false, "");
		mql.addParam(new MQLParam("jobId", "The job id", "string", true));
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job scheduler stop", "To stop the job scheduler", "job scheduler stop", "Job scheduler stoped with successful.", null, null, null, null, false, "");
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job scheduler start", "To start the job scheduler", "job scheduler start", "Job scheduler started with successful.", null, null, null, null, false, "");
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job scheduler restart", "To restart the job scheduler", "job scheduler restart", "Job scheduler restarted with successful.", null, null, null, null, false, "");
		functions.get("Job").add(mql);
		mql = new MQLDocumentation(true, "job scheduler status", "To show the status of the job scheduler", "job scheduler status", "Running...", null, null, null, null, false, "");
		functions.get("Job").add(mql);
		
		functions.put("Sequence", new Vector<MQLDocumentation>());
		page_description.put("Sequence", "<img src='images/p.png' style='vertical-align: middle;'>You can manage sequences (base 26) as a standard database.");
		mql = new MQLDocumentation(true, "sequence show", "To show all sequences", "sequence show", "{<br>  \"relation-O\": \"-1\",<br>  \"thought\": \"89\",<br>  \"relation-I\": \"5\",<br>  \"relation-L\": \"54\",<br>  \"relation-U\": \"n\",<br>  \"relation-G\": \"w\",<br>  \"circle\": \"0\",<br>  \"relation-R\": \"3\",<br>  \"relation-A\": \"7\",<br>  \"relation-D\": \"1\",<br>  \"relation-S\": \"1\"<br>}", null, null, null, null, false, "");
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence get current", "To get the current sequence value", "sequence get current \"thought\"", "6z", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence add", "To add a new sequence", "sequence add \"test\" 0", "Sequence added with successful.", null, null, null, null, false, "sequence create|insert|add");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("value", "The value", "integer", true));
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence exist", "To check if a sequence exist", "sequence exist \"test\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence increment", "To increment a sequence value+1", "sequence increment \"test\"", "1", "sequence increment \"test\"", "2", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence update", "To update a sequence", "sequence update \"test\" 67", "Sequence updated with successful.", "sequence get current \"test\"", "67", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("value", "The value", "integer", true));
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence generate update", "To generate an update sequence update request", "sequence generate update \"test\"", "sequence update \"test\" \"67\";", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Sequence").add(mql);
		mql = new MQLDocumentation(true, "sequence remove", "To remove a sequence", "sequence remove \"test\"", "Sequence removed with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Sequence").add(mql);
		
		functions.put("JSON", new Vector<MQLDocumentation>());
		page_description.put("Node", "<img src='images/p.png' style='vertical-align: middle;'>The nodes are stored in the storage structure in JSON."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can change them at any time."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>Beware, however, of data corruptions."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You also have MQL functions to handle JSON in memory.");
		mql = new MQLDocumentation("Record", true, "node create", "To create a new empty node", "node create \"keyId\"", "Node keyId created with successful.", null, null, null, null, false, "node create|insert|add");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node exist", "To check if a node already exist", "node exist \"keyId\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node insert into object", "To insert an element into an object in a node", 
				"node insert into object \"keyId\" / a 5 NUM", "Inserted with successful.", "node insert into object \"keyId\" / tab \"[]\" ARRAY", "Inserted with successful.", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("fieldname", "The fieldname", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node show", "To show a specific node", "node show \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [],<br>  \"key\": \"keyId\"<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node show detailed", "To show a node with detailed information", "node show detailed \"keyId\"", "{<br>  \"dataFileId\": 0,<br>  \"dataFilePosition\": 1579520,<br>  \"data\": \"{\\\"a\\\":5.0,\\\"tab\\\":[],\\\"key\\\":\\\"keyId\\\"}\",<br>  \"size\": 32,<br>  \"indexFileTabPosition\": 5,<br>  \"key\": \"keyId\",<br>  \"indexFileToWrite\": 282<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node count element", "To count element in a node", "node count element \"keyId\" /", "3", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node fields", "To get all fields into an object in a node", "node fields \"keyId\" /", "[<br>  \"a\",<br>  \"tab\",<br>  \"key\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node insert into array", "To insert an element into an array in a node", "node insert into array \"keyId\" /tab test STR", "Inserted with successful.", "node insert into array \"keyId\" /tab test2 STR", "Inserted with successful.", "node show \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test2\"<br>  ],<br>  \"key\": \"keyId\"<br>}", false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node insert into array", "To insert an element into an array in a node", "node insert into array \"keyId\" /tab 0 test STR", "Inserted with successful.", "node insert into array \"keyId\" /tab test2 STR", "Inserted with successful.", "node show \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test\",<br>    \"test2\",<br>    \"test2\"<br>  ],<br>  \"key\": \"keyId\"<br>}", false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index position", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node is array", "To check if an element is an array in a node", "node is array \"keyId\" /tab", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node is object", "To check if an element is an object in a node", "node is object \"keyId\" /", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node select element", "To select an element in a node", "node select element \"keyId\" /a", "5.0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node update array", "To update an element into an array in a node", "node update array \"keyId\" /tab 1 test2 STR", "Updated with successful.", "node show \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ],<br>  \"key\": \"keyId\"<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index", "integer", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node update object", "To update an element into an object in a node", "node update object \"keyId\" / a 8 NUM", "Updated with successful.", "node show \"keyId\"", "{<br>  \"a\": 8.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ],<br>  \"key\": \"keyId\"<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("fieldname", "The fieldname", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node delete from array", "To delete an element from an array in a node", "node delete from array \"keyId\" /tab 0", "Deleted with successful.", "node show \"keyId\"", "{<br>  \"a\": 8.0,<br>  \"tab\": [<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ],<br>  \"key\": \"keyId\"<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index", "integer", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "node delete from object", "To delete an element from an object in a node", "node delete from object \"keyId\" / a", "Deleted with successful.", "node show \"keyId\"", "{<br>  \"tab\": [<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ],<br>  \"key\": \"keyId\"<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("fieldname", "The fieldname", "string", true));
		functions.get("JSON").add(mql);
		
		mql = new MQLDocumentation("In memory", true, "json load", "To load a json string", "json load \"keyId\" \"{}\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jsonString", "The JSON string", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json load select", "To load a json string and select an element", "json load select \"keyId\" \"{\\\"a\\\": 5.0, \\\"b\\\": {\\\"c\\\": 7.0}}\" /a", "5.0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jsonString", "The JSON string", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json load by ref", "To load by ref a json string from another json object", "json load by ref \"keyId\" /b \"newKeyId\"", "{<br>  \"c\": 7.0<br>}", "json update object \"newKeyId\" / c 8 NUM;<br>json select element \"newKeyId\" /c;", "8.0", "json doc \"keyId\";", "{<br>  \"a\": 5.0,<br>  \"b\": {<br>    \"c\": 8.0<br>  }<br>}", false, "");
		mql.addParam(new MQLParam("fromKey", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("new KeyId", "The JSON string", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json show key", "To show all JSON object saved into the session", "json show key", "[\"keyId\"]", null, null, null, null, false, "");
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json insert into object", "To insert an element into an object in a node", 
				"json insert into object \"keyId\" / a 5 NUM", "1", "json insert into object \"keyId\" / tab \"[]\" ARRAY", "1", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("fieldname", "The fieldname", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json count element", "To count element in a node", "json count element \"keyId\" /", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json select element", "To select an element in a node", "json select element \"keyId\" /a", "5.0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json is object", "To check if an element is an object in a node", "json is object \"keyId\" /", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json is array", "To check if an element is an array in a node", "json is array \"keyId\" /tab", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json is loaded", "To check if a key is loaded", "json is loaded \"keyId\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json doc", "To get a json document", "json doc \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": []<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json fields", "To get all fields into an object in a node", "json fields \"keyId\" /", "[<br>  \"a\",<br>  \"tab\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json insert into array", "To insert an element into an array in a node", "json insert into array \"keyId\" /tab test STR", "1", "json insert into array \"keyId\" /tab test2 STR", "1", "json doc \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test2\"<br>  ]<br>}", false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json insert into array", "To insert an element into an array in a node", "json insert into array \"keyId\" /tab 0 test STR", "1", "json insert into array \"keyId\" /tab test2 STR", "1", "json doc \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test\",<br>    \"test2\",<br>    \"test2\"<br>  ]<br>}", false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index position", "number", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json set into array", "To set an element into an array in a node", "json set into array \"keyId\" /tab 0 test STR", "1", "json insert into array \"keyId\" /tab test2 STR", "1", "json doc \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ]<br>}", false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index position", "number", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json update array", "To update an element into an array in a node", "json update array \"keyId\" /tab 1 test2 STR", "1", "json doc \"keyId\"", "{<br>  \"a\": 5.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ]<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index", "integer", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json update object", "To update an element into an object in a node", "json update object \"keyId\" / a 8 NUM", "1", "json doc \"keyId\"", "{<br>  \"a\": 8.0,<br>  \"tab\": [<br>    \"test\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ]<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("fieldname", "The fieldname", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		mql.addParam(new MQLParam("type", "The type (NUM|STR|BOOL|ARRAY|OBJ)", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json delete from array", "To delete an element from an array in a node", "json delete from array \"keyId\" /tab 0", "1", "json doc \"keyId\"", "{<br>  \"a\": 8.0,<br>  \"tab\": [<br>    \"test2\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ]<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("index", "The index", "integer", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json delete from object", "To delete an element from an object in a node", "json delete from object \"keyId\" / a", "1", "json doc \"keyId\"", "{<br>  \"tab\": [<br>    \"test2\",<br>    \"test2\",<br>    \"test2\",<br>    \"test2\"<br>  ]<br>}", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("jPath", "The JSON path", "string", true));
		mql.addParam(new MQLParam("fieldname", "The fieldname", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json unload", "To unload a json document", "json unload \"keyId\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("JSON").add(mql);
		mql = new MQLDocumentation(true, "json unload_all", "To unload all json documents", "json unload_all", "0", null, null, null, null, false, "");
		functions.get("JSON").add(mql);

		functions.put("Transaction", new Vector<MQLDocumentation>());
		page_description.put("Transaction", "<img src='images/p.png' style='vertical-align: middle;'>Whenever you run MQL commands, you are in a micro transaction."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>If one of your orders returns an error, all your micro transaction is rollbacked."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>The storage system is ACID.");
		mql = new MQLDocumentation(true, "commit", "To commit the current transaction", "word create duck en;<br>commit;<br>word create rabbit en;<br>exception (1) (\"here a rollback ...\");", "1: here a rollback ...", "node show W[duck]", "{<br>  \"fl\": {<br>    \"en\": {<br>      \"ltl\": \"TH[77]\",<br>      \"ftl\": \"TH[77]\"<br>    }<br>  },<br>  \"w\": 1,<br>  \"k\": \"W[duck]\"<br>}", "node show W[rabbit]", "Sorry, the node 'W[rabbit]' does not exist.", false, "");
		functions.get("Transaction").add(mql);
		mql = new MQLDocumentation(true, "rollback", "To rollback the current transaction", "word create wolf en;<br>commit;<br>word create panda en;<br>rollback;", "Rollback done.", "node show W[wolf]", "{<br>  \"fl\": {<br>    \"en\": {<br>      \"ltl\": \"TH[78]\",<br>      \"ftl\": \"TH[78]\"<br>    }<br>  },<br>  \"w\": 1,<br>  \"k\": \"W[wolf]\"<br>}", "node show W[panda]", "Sorry, the node 'W[panda]' does not exist.", false, "");
		functions.get("Transaction").add(mql);
		mql = new MQLDocumentation(true, "transaction logs", "To show the log block list to replay", "word create hippo en;<br>env set var \"[transaction_logs]\" (transaction logs 10;);<br>[transaction_logs]", "[<br>  \"1\\tTO_OVERWRITE_COUNTER_INDEX\\tIndexFileId:651\\tCounter:4\",<br>  \"2\\tTO_SET_OLD_INDEX_BLOCK\\tIndexFileId:651\\tIndexPosition:104\\tDataFileId:-1\\tDataPosition:-1\",<br>  \"3\\tTO_DELETE_BLOCK\\tDataFileId:0\\tPosition:1589248\",<br>  \"4\\tTO_REWRITE_UNUSED\\tDataFileId:0\\tPosition:1589248\",<br>  \"5\\tTO_DELETE_BLOCK\\tDataFileId:0\\tPosition:1589504\",<br>  \"6\\tTO_REWRITE_UNUSED\\tDataFileId:0\\tPosition:1589504\",<br>  \"7\\tTO_OVERWRITE_COUNTER_INDEX\\tIndexFileId:651\\tCounter:5\",<br>  \"8\\tTO_SET_OLD_INDEX_BLOCK\\tIndexFileId:651\\tIndexPosition:104\\tDataFileId:0\\tDataPosition:1589504\",<br>  \"9\\tTO_DELETE_UNUSED\\tDataFileId:0\\tUnusedLength:8\",<br>  \"10\\tTO_REWRITE_BLOCK\\tDataFileId:0\\tPosition:1589504\\tNextPosition:-1\\tData:W1siV1toaXBwb10iLDAsMTU4OTI0OF1d                                                                                                                                                                                                        \\tSize:32\"<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("limit", "The limit value", "integer >= 0", true));
		functions.get("Transaction").add(mql);

		functions.put("File System", new Vector<MQLDocumentation>());
		page_description.put("File System", "<img src='images/p.png' style='vertical-align: middle;'>Each of your JSON nodes are saved in blocks to form a record."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>All blocks and records are searchable through position calculations in index and data files."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>You can see the sizes of the data and index files at any time.");
		//video.put("File System", "https://www.youtube.com/embed/XGSy3_Czz8k");
		mql = new MQLDocumentation(true, "fs index size", "To show the size of all index files (by page)", "fs index size 1 10", "[<br>  {<br>    \"percent_used\": 6,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 1,<br>    \"percent_graph\": \"|-----------------------------\",<br>    \"indexFileId\": 0<br>  },<br>  {<br>    \"percent_used\": 12,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 2,<br>    \"percent_graph\": \"|||---------------------------\",<br>    \"indexFileId\": 1<br>  },<br>  {<br>    \"percent_used\": 31,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 5,<br>    \"percent_graph\": \"|||||||||---------------------\",<br>    \"indexFileId\": 2<br>  },<br>  {<br>    \"percent_used\": 25,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 4,<br>    \"percent_graph\": \"|||||||-----------------------\",<br>    \"indexFileId\": 3<br>  },<br>  {<br>    \"percent_used\": 37,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 6,<br>    \"percent_graph\": \"|||||||||||-------------------\",<br>    \"indexFileId\": 4<br>  },<br>  {<br>    \"percent_used\": 0,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 0,<br>    \"percent_graph\": \"------------------------------\",<br>    \"indexFileId\": 5<br>  },<br>  {<br>    \"percent_used\": 0,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 0,<br>    \"percent_graph\": \"------------------------------\",<br>    \"indexFileId\": 6<br>  },<br>  {<br>    \"percent_used\": 6,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 1,<br>    \"percent_graph\": \"|-----------------------------\",<br>    \"indexFileId\": 7<br>  },<br>  {<br>    \"percent_used\": 12,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 2,<br>    \"percent_graph\": \"|||---------------------------\",<br>    \"indexFileId\": 8<br>  },<br>  {<br>    \"percent_used\": 25,<br>    \"current_max_block_size\": 16,<br>    \"size\": 200,<br>    \"current_block_size\": 4,<br>    \"percent_graph\": \"|||||||-----------------------\",<br>    \"indexFileId\": 9<br>  }<br>]", "metric index files 1 10;", "Total block: 1000x16=16000<br>Total size: 1000x200=200000b<br>|----------------------------- 6% - 0 - used: 1, free: 15, total block: 16, size: 200b<br>|||--------------------------- 12% - 1 - used: 2, free: 14, total block: 16, size: 200b<br>|||||||||--------------------- 31% - 2 - used: 5, free: 11, total block: 16, size: 200b<br>|||||||----------------------- 25% - 3 - used: 4, free: 12, total block: 16, size: 200b<br>|||||||||||------------------- 37% - 4 - used: 6, free: 10, total block: 16, size: 200b<br>------------------------------ 0% - 5 - used: 0, free: 16, total block: 16, size: 200b<br>------------------------------ 0% - 6 - used: 0, free: 16, total block: 16, size: 200b<br>|----------------------------- 6% - 7 - used: 1, free: 15, total block: 16, size: 200b<br>|||--------------------------- 12% - 8 - used: 2, free: 14, total block: 16, size: 200b<br>|||||||----------------------- 25% - 9 - used: 4, free: 12, total block: 16, size: 200b<br>", null, null, false, "");
		mql.addParam(new MQLParam("page", "The page", "integer >= 1", true));
		mql.addParam(new MQLParam("nbByPage", "The number by page", "integer>0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs data size", "To show the size of all data files in a page", "fs data size 1 2", "[<br>  {<br>    \"dataFileId\": 0,<br>    \"percent_used\": 0,<br>    \"current_max_block_size\": 10000000,<br>    \"size\": 1583616,<br>    \"current_block_size\": 6186,<br>    \"nbUnusedBlock\": 22,<br>    \"percent_graph\": \"------------------------------\",<br>    \"lastUnusedBlock\": 1577216<br>  }<br>]", "metric data files 1 2;", "------------------------------ 0% - 0 - used: 6186, free: 9993814, total block: 10000000, size: 1583616b, max size: 2560Mb, last unused block: 1577216, nb unused block: 22", null, null, false, "");
		mql.addParam(new MQLParam("page", "The page", "integer >= 1", true));
		mql.addParam(new MQLParam("nbByPage", "The number by page", "integer>0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs index calcul pos", "To calculate the index position for a specific key", "fs index calcul pos \"W[cat]\"", "{<br>  \"indexPosition\": 8,<br>  \"key\": \"W[cat]\",<br>  \"indexFileId\": 115<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs index block", "To show a specific block at a position in an index file (The first 8 bytes is the number of block from the index file)", "fs index block 115 8", "{<br>  \"DataFileId\": 0,<br>  \"DataPosition\": 1575168<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("indexFileId", "The index file id", "integer >= 0", true));
		mql.addParam(new MQLParam("position", "The position", "integer >= 0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs index count", "To show the number of block from an index file", "fs index count 115", "2", null, null, null, null, false, "");
		mql.addParam(new MQLParam("indexFileId", "The index file id", "integer >= 0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs data block", "To show a specific block at a position from a data file", "#Here the collision block;<br>fs data block 0 1575168", "{<br>  \"nextBlockAddress\": -1,<br>  \"data\": \"W1siV1tjYXRdIiwwLDE1NzU0MjRdXQ\\u003d\\u003d\"<br>}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dataFileId", "The data file id", "integer >= 0", true));
		mql.addParam(new MQLParam("position", "The position", "integer >= 0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs data record", "To show all blocks in a record at the position in a data file", "#Here the collision record;<br>fs data record 0 1575168", "[<br>  {<br>    \"blockId\": 1,<br>    \"block\": {<br>      \"nextBlockAddress\": -1,<br>      \"data\": \"W1siV1tjYXRdIiwwLDE1NzU0MjRdXQ\\u003d\\u003d\"<br>    },<br>    \"position\": 1575168<br>  },<br>  \"[[\\\"W[cat]\\\",0,1575424]]\"<br>]", "#Here the data record W[cat];<br>fs data record 0 1575424", "[<br>  {<br>    \"blockId\": 1,<br>    \"block\": {<br>      \"nextBlockAddress\": -1,<br>      \"data\": \"eyJmbCI6eyJlbiI6eyJsdGwiOiJUSFs2dF0iLCJmdGwiOiJUSFs2dF0ifX0sInciOjIsImsiOiJXW2NhdF0ifQ\\u003d\\u003d\"<br>    },<br>    \"position\": 1575424<br>  },<br>  \"{\\\"fl\\\":{\\\"en\\\":{\\\"ltl\\\":\\\"TH[6t]\\\",\\\"ftl\\\":\\\"TH[6t]\\\"}},\\\"w\\\":2,\\\"k\\\":\\\"W[cat]\\\"}\"<br>]", "#Here the data node W[cat];<br>node show W[cat]", "{<br>  \"fl\": {<br>    \"en\": {<br>      \"ltl\": \"TH[72]\",<br>      \"ftl\": \"TH[6t]\"<br>    }<br>  },<br>  \"w\": 10,<br>  \"k\": \"W[cat]\"<br>}", false, "");
		mql.addParam(new MQLParam("dataFileId", "The data file id", "integer >= 0", true));
		mql.addParam(new MQLParam("position", "The position", "integer >= 0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs cache size", "To show the cache size", "fs cache size", "512", null, null, null, null, false, "");
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs unused data block", "To show the unused block list (after a block was deleted). The list return the last block before.", "fs unused data block 0 10", "[<br>  1577728,<br>  1578752,<br>  1579008,<br>  57344,<br>  1579520,<br>  1579776,<br>  1581312,<br>  1580288,<br>  1580032,<br>  1579264,<br>  1581824<br>]", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dataFileId", "The data file id", "integer >= 0", true));
		mql.addParam(new MQLParam("limit", "The limit", "integer >= 0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs unused nb block", "To show the number of unused data block (after a block was deleted)", "fs unused nb block 0", "22", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dataFileId", "The data file id", "integer >= 0", true));
		functions.get("File System").add(mql);
		mql = new MQLDocumentation(true, "fs unused last block", "To show the last unused data block (after a block was deleted)", "fs unused last block 0", "1577728", null, null, null, null, false, "");
		mql.addParam(new MQLParam("dataFileId", "The data file id", "integer >= 0", true));
		functions.get("File System").add(mql);
		
		functions.put("Metric", new Vector<MQLDocumentation>());
		page_description.put("Metric", "<img src='images/p.png' style='vertical-align: middle;'>You have a lot of metric functions.");
		mql = new MQLDocumentation(true, "metric sessions", "To show all metrics about sessions", "metric sessions", "Cmd on 31 days (total: 16): [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16]<br>Session on 31 days (total: 4): [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4]<br>Open sessions: 2/20<br>  - id: 1, user: ai, nbExecution=2, life: 142/600s, used: 142, maxUsable: 600<br>  - id: 3, user: admin, nbExecution=8, life: 0/600s, used: 0, maxUsable: 600", null, null, null, null, false, "");
		functions.get("Metric").add(mql);
		mql = new MQLDocumentation(true, "metric index files", "To show all stats about index files", "metric index files 1 2", "Total block: 1000x16=16000<br>Total size: 1000x200=200000b<br>|----------------------------- 6% - 0 - used: 1, free: 15, total block: 16, size: 200b<br>------------------------------ 0% - 1 - used: 0, free: 16, total block: 16, size: 200b", null, null, null, null, false, "");
		mql.addParam(new MQLParam("page", "The page", "integer>0", true));
		mql.addParam(new MQLParam("nbByPage", "The number by page", "integer>0", true));
		functions.get("Metric").add(mql);
		mql = new MQLDocumentation(true, "metric data files", "To show all stats about data files", "metric data files 1 2", "------------------------------ 0% - 0 - used: 3699, free: 9996301, total block: 10000000, size: 458676b, max size: 1240Mb, last unused block: -1, nb unused block: 0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("page", "The page", "integer>0", true));
		mql.addParam(new MQLParam("nbByPage", "The number by page", "integer>0", true));
		functions.get("Metric").add(mql);
		mql = new MQLDocumentation(true, "metric system", "To show all stats about system", "metric system", "AI: Liza Payet<br>ID: NNeWO1VGLmeI<br>MentDB: v_alpha-0.3.1<br>Author: Jimmitry PAYET, Programmer Analyst<br>Contact: contact@jpayet.re<br>Java: v_1.8.0_25, Oracle Corporation<br>Mac OS X: v_10.10.5, x86_64, nb_proc: 4<br>Cpu jvm: 0.0%, Cpu sys: 76.1%, Used jvm mem: 255/757/1820M, Used mem: 7490/8192M, Used swap mem: 2456/4096M<br>Cpu on 24 hours: [0.0]<br>Mem on 24 hours: [85]<br>File system roots:<br>  - [/] 81%, usedSpace: 195027M, freeSpace: 43220M, totalSpace: 238248M<br>", null, null, null, null, false, "");
		functions.get("Metric").add(mql);
		functions.get("Metric").add(new MQLDocumentation(true, "metric count cmd", "To count the number of command executed at the last 31 days", "metric count cmd", "14", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric count session", "To count the number of session opened at the last 31 days", "metric count session", "5", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric cmd", "To show the number of command executed at the last 31 days", "metric cmd", "[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,18]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric cpu", "To show all cpu values at the last 10m/24h", "metric cpu", "[[1459140133990,0.0],[1459143173857,9.6],[1459143773878,3.0],[1459144373901,3.9],[1459144973920,7.6],[1459145573941,8.4],[1459146173960,7.2],[1459146773980,5.4],[1459147374003,3.1],[1459147974024,7.0]]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current cpu jvm", "To show current jvm cpu value", "metric current cpu jvm", "0.0", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current cpu system", "To show current system cpu value", "metric current cpu system", "9.0", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current free mem", "To show current free memory value", "metric current free mem", "503", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current free swap mem", "To show current free swap memory value", "metric current free swap mem", "1093", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current mem jvm", "To show current jvm memory status", "metric current mem jvm", "32/123/1820", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current used mem", "To show current usable memory value", "metric current used mem", "7690", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric current used swap mem", "To show current used swap memory value", "metric current used swap mem", "1978", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric date", "To show all dates at the last 31 days", "metric date", "[2016-02-27, 2016-02-28, 2016-02-29, 2016-03-01, 2016-03-02, 2016-03-03, 2016-03-04, 2016-03-05, 2016-03-06, 2016-03-07, 2016-03-08, 2016-03-09, 2016-03-10, 2016-03-11, 2016-03-12, 2016-03-13, 2016-03-14, 2016-03-15, 2016-03-16, 2016-03-17, 2016-03-18, 2016-03-19, 2016-03-20, 2016-03-21, 2016-03-22, 2016-03-23, 2016-03-24, 2016-03-25, 2016-03-26, 2016-03-27, 2016-03-28]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric file system roots", "To show statistics to all system roots", "metric file system roots", "[{\"usedSpace\":\"148708\",\"freeSpace\":\"89539\",\"absolutePath\":\"\\/\",\"totalSpace\":\"238248\"}]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric java vendor", "To show the Java vendor", "metric java vendor", "Oracle Corporation", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric java version", "To show the Java version", "metric java version", "1.8.0_25", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric mem", "To show all mem values at the last 10m/24h", "metric mem", "[[1459140133991,81],[1459143173857,84],[1459143773878,84],[1459144373901,88],[1459144973920,89],[1459145573941,89],[1459146173960,89],[1459146773980,87],[1459147374003,87],[1459147974024,87],[1459148574049,93]]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric sess", "To show the number of session opened at the last 31 days", "metric sess", "[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric session", "To show the number of session opened at the last 31 days", "metric session", "[[1456516800000,0],[1456603200000,0],[1456689600000,0],[1456776000000,0],[1456862400000,0],[1456948800000,0],[1457035200000,0],[1457121600000,0],[1457208000000,0],[1457294400000,0],[1457380800000,0],[1457467200000,0],[1457553600000,0],[1457640000000,0],[1457726400000,0],[1457812800000,0],[1457899200000,0],[1457985600000,0],[1458072000000,0],[1458158400000,0],[1458244800000,0],[1458331200000,0],[1458417600000,0],[1458504000000,0],[1458590400000,0],[1458676800000,0],[1458763200000,0],[1458849600000,0],[1458936000000,0],[1459022400000,0],[1459108800000,5]]", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric system architecture", "To show the architecture of the operating system", "metric system architecture", "x86_64", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric system name", "To show the name of the operating system", "metric system name", "Mac OS X", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric system nb processor", "To show the processor number of the operating system", "metric system nb processor", "4", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric system version", "To show the version of the operating system", "metric system version", "10.10.5", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric total mem", "To show current total memory value", "metric total mem", "8192", null, null, null, null, false, ""));
		functions.get("Metric").add(new MQLDocumentation(true, "metric total swap mem", "To show current total swap memory value", "metric total swap mem", "3072", null, null, null, null, false, ""));

		functions.put("Environment variable", new Vector<MQLDocumentation>());
		page_description.put("Environment variable", "<img src='images/p.png' style='vertical-align: middle;'>You can save data into variables in your session.");
		mql = new MQLDocumentation(true, "env set var", "To set a variable into the environment", "env set var \"[var1]\" 15", "15", "-> \"[var2]\" 16", "16", null, null, false, "");
		mql.addParam(new MQLParam("varName", "The variable name", "string", true));
		mql.addParam(new MQLParam("value", "The value", "string", true));
		functions.get("Environment variable").add(mql);
		mql = new MQLDocumentation(true, "env incr var", "To increment a variable into the environment", "env incr var \"[var1]\" 1", "16", "++ \"[var1]\"", "17", null, null, false, "");
		mql.addParam(new MQLParam("varName", "The variable name", "string", true));
		mql.addParam(new MQLParam("increment", "The increment", "integer", true));
		functions.get("Environment variable").add(mql);
		mql = new MQLDocumentation(true, "env decr var", "To decrement a variable into the environment", "env decr var \"[var1]\" 1", "16", "-- \"[var1]\"", "15", null, null, false, "");
		mql.addParam(new MQLParam("varName", "The variable name", "string", true));
		mql.addParam(new MQLParam("decrement", "The decrement", "integer", true));
		functions.get("Environment variable").add(mql);
		mql = new MQLDocumentation(true, "env get var", "To get a variable from into the environment", "env get var \"[var1]\"", "15", null, null, null, null, false, "");
		mql.addParam(new MQLParam("varName", "The variable name", "string", true));
		functions.get("Environment variable").add(mql);
		mql = new MQLDocumentation(true, "env exist var", "To check if a variable exist into the environment", "env exist var \"[var1]\"", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("varName", "The variable name", "string", true));
		functions.get("Environment variable").add(mql);
		mql = new MQLDocumentation(true, "env del var", "To delete a variable from the environment", "env del var \"[var1]\"", "Variable deleted with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("varName", "The variable name", "string", true));
		functions.get("Environment variable").add(mql);
		mql = new MQLDocumentation(true, "env show", "To show the environment variable", "env show", "{<br>  \"[err]\": \"1: your messffage ...\",<br>  \"[var2]\": \"16\",<br>  \"[i]\": \"45\",<br>  \"[ERROR]\": \"1: your messffage ...\"<br>}", null, null, null, null, false, "");
		functions.get("Environment variable").add(mql);

		functions.put("Session", new Vector<MQLDocumentation>());
		page_description.put("Session", "<img src='images/p.png' style='vertical-align: middle;'>You have control functions of web socket session.<br>"+
				"<img src='images/p.png' style='vertical-align: middle;'>You have control functions of database session.");
		functions.get("Session").add(new MQLDocumentation("WS Session", true, "chat", "To change the mode to CHAT.", "@chat", "CHAT mode activated (do not use directly with the editor, only in your application with the web socket driver).", null, null, null, null, true, ""));
		functions.get("Session").add(new MQLDocumentation(true, "cmdid", "To show the number of web socket command executed by the current user", "@cmdid", "5", null, null, null, null, true, ""));
		functions.get("Session").add(new MQLDocumentation(true, "count sessions", "To count all open web socket sessions", "@count sessions", "1", null, null, null, null, true, ""));
		functions.get("Session").add(new MQLDocumentation(true, "sessions", "To show open web socket sessions", "@sessions", "[<br>  {<br>    \"maxUsable\": 600,<br>    \"nbExecution\": 5,<br>    \"id\": 2,<br>    \"used\": 0,<br>    \"user\": \"admin\",<br>    \"life\": \"0/600s\"<br>  }<br>]", null, null, null, null, true, ""));
		functions.get("Session").add(new MQLDocumentation(true, "sid", "To show current web socket session id", "@sid", "3", null, null, null, null, true, ""));
		
		functions.get("Session").add(new MQLDocumentation("DB Session", true, "bye", "To close the session (or exit or quit)", "bye", "Session close with successful.", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "cmdid", "To show the number of command executed by the current user", "cmdid", "25", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "count sessions", "To count all open sessions", "count sessions", "3", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "exit", "To close the session (or bye or quit)", "exit", "Session close with successful.", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "help", "To show this help menu", null, null, null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "quit", "To close the session (or exit or quit)", "quit", "Session close with successful.", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "sessions", "To show all open sessions", "sessions", "[<br>  {<br>    \"maxUsable\": 600,<br>    \"nbExecution\": 2,<br>    \"used\": 0,<br>    \"id\": 4,<br>    \"user\": \"admin\",<br>    \"life\": \"0/600s\"<br>  }<br>]", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "sid", "To show current session id", "sid", "5", null, null, null, null, false, ""));
		functions.get("Session").add(new MQLDocumentation(true, "who", "To show the current user", "who", "admin", null, null, null, null, false, ""));

		functions.put("Parameter", new Vector<MQLDocumentation>());
		page_description.put("Parameter", "<img src='images/p.png' style='vertical-align: middle;'>You have an ultra-fast global parameter engine.");
		mql = new MQLDocumentation(true, "parameter add", "To add a new parameter", "parameter add test value 0", "Parameter added with successful.", null, null, null, null, false, "sequence create|insert|add");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("value", "The value", "integer", true));
		mql.addParam(new MQLParam("looked", "The field is looked", "boolean (1 or 0)", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter exist", "To check if a parameter exist", "parameter exist test", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter get value", "To get the parameter value", "parameter get value test", "value", "get_param test", "value", null, null, false, "get_param");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter get looked", "To get the parameter looked value", "parameter get looked test", "0", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter look", "To look a parameter", "parameter look test", "Parameter looked with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter remove", "To remove a parameter", "parameter remove test", "Sorry, the parameter test is looked.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter show", "To show all parameters", "parameter show", "{<br>  \"C[basic-search]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"10\"<br>  },<br>  \"C[symbol]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"200\"<br>  },<br>  \"test\": {<br>    \"looked\": \"1\",<br>    \"value\": \"value\"<br>  },<br>  \"C[relation]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"200\"<br>  },<br>  \"C[language]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"25\"<br>  },<br>  \"C[user]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"25\"<br>  },<br>  \"C[emotion]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"25\"<br>  },<br>  \"C[stimulation-tab]\": {<br>    \"looked\": \"1\",<br>    \"value\": \"25\"<br>  }<br>}", null, null, null, null, false, "");
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter unlook", "To unlook a parameter", "parameter unlook test", "Parameter unlooked with successful.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter update", "To update a parameter", "parameter update test 67", "Parameter updated with successful.", "parameter get value test", "67", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		mql.addParam(new MQLParam("value", "The value", "integer", true));
		functions.get("Parameter").add(mql);
		mql = new MQLDocumentation(true, "parameter generate update", "To generate an update parameter update request", "parameter generate update test", "parameter update \"test\" \"67\";", "parameter remove test", "Parameter removed with successful.", null, null, false, "");
		mql.addParam(new MQLParam("key", "The key", "string", true));
		functions.get("Parameter").add(mql);
		
		functions.put("Configuration", new Vector<MQLDocumentation>());
		page_description.put("Configuration", "<img src='images/p.png' style='vertical-align: middle;'>You have some read-only access functions on the configuration.");
		functions.get("Configuration").add(new MQLDocumentation(true, "config max session", "To show the max session", "config max session", "20", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config remote access", "To show the remote access mode", "config remote access", "DENIED", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config server", "To show the server Server", "config server", "[ADMIN]<br>PWD=pwd<br><br>[AI]<br>FIRSTNAME=liza<br>LASTNAME=payet<br><br>[WEBSERVER]<br>#Web server port [integer]<br>WEB_SERVER_PORT=9999<br><br>[DATABASE]<br>#Server port [integer]<br>SERVER_PORT=9998<br>#The remote access [ALLOW|DENIED]<br>REMOTE_ACCESS=DENIED<br>#The session connect timeout [long]<br>SESSION_CONNECT_TIMEOUT=10000<br>#The session timeout [long]<br>SESSION_TIMEOUT=3600000<br>#Max session open in the same time [integer]<br>MAX_SESSION=20<br>#Number of index file (do not change when data was created) integer]<br>NB_INDEX_FILE=1000<br>#Number of index file open [integer]<br>NB_INDEX_FILE_OPEN=1000<br>#Number of data file open [integer]<br>NB_DATA_FILE_OPEN=20<br>#Data block size (do not change when data was created) + 24 < 65535 [integer]<br>DATA_BLOCK_SIZE=1000<br>#Max data block size [long]<br>MAX_DATA_BLOCK_SIZE=10000000<br>#RECORD CACHE SIZE [long]<br>RECORD_CACHE_SIZE=5000<br>#First language<br>FIRST_LANGUAGE=fr<br><br>[DOCUMENTATION]<br>#do not use<br>check_documentation=1<br>", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config server port", "To show the server port", "config server port", "9998", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config session connect timeout", "To show the session connect timeout", "config session connect timeout", "10000", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config session timeout", "To show the session timeout", "config session timeout", "3600000", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config web server port", "To show the web server port", "config web server port", "9999", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config first language", "To show the first language", "config first language", "fr", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config data block size", "To show the data block size (do not change when data was created) + 14 < 65535", "config data block size", "232", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config max data block size", "To show the max data block size", "config max data block size", "10000000", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config nb data file open", "To show the number of data file open", "config nb data file open", "20", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config nb index file", "To show the number of index file (do not change when data was created)", "config nb index file", "1000", null, null, null, null, false, ""));
		functions.get("Configuration").add(new MQLDocumentation(true, "config nb index file open", "To show the number of index file open", "config nb index file open", "1000", null, null, null, null, false, ""));
		
		functions.put("Log", new Vector<MQLDocumentation>());
		page_description.put("Log", "<img src='images/p.png' style='vertical-align: middle;'>You have a log engine in rolling.");
		mql = new MQLDocumentation(true, "log trace", "To write a message into the log system", "log trace test", "1", null, null, null, null, false, "");
		mql.addParam(new MQLParam("text", "The text", "string", true));
		functions.get("Log").add(mql);
		mql = new MQLDocumentation(true, "log show", "To show last write text into the log system", "log show 5", "2017-06-05 10:35:42.375+0400: 43<br>2017-06-05 10:35:42.375+0400: 44<br>2017-06-05 10:35:42.380+0400: 1: your messffage ...<br>2017-06-05 10:35:42.896+0400: test<br>", "log show", "2017-06-05 10:35:42.375+0400: 43<br>2017-06-05 10:35:42.375+0400: 44<br>2017-06-05 10:35:42.380+0400: 1: your messffage ...<br>2017-06-05 10:35:42.896+0400: test<br>", null, null, false, "");
		mql.addParam(new MQLParam("nbLine", "The number of line", "integer", false));
		functions.get("Log").add(mql);
		mql = new MQLDocumentation(true, "log current id", "To show the current log id", "log current id", "1", null, null, null, null, false, "");
		functions.get("Log").add(mql);

		functions.put("Graph", new Vector<MQLDocumentation>());
		page_description.put("Graph", "<img src='images/p.png' style='vertical-align: middle;'>You have a graph function to view the storage structure.");
		mql = new MQLDocumentation(true, "graph symbol", "To refresh graph symbols in the editor", "graph symbol \"ai\"", "j23i88m90m76i39t04r09y35p14a96y09e57t37{\"e\":[{\"n1\":\"S0_a\",\"n2\":\"L1_io\",\"k\":\"0_1\"},{\"n1\":\"S0_a\",\"n2\":\"L2_en\",\"k\":\"0_2\"},{\"n1\":\"L2_en\",\"n2\":\"W3_snake\",\"k\":\"2_3\"},{\"n1\":\"W3_snake\",\"n2\":\"W4_cat\",\"k\":\"3_4\"},{\"n1\":\"S0_a\",\"n2\":\"L5_fr\",\"k\":\"0_5\"},{\"n1\":\"L5_fr\",\"n2\":\"W6_admin\",\"k\":\"5_6\"},{\"n1\":\"W6_admin\",\"n2\":\"W7_ai\",\"k\":\"6_7\"},{\"n1\":\"W7_ai\",\"n2\":\"W8_orange\",\"k\":\"7_8\"},{\"n1\":\"W8_orange\",\"n2\":\"W9_jaune\",\"k\":\"8_9\"},{\"n1\":\"W9_jaune\",\"n2\":\"W10_blanc\",\"k\":\"9_10\"},{\"n1\":\"W10_blanc\",\"n2\":\"W11_espace\",\"k\":\"10_11\"},{\"n1\":\"W11_espace\",\"n2\":\"W12_par\",\"k\":\"11_12\"},{\"n1\":\"W12_par\",\"n2\":\"W13_séparé\",\"k\":\"12_13\"},{\"n1\":\"W13_séparé\",\"n2\":\"W14_phrase\",\"k\":\"13_14\"},{\"n1\":\"W14_phrase\",\"n2\":\"W15_ajout\",\"k\":\"14_15\"},{\"n1\":\"S0_a\",\"n2\":\"S16_i\",\"k\":\"#16_0\"},{\"n1\":\"S16_i\",\"n2\":\"L17_io\",\"k\":\"16_17\"},{\"n1\":\"S16_i\",\"n2\":\"L18_en\",\"k\":\"16_18\"},{\"n1\":\"S16_i\",\"n2\":\"L19_fr\",\"k\":\"16_19\"},{\"n1\":\"L19_fr\",\"n2\":\"W20_io\",\"k\":\"19_20\"},{\"n1\":\"W20_io\",\"n2\":\"W21_admin\",\"k\":\"20_21\"},{\"n1\":\"W21_admin\",\"n2\":\"W22_ai\",\"k\":\"21_22\"},{\"n1\":\"W22_ai\",\"n2\":\"W23_noire\",\"k\":\"22_23\"},{\"n1\":\"W23_noire\",\"n2\":\"W24_technique\",\"k\":\"23_24\"},{\"n1\":\"W24_technique\",\"n2\":\"W25_divise\",\"k\":\"24_25\"},{\"n1\":\"W25_divise\",\"n2\":\"W26_divise\",\"k\":\"25_26\"},{\"n1\":\"W26_divise\",\"n2\":\"W27_multiplie\",\"k\":\"26_27\"},{\"n1\":\"W27_multiplie\",\"n2\":\"W28_multiplie\",\"k\":\"27_28\"},{\"n1\":\"W28_multiplie\",\"n2\":\"W29_supprime\",\"k\":\"28_29\"}],\"n\":{\"L1_io\":\"L[io]\",\"L18_en\":\"L[en]\",\"W26_divise\":\"divise\",\"L19_fr\":\"L[fr]\",\"S16_i\":\"S[i]\",\"W25_divise\":\"divise\",\"W12_par\":\"par\",\"W21_admin\":\"admin\",\"W14_phrase\":\"phrase\",\"W27_multiplie\":\"multiplie\",\"W29_supprime\":\"supprime\",\"W13_séparé\":\"séparé\",\"W20_io\":\"io\",\"W22_ai\":\"ai\",\"W24_technique\":\"technique\",\"W7_ai\":\"ai\",\"L2_en\":\"L[en]\",\"L5_fr\":\"L[fr]\",\"W4_cat\":\"cat\",\"W15_ajout\":\"ajout\",\"W23_noire\":\"noire\",\"L17_io\":\"L[io]\",\"W9_jaune\":\"jaune\",\"W28_multiplie\":\"multiplie\",\"W11_espace\":\"espace\",\"W3_snake\":\"snake\",\"W6_admin\":\"admin\",\"S0_a\":\"S[a]\",\"W10_blanc\":\"blanc\",\"W8_orange\":\"orange\"}}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		functions.get("Graph").add(mql);
		mql = new MQLDocumentation(true, "graph thought", "To refresh graph thought in a word in the editor", "graph thought \"ai\"", "j23i88m90m76i39t04r09y35p14a96y09e57t38{\"e\":[{\"n1\":\"X0_ai\",\"n2\":\"L1_fr\",\"k\":\"#0_1\"},{\"n1\":\"L1_fr\",\"n2\":\"T2_TH[6l]\",\"k\":\"#1_2\"}],\"n\":{\"X0_ai\":\"ai\",\"L1_fr\":\"L[fr]\",\"T2_TH[6l]\":\"TH[6l]\"}}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("word", "The word", "string", true));
		functions.get("Graph").add(mql);
		mql = new MQLDocumentation(true, "graph relation", "To refresh graph relation in a relation/thought in the editor", "graph relation \"TH[67]\"", "j23i88m90m76i39t04r09y35p14a96y09e57t39{\"e\":[{\"n1\":\"Y0_TH[67]\",\"n2\":\"B1_L\",\"k\":\"Y0_TH[67]_B1_L\"},{\"n1\":\"B1_L\",\"n2\":\"R2_RL[4b]\",\"k\":\"B1_L_R2_RL[4b]\"}],\"n\":{\"R2_RL[4b]\":\"RL[4b]\",\"B1_L\":\"L\",\"Y0_TH[67]\":\"TH[67]\"}}", null, null, null, null, false, "");
		mql.addParam(new MQLParam("thoughtId_or_relationId", "The thought/relation id", "string", true));
		functions.get("Graph").add(mql);
		mql = new MQLDocumentation(true, "graph refresh", "To refresh the last graph in the editor", "graph refresh", "j23i88m90m76i39t04r09y35p14a96y09e57t39{\"e\":[{\"n1\":\"Y0_TH[67]\",\"n2\":\"B1_L\",\"k\":\"Y0_TH[67]_B1_L\"},{\"n1\":\"B1_L\",\"n2\":\"R2_RL[4b]\",\"k\":\"B1_L_R2_RL[4b]\"}],\"n\":{\"R2_RL[4b]\":\"RL[4b]\",\"B1_L\":\"L\",\"Y0_TH[67]\":\"TH[67]\"}}", null, null, null, null, false, "");
		functions.get("Graph").add(mql);
		
		functions.put("Server", new Vector<MQLDocumentation>());
		page_description.put("Server", "<img src='images/p.png' style='vertical-align: middle;'>You have some functions to control the execution of the server."+
				"<br><img src='images/p.png' style='vertical-align: middle;'>To learn more about the version of the server, and see the AI ID card.");
		functions.get("Server").add(new MQLDocumentation(true, "id", "To get the id of the server", "id", "WLMJ4JpIGAYY", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "ai firstname", "To show the AI firstname", "ai firstname", "liza", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "ai lastname", "To show the AI lastname", "ai lastname", "payet", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "name", "To show the AI name", "name", "liza payet", null, null, null, null, false, "ai name"));
		mql = new MQLDocumentation(true, "shutdown", "To stop the server (or stop)", "shutdown \"ZVQNt8PJDt1h_\"", "MentDB: Shutdown with successful.<br>Bye.<br><br>Disconnected.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("identificationCode", "The identification code", "string", true));
		functions.get("Server").add(mql);
		mql = new MQLDocumentation(true, "stop", "To stop the server (or shutdown)", "stop \"ZVQNt8PJDt1h_\"", "MentDB: Shutdown with successful.<br>Bye.<br><br>Disconnected.", null, null, null, null, false, "");
		mql.addParam(new MQLParam("identificationCode", "The identification code", "string", true));
		functions.get("Server").add(mql);
		functions.get("Server").add(new MQLDocumentation(true, "version", "To show the version of the server", "version", Start.version, null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "exceeded sessions", "To show the number of exceeded sessions", "exceeded sessions", "0", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "reset exceeded sessions", "To reset the number of exceeded sessions", "reset exceeded sessions", "Exceeded sessions has been reset successfully.", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "exceeded sessions", "To show the number of exceeded web socket sessions", "@exceeded sessions", "0", null, null, null, null, true, ""));
		functions.get("Server").add(new MQLDocumentation(true, "reset exceeded sessions", "To reset the number of exceeded web socket sessions", "@reset exceeded sessions", "Exceeded sessions has been reset successfully.", null, null, null, null, true, ""));
		functions.get("Server").add(new MQLDocumentation(true, "src count", "To count the number of Java lines", "src count", "35679", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "refresh tree", "To refresh the tree in the editor", "refresh tree", "", null, null, null, null, false, ""));
		functions.get("Server").add(new MQLDocumentation(true, "in editor", "To refresh the tree in the editor", "in editor {script generate update \"brain.stimulate\"}", "j23i88m90m76i39t04r09y35p14a96y09e57t36script update \"brain.stimulate\" 1 \"[str],[level]\" {<br><br>	env set var \"[fromUser]\" (who;);<br><br>	if (and (type is_int [level];) (>= [level] 0);) {<br><br>		#Initialize the stimulation;<br>		stimulation initialize [str];<br><br>		#Get the last simulation;<br>		include \"stimulation.last\" \"[fromUser]\" [fromUser];<br><br>		env set var \"[iStim]\" 0;<br>		while (and <br>				(< [iStim] [level];) <br>				(not equal (json select element lastStimulation \"/_current_action_\";) \"CLOSED\";);) {<br><br>			#Make a stimulation;<br>			stimulation click;<br><br>			#Get the last simulation;<br>			include \"stimulation.last\" \"[fromUser]\" [fromUser];<br><br>			env incr var \"[iStim]\" 1;<br>		<br>		};<br>		<br>		stimulation get [fromUser] (- (stimulation size [fromUser];) 1);<br>	<br>	} {<br>	<br>		#Check the stop level;<br>		if (case<br>			(equal [level] \"INITIALIZATION\";) {true;}<br>			(equal [level] \"BASIC_SEARCH\";) {true;}<br>			(equal [level] \"SEARCH_STRATEGY\";) {true;}<br>			(equal [level] \"SEARCH_LOGICAL\";) {true;}<br>			(equal [level] \"CHECK_RELATION\";) {true;}<br>			(equal [level] \"GET_LAST_RELATION\";) {true;}<br>			(equal [level] \"EXECUTE_RELATION_RS\";) {true;}<br>			(equal [level] \"EXECUTE_RELATION_RA\";) {true;}<br>			(equal [level] \"EXECUTE_RELATION_RL\";) {true;}<br>			(equal [level] \"REDIRECTION_RELATION\";) {true;}<br>			(equal [level] \"RELATION_RESPONSE\";) {true;}<br>			(equal [level] \"RELATION_REDIRECTION\";) {true;}<br>			(equal [level] \"CLOSED\";) {true;}<br>			{false};) {<br>	<br>			#The stop level is valid;<br>	<br>			#Initialize the stimulation;<br>			stimulation initialize [str];<br>	<br>			#Get the last simulation;<br>			include \"stimulation.last\" \"[fromUser]\" [fromUser];<br>	<br>			while (and <br>					(not equal (json select element lastStimulation \"/_current_action_\";) [level];) <br>					(not equal (json select element lastStimulation \"/_current_action_\";) \"CLOSED\";);) {<br>	<br>				#Make a stimulation;<br>				stimulation click;<br>	<br>				#Get the last simulation;<br>				include \"stimulation.last\" \"[fromUser]\" [fromUser];<br>			<br>			};<br>	<br>			stimulation get [fromUser] (- (stimulation size [fromUser];) 1);<br>		<br>		} {<br>	<br>			#The stop level is not valid;<br>			exception (1) (\"The stop level is not valid (INITIALIZATION | BASIC_SEARCH | SEARCH_STRATEGY | SEARCH_LOGICAL | CHECK_RELATION | GET_LAST_RELATION | EXECUTE_RELATION_RS | EXECUTE_RELATION_RA | EXECUTE_RELATION_RL | REDIRECTION_RELATION | RELATION_RESPONSE | RELATION_REDIRECTION | CLOSED).\");<br>		<br>		};<br>	<br>	};<br>	<br>};", null, null, null, null, false, ""));
		
		if ((new File(".."+File.separator+"MentDB_Editor"+File.separator+"tools"+File.separator+"mql.json")).exists()) {
		
			//Write the MQL file
			try (Writer writer = new BufferedWriter(new OutputStreamWriter(
		              new FileOutputStream(".."+File.separator+"MentDB_Editor"+File.separator+"tools"+File.separator+"mql.json"), "utf-8"))) {
				
				JSONArray json = new JSONArray();
	
				for (Entry<String, Vector<MQLDocumentation>> e : MQLDocumentation.functions.entrySet()) {
					
					for(int i=0;i<e.getValue().size();i++) {
						
						String fx = "", fxAll = "", fxDesc = "";
											
						//Get the current MQLDocumentation object
						
						MQLDocumentation currentFunction = e.getValue().get(i);
						
						fx+= (currentFunction.webSocket?"@":"")+currentFunction.functionId;
						fxDesc += "Description: "+currentFunction.description+"<hr>";
						if (currentFunction.synonymous!= null && !currentFunction.synonymous.equals("")) fxDesc += "Synonymous: "+currentFunction.synonymous+"<hr>";
						for(int r=0;r<currentFunction.parameters.size();r++) {
							
							fxAll += " \"<b>"+currentFunction.parameters.get(r).name+"</b>\"";
							fxDesc += "<b>"+currentFunction.parameters.get(r).name+"</b>: "+currentFunction.parameters.get(r).description+" - "+currentFunction.parameters.get(r).type+" - required="+currentFunction.parameters.get(r).required+"<br>";
	
						}
						if (currentFunction.parameters.size()>0) fxDesc += "<hr>";
						if (currentFunction.example1!=null) fxDesc += "Example 1<br><pre style='color:#2A7CD9'>"+currentFunction.example1.replace("<br>", "\n")+"</pre>";
						else fxDesc += "Example 1<br>?????<br>";
						fxDesc += "Result 1<br><pre style='color:#2AB77C'>"+currentFunction.result1+"</pre>";
						if (currentFunction.example2!= null && !currentFunction.example2.equals("")) fxDesc += "Example 2<br><pre style='color:#2A7CD9'>"+currentFunction.example2+"</pre>";
						if (currentFunction.result2!= null && !currentFunction.result2.equals("")) fxDesc += "Result 2<br><pre style='color:#2AB77C'>"+currentFunction.result2+"</pre><br>";
						if (currentFunction.example3!= null && !currentFunction.example3.equals("")) fxDesc += "Example 3<br><pre style='color:#2A7CD9'>"+currentFunction.example3+"</pre>";
						if (currentFunction.result3!= null && !currentFunction.result3.equals("")) fxDesc += "Result 3<br><pre style='color:#2AB77C'>"+currentFunction.result3+"</pre>";
						
						json.add(fx);
						if (currentFunction.example1!=null) json.add(currentFunction.example1.replace("<br>", "\n"));
						else json.add(fx+fxAll);
						json.add("<div style='white-space: nowrap;'><span style='color:#995500'><b>"+fx+"</b></span>"+fxAll+"<hr>"+fxDesc+"</div>");
						json.add(e.getKey());
						
						if (!currentFunction.synonymous.equals("")) {
							
							fx = "";
							fxAll = "";
							fxDesc = "";
							
							//Get the current synonymous MQLDocumentation object
							
							fx+= (currentFunction.webSocket?"@":"")+currentFunction.synonymous;
							fxDesc += "Description: "+currentFunction.description+"<hr>";
							if (currentFunction.synonymous!= null && !currentFunction.synonymous.equals("")) fxDesc += "Synonymous: "+currentFunction.synonymous+"<hr>";
							
							for(int r=0;r<currentFunction.parameters.size();r++) {
								
								fxAll += " \"<b>"+currentFunction.parameters.get(r).name+"</b>\"";
								fxDesc += "<b>"+currentFunction.parameters.get(r).name+"</b>: "+currentFunction.parameters.get(r).description+" - "+currentFunction.parameters.get(r).type+" - required="+currentFunction.parameters.get(r).required+"<br>";
	
							}
							if (currentFunction.parameters.size()>0) fxDesc += "<hr>";
							if (currentFunction.example1!=null) fxDesc += "Example 1<br>"+currentFunction.example1.replace("<br>", "\n")+"<br>";
							else fxDesc += "Example 1<br>?????<br>";
							fxDesc += "Result 1<br>"+currentFunction.result1+"<br><br>";
							if (currentFunction.example2!= null && !currentFunction.example2.equals("")) fxDesc += "Example 2<br>"+currentFunction.example2+"<br>";
							if (currentFunction.result2!= null && !currentFunction.result2.equals("")) fxDesc += "Result 2<br>"+currentFunction.result2+"<br><br>";
							if (currentFunction.example3!= null && !currentFunction.example3.equals("")) fxDesc += "Example 3<br>"+currentFunction.example3+"<br>";
							if (currentFunction.result3!= null && !currentFunction.result3.equals("")) fxDesc += "Result 3<br>"+currentFunction.result3;
							
							json.add(fx);
							if (currentFunction.example1!=null) json.add(currentFunction.example1.replace("<br>", "\n"));
							else json.add(fx+fxAll);
							json.add("<div style='white-space: nowrap;'><span style='color:#995500'><b>"+fx+"</b></span>"+fxAll+"<hr>"+fxDesc+"</div>");
							json.add(e.getKey());
							
						}
						
					}
				}
	
				writer.write(json.toJSONString());
				
				
			}
			
		}
		
		Misc.create("docs"+File.separator+"mql.html", Misc.load("docs"+File.separator+"mql_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", "MQL").replace("[MQL]", MQLDocumentation.menu_to_html()).replace("[MQL_VERSION]", Start.version));
		Misc.create("docs"+File.separator+"index.html", Misc.load("docs"+File.separator+"index_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[MQL_VERSION]", Misc.atom(Start.version, 2, "-")));
		
		Doc_changelog.load();Misc.create("docs"+File.separator+"changelog.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("changelog")).get("title")+"").replace("[DOC]", Doc.to_html("changelog")).replace("[MQL_VERSION]", Start.version));
		Doc_faq.load();Misc.create("docs"+File.separator+"faq.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("faq")).get("title")+"").replace("[DOC]", Doc.to_html("faq")).replace("[MQL_VERSION]", Start.version));
		Doc_demo.load();Misc.create("docs"+File.separator+"demo.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("demo")).get("title")+"").replace("[DOC]", Doc.to_html("demo")).replace("[MQL_VERSION]", Start.version));
		Doc_limitation.load();Misc.create("docs"+File.separator+"limitation.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("limitation")).get("title")+"").replace("[DOC]", Doc.to_html("limitation")).replace("[MQL_VERSION]", Start.version));
		Doc_structure.load();Misc.create("docs"+File.separator+"structure.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("structure")).get("title")+"").replace("[DOC]", Doc.to_html("structure")).replace("[MQL_VERSION]", Start.version));
		Doc_storage_engine.load();Misc.create("docs"+File.separator+"storage_engine.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("storage_engine")).get("title")+"").replace("[DOC]", Doc.to_html("storage_engine")).replace("[MQL_VERSION]", Start.version));
		Doc_mental_process.load();Misc.create("docs"+File.separator+"mental_process.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("mental_process")).get("title")+"").replace("[DOC]", Doc.to_html("mental_process")).replace("[MQL_VERSION]", Start.version));
		Doc_emotion_manager.load();Misc.create("docs"+File.separator+"emotion_manager.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("emotion_manager")).get("title")+"").replace("[DOC]", Doc.to_html("emotion_manager")).replace("[MQL_VERSION]", Start.version));
		Doc_performance.load();Misc.create("docs"+File.separator+"performance.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("performance")).get("title")+"").replace("[DOC]", Doc.to_html("performance")).replace("[MQL_VERSION]", Start.version));
		Doc_license.load();Misc.create("docs"+File.separator+"license.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("license")).get("title")+"").replace("[DOC]", Doc.to_html("license")).replace("[MQL_VERSION]", Start.version));
		Doc_required.load();Misc.create("docs"+File.separator+"required.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("required")).get("title")+"").replace("[DOC]", Doc.to_html("required")).replace("[MQL_VERSION]", Start.version));
		Doc_download.load();Misc.create("docs"+File.separator+"download.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("download")).get("title")+"").replace("[DOC]", Doc.to_html("download")).replace("[MQL_VERSION]", Start.version));
		Doc_install.load();Misc.create("docs"+File.separator+"install.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("install")).get("title")+"").replace("[DOC]", Doc.to_html("install")).replace("[MQL_VERSION]", Start.version));
		Doc_config.load();Misc.create("docs"+File.separator+"config.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("config")).get("title")+"").replace("[DOC]", Doc.to_html("config")).replace("[MQL_VERSION]", Start.version));
		Doc_first_start.load();Misc.create("docs"+File.separator+"first_start.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("first_start")).get("title")+"").replace("[DOC]", Doc.to_html("first_start")).replace("[MQL_VERSION]", Start.version));
		Doc_clients.load();Misc.create("docs"+File.separator+"clients.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("clients")).get("title")+"").replace("[DOC]", Doc.to_html("clients")).replace("[MQL_VERSION]", Start.version));
		Doc_ai_mql.load();Misc.create("docs"+File.separator+"mql_ai_mode.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("mql_ai")).get("title")+"").replace("[DOC]", Doc.to_html("mql_ai")).replace("[MQL_VERSION]", Start.version));
		Doc_language.load();Misc.create("docs"+File.separator+"language.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("language")).get("title")+"").replace("[DOC]", Doc.to_html("language")).replace("[MQL_VERSION]", Start.version));
		Doc_word.load();Misc.create("docs"+File.separator+"word.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("word")).get("title")+"").replace("[DOC]", Doc.to_html("word")).replace("[MQL_VERSION]", Start.version));
		Doc_merge.load();Misc.create("docs"+File.separator+"merge.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("merge")).get("title")+"").replace("[DOC]", Doc.to_html("merge")).replace("[MQL_VERSION]", Start.version));
		Doc_to_do.load();Misc.create("docs"+File.separator+"to_do.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("to_do")).get("title")+"").replace("[DOC]", Doc.to_html("to_do")).replace("[MQL_VERSION]", Start.version));
		Doc_chat_room.load();Misc.create("docs"+File.separator+"chat_room.html", Misc.load("docs"+File.separator+"doc_page_tmp.html").replace("[SERVER_YEAR]", Start.copyright).replace("[PAGE_TITLE]", ((JSONObject) Doc.pages.get("chat_room")).get("title")+"").replace("[DOC]", Doc.to_html("chat_room")).replace("[MQL_VERSION]", Start.version));
		
		mqlPage_to_html();
		
	}

}